Package: crystract
Type: Package
Title: Crystallographic Information File (CIF) Data Processing Tools
Version: 1.0.1
Authors@R: c(person(given = "Don", 
                  family = "Ngo", 
                  email = "dngo@carnegiescience.edu", 
                  role = c("aut", "cre"),
                  comment = c(ORCID = "0009-0001-2779-2146")),
           person(given = "Anirudh", 
                  family = "Prabhu", 
                  email = "aprabhu@carnegiescience.edu", 
                  role = c("aut"),
                  comment = c(ORCID = "0000-0002-9921-6084")),
           person(given = "Julia-Maria", 
                  family = "Hubner", 
                  email = "julia-maria.huebner@tu-dresden.de",
                  role = c("aut"),
                  comment = c(ORCID = "0000-0003-2048-6629"))
                  )
Maintainer: Don Ngo <dngo@carnegiescience.edu>
Description: Provides a suite of functions to parse Crystallographic Information
    Files (.cif), extracting essential data such as chemical formulas, unit cell
    parameters, atomic coordinates, and symmetry operations. It also includes
    tools to calculate interatomic distances, identify bonded pairs using various
    algorithms (minimum_distance, brunner_nn_reciprocal, econ_nn, crystal_nn), determine nearest
    neighbor counts, and calculate bond angles. The package is designed to
    facilitate the preparation of crystallographic data for further analysis,
    including machine learning applications in materials science.
    Methods are described in:
    Brunner (1977) <doi:10.1107/S0567739477000461>;
    Hoppe (1979) <doi:10.1524/zkri.1979.150.14.23>;
    O'Keeffe (1979) <doi:10.1107/S0567739479001765>;
    Shannon (1976) <doi:10.1107/S0567739476001551>;
    Pan et al. (2021) <doi:10.1021/acs.inorgchem.0c02996>;
    Pauling (1960, ISBN:978-0801403330).
License: GPL (>= 3)
Encoding: UTF-8
RoxygenNote: 7.3.3
Depends: R (>= 3.5.0)
Imports: data.table, dplyr, stringr, future, future.apply, geometry
Suggests: DT, ggplot2, htmlwidgets, knitr, plotly, rmarkdown, testthat
        (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
LazyData: true
URL: https://prabhulab.github.io/ml-crystals/
NeedsCompilation: no
Packaged: 2026-05-28 17:41:44 UTC; donngo
Author: Don Ngo [aut, cre] (ORCID: <https://orcid.org/0009-0001-2779-2146>),
  Anirudh Prabhu [aut] (ORCID: <https://orcid.org/0000-0002-9921-6084>),
  Julia-Maria Hubner [aut] (ORCID:
    <https://orcid.org/0000-0003-2048-6629>)
Repository: CRAN
Date/Publication: 2026-05-28 18:30:02 UTC

# Generated by roxygen2: do not edit by hand

export(aggregate_batch_results)
export(analyze_cif_files)
export(analyze_single_cif)
export(apply_symmetry_operations)
export(brunner_nn_reciprocal)
export(calculate_angles)
export(calculate_distances)
export(calculate_expansion_factors)
export(calculate_neighbor_counts)
export(calculate_weighted_average_network_distance)
export(calculate_weighted_neighbor_counts)
export(covalent_radii)
export(crystal_nn)
export(econ_nn)
export(expand_transformed_coords)
export(export_analysis_to_csv)
export(extract_atomic_coordinates)
export(extract_chemical_formula)
export(extract_database_code)
export(extract_space_group_name)
export(extract_space_group_number)
export(extract_structure_type)
export(extract_symmetry_operations)
export(extract_unit_cell_metrics)
export(filter_atoms_by_symbol)
export(filter_by_elements)
export(filter_by_wyckoff_symbol)
export(filter_ghost_distances)
export(merge_sites_pbc)
export(minimum_distance)
export(propagate_angle_error)
export(propagate_distance_error)
export(read_cif_files)
export(set_radii_data)
export(voronoi_nn)
importFrom(data.table,":=")
importFrom(data.table,.N)
importFrom(data.table,.SD)
importFrom(data.table,as.data.table)
importFrom(data.table,copy)
importFrom(data.table,data.table)
importFrom(data.table,fread)
importFrom(data.table,rbindlist)
importFrom(data.table,set)
importFrom(data.table,setkey)
importFrom(data.table,setnames)
importFrom(dplyr,distinct)
importFrom(geometry,delaunayn)
importFrom(stats,as.dist)
importFrom(stats,cutree)
importFrom(stats,hclust)
importFrom(stats,na.omit)
importFrom(stringr,str_match)
importFrom(stringr,str_split_fixed)
importFrom(utils,combn)
importFrom(utils,globalVariables)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# crystract

<!-- badges: start -->

<!-- After setting up, uncomment and update these badges -->

<!-- [![Lifecycle: experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental) -->

<!-- [![R-CMD-check](https://github.com/PrabhuLab/ml-crystals/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/PrabhuLab/ml-crystals/actions/workflows/R-CMD-check.yaml) -->

<!-- badges: end -->

**crystract** provides a suite of functions to parse Crystallographic
Information Files (`.cif`), extracting essential data such as chemical
formulas, unit cell parameters, atomic coordinates, and symmetry
operations. It also includes tools to calculate interatomic distances,
identify bonded pairs using various algorithms (Minimum Distance,
Brunner’s, Hoppe’s, Voronoi, CrystalNN), determine nearest neighbor
counts, and calculate bond angles. All data is extracted into nested
`data.table`s, which can then be exported as an R Data Structure (RDS)
or folders of .csv files. The package is designed to facilitate the
preparation of crystallographic data for further analysis, including
machine learning applications in materials science.

> **Note on Repository Structure**
>
> The `crystract` package is located within the `packages/crystract/`
> subdirectory of the `PrabhuLab/ml-crystals` GitHub repository. You
> must use the `subdir` argument during installation, as shown below.

## Key Features

- **Efficient CIF Parsing**: Utilizes `data.table` for fast and robust
  extraction of metadata, unit cell parameters, atomic coordinates, and
  symmetry operations.
- **Symmetry and Supercell Generation**: Applies symmetry operations to
  generate a full unit cell from the asymmetric unit and expands
  coordinates into a 3x3x3 supercell for neighbor searching.
- **Geometric Calculations**: Computes interatomic distances using the
  metric tensor (correct for all crystal systems) and calculates bond
  angles.
- **Multiple Bonding Algorithms**: Implements several algorithms to
  identify bonded atoms, including the `minimum_distance` (default),
  `brunner`, `econ` (Hoppe’s), `voronoi`, and `crystal_nn` methods.
- **Rigorous Error Propagation**: Calculates and propagates experimental
  uncertainties from the CIF file into the final calculated bond lengths
  and angles.
- **Powerful Post-Processing Tools**: Includes functions to filter
  results by chemical element, Wyckoff site, or to remove non-physical
  “ghost” distances caused by site disorder using a customizable atomic
  radii table.
- **Batch Processing & Export**: The main `analyze_cif_files()` function
  is designed to process hundreds of files in a single run, and results
  can be easily exported to a structured directory of CSV files with
  `export_analysis_to_csv()`.

------------------------------------------------------------------------

## Workflow Overview

The following diagram illustrates the primary data pipeline in
`crystract`, from raw CIF input to final CSV export.

<div align="center">

<img src="man/figures/workflow_diagram.png" width="90%" alt="crystract Workflow Diagram">

</div>

------------------------------------------------------------------------

## Best Practices & Decision Logic

To assist researchers in configuring `crystract` for their specific
datasets, we provide the following decision trees for selecting atomic
radii and choosing the most appropriate bonding algorithm.

### 1. Choosing a Bonding Algorithm

When invoking `analyze_cif_files(..., bonding_algorithms = c(...))`, we
recommend choosing your target algorithm based on the chemical makeup of
your structure and the electronegativity differences ($\Delta EN$) of
the expected bonds.

<div align="center">

<img src="man/figures/bonding_logic.png" width="60%" alt="Bonding Algorithm Selection Logic">

</div>

### 2. Atomic Radii Selection

When utilizing functions like `filter_ghost_distances()` or algorithms
that rely on distance cutoffs, `crystract` employs an internal logic to
select the most appropriate atomic radius. You can also override this by
injecting your own custom radii dictionary via `set_radii_data()`.

<div align="center">

<img src="man/figures/radii_logic.png" width="70%" alt="Atomic Radii Selection Logic">

</div>

------------------------------------------------------------------------

## Quickstart: A Complete Workflow & Real Outputs

The `analyze_single_cif()` (and its batch counterpart
`analyze_cif_files()`) provides a complete, one-step workflow. Here we
run it on an example crystal structure included inside the package
itself, demonstrating the exact data outputs you can expect.

``` r
library(crystract)
library(data.table)

# 1. Load the built-in demo CIF file (Strontium Silicide)
cif_path <- system.file("extdata", "1590946.cif", package = "crystract")

# 2. Analyze the file
# This single function handles parsing, supercell expansion, geometric calculations, 
# bonding detection, and error propagation.
analysis_results <- analyze_single_cif(
  cif_path,
  bonding_algorithms = c("minimum_distance", "crystal_nn")
)
```

### 1. Extracted Metadata

The returned object is a single row `data.table` containing both
high-level metadata and list-columns storing the detailed extracted
measurements.

| database_code                    | chemical_formula | space_group_name |
|:---------------------------------|:-----------------|:-----------------|
| depnum_ccdc_archive CCDC 1590946 | Si1 Sr2          | P n m a          |

High-Level Crystal Information

### 2. Unit Cell Metrics

The parameters defining the size and shape of the unit cell are securely
parsed along with their experimental uncertainties (if available in the
CIF).

| \_cell_length_a | \_cell_length_b | \_cell_length_c | \_cell_angle_alpha | \_cell_angle_beta | \_cell_angle_gamma |
|---:|---:|---:|---:|---:|---:|
| 8.11 | 5.15 | 9.54 | 90 | 90 | 90 |

Extracted Unit Cell Parameters (Å and Degrees)

### 3. Detected Bonded Pairs

`crystract` identifies bonded pairs using your chosen algorithms. Below
is the output from the **Minimum Distance** algorithm. Notice the
rigorous propagation of experimental error (`DistanceError`).

| Atom1 | Atom2           | Distance | DistanceError |    Weight |
|:------|:----------------|---------:|--------------:|----------:|
| Si1   | Sr1_1_0_0_0     | 3.163544 |             0 | 1.0000000 |
| Si1   | Sr1_2_0_0_0     | 3.245310 |             0 | 0.9748050 |
| Si1   | Sr1_4_0\_-1\_-1 | 3.184477 |             0 | 0.9934267 |
| Si1   | Sr1_4_0_0\_-1   | 3.184477 |             0 | 0.9934267 |
| Si1   | Sr2_1_0_0\_-1   | 3.261366 |             0 | 0.9700058 |
| Si1   | Sr2_3\_-1\_-1_0 | 3.465249 |             0 | 0.9129342 |

Predicted Bonded Pairs (Minimum Distance Method)

### 4. Calculated Bond Angles

Using the metric tensor, all connected triplets are evaluated to
calculate the exact internal bond angles across the repeating periodic
boundaries.

| CentralAtom | Neighbor1   | Neighbor2       |     Angle | AngleError |
|:------------|:------------|:----------------|----------:|-----------:|
| Si1         | Sr1_1_0_0_0 | Sr1_2_0_0_0     | 109.37260 |          0 |
| Si1         | Sr1_1_0_0_0 | Sr1_4_0\_-1\_-1 | 125.55190 |          0 |
| Si1         | Sr1_1_0_0_0 | Sr1_4_0_0\_-1   | 125.55190 |          0 |
| Si1         | Sr1_1_0_0_0 | Sr2_1_0_0\_-1   | 129.28796 |          0 |
| Si1         | Sr1_1_0_0_0 | Sr2_3\_-1\_-1_0 |  69.08689 |          0 |
| Si1         | Sr1_1_0_0_0 | Sr2_3\_-1_0_0   |  69.08689 |          0 |

Calculated Interatomic Angles

------------------------------------------------------------------------

### Data Dictionary

Here is a comprehensive overview of the columns generated in the Master
Analysis Object and its nested tables.

#### 1. Master Analysis Object

| Column Name | Data Type | Description |
|:---|:---|:---|
| `file_name` | Character | The name of the processed CIF file. |
| `database_code` | Character | The unique identifier from the source database. |
| `chemical_formula` | Character | The chemical sum formula extracted from the CIF. |
| `structure_type` | Character | The name of the structure type. |
| `space_group_name` | Character | Hermann-Mauguin space group symbol. |
| `space_group_number` | Character | International Tables space group number. |
| `unit_cell_metrics` | List (DT) | Nested table containing lattice parameters. |
| `atomic_coordinates` | List (DT) | Nested table of primary asymmetric atoms. |
| `symmetry_operations` | List (DT) | Nested table of symmetry operators. |
| `transformed_coords` | List (DT) | Nested table of the full unit cell atoms. |
| `expanded_coords` | List (DT) | Nested table of the supercell (3x3x3) atoms. |
| `distances` | List (DT) | Nested table of all calculated interatomic distances. |
| `bonded_pairs_*` | List (DT) | Nested table of bonds detected via requested methods (e.g. `_minimum_distance`). |
| `neighbor_counts_*` | List (DT) | Nested table of coordination numbers for requested methods. |
| `bond_angles_*` | List (DT) | Nested table of calculated bond angles for requested methods. |

#### 2. Atomic Coordinates (Nested)

| Column Name | Data Type | Description |
|:---|:---|:---|
| `Label` | Character | Unique atom label (e.g., “Fe1”). |
| `WyckoffSymbol` | Character | The Wyckoff letter (e.g., “c”). |
| `WyckoffMultiplicity` | Numeric | The site multiplicity (e.g., 4). |
| `Occupancy` | Numeric | Site occupancy factor (0.0 to 1.0). |
| `x_a`, `y_b`, `z_c` | Numeric | Fractional coordinates along axis $a, b, c$. |
| `*_error` | Numeric | Standard uncertainties for coordinates. |

#### 3. Bonded Pairs (Nested)

| Column Name | Data Type | Description |
|:---|:---|:---|
| `Atom1` | Character | Label of the central atom (from the asymmetric unit). |
| `Atom2` | Character | Label of the neighbor atom (from the expanded supercell). |
| `Distance` | Numeric | Calculated Euclidean distance in Angstroms (Å). |
| `DistanceError` | Numeric | Propagated standard uncertainty of the distance. |
| `DeltaX`, `DeltaY`, `DeltaZ` | Numeric | Difference in fractional coordinates ($x_1 - x_2$). |
| `Weight` | Numeric | Calculated bond weight/strength depending on the algorithm. |

## Licensing

`crystract` is offered under a dual-license model to accommodate a
variety of use cases:

- **For Open-Source Projects:** The package is licensed under the **GNU
  General Public License v3.0 (GPL-3.0)**. If you are developing other
  open-source software, you are free to use, modify, and distribute
  `crystract` under the terms of the GPL-3.0.

- **For Commercial Use:** If you wish to use `crystract` in a commercial
  product, for commercial services, or for any other commercial purpose,
  you must obtain a separate commercial license. Please contact the
  package maintainer to arrange the terms.

## Installation

Installing `crystract` involves a few steps, as it is currently hosted
on GitHub. We use the `remotes` package to facilitate installation
directly from the repository.

### Prerequisites

- Install the latest version of **[R](https://cran.r-project.org/)**.
- Install the free **[RStudio Desktop
  IDE](https://posit.co/download/rstudio-desktop/)**.

### Installation Steps

Open R or RStudio and run the following commands:

``` r
# First, ensure you have the remotes package
install.packages("remotes")

# Install crystract from the GitHub repository
remotes::install_github("PrabhuLab/ml-crystals", subdir = "packages/crystract", build_vignettes = TRUE)
```

### Verifying the Installation

To make sure the package was installed correctly, load it into your R
session.

``` r
library(crystract)
```

If this command runs without any errors, the installation was
successful.

## Learning More

For a detailed, step-by-step guide explaining each function, the
crystallographic principles, and the formulas used for calculations,
please see the package vignette.

You can access it with the following command after you have successfully
installed the package:

``` r
# This command opens the detailed package guide
vignette("crystract")
```

## Community Guidelines

We welcome and appreciate all forms of community engagement. To ensure a
smooth and productive collaboration, we have established guidelines for
contributing, reporting issues, and seeking support.

All participants in this project are expected to abide by our **[Code of
Conduct](https://github.com/PrabhuLab/ml-crystals/blob/main/packages/crystract/CODE_OF_CONDUCT.md)**.
Please read it to understand the standards of behavior we expect.

For detailed instructions on how to contribute to the software, report
bugs, or suggest new features, please review our **[Contributing
Guidelines](https://github.com/PrabhuLab/ml-crystals/blob/main/packages/crystract/CONTRIBUTING.md)**.

## Author

**Author and Maintainer:** Don Ngo (`dngo@carnegiescience.edu`)

## ----include = FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
# Save current options and apply new ones for the vignette rendering
old_options <- options(width = 75, datatable.print.width = 80)

## ----setup, message=FALSE, warning=FALSE---------------------------------
library(crystract)
library(data.table)

## ----full-pipeline-with-comments, eval=TRUE------------------------------
# IMPORTANT: Update this path to point to your own downloaded CIF file.
# 1. Try to find the file in the installed package
cif_path <- system.file("extdata", "1590946.cif", package = "crystract")

# 2. If that fails, look in the source directory
if (cif_path == "") {
  cif_path <- "../inst/extdata/1590946.cif"
}

# 3. Final check to provide a clear error if both fail
if (!file.exists(cif_path)) {
  stop("Could not find 1590946.cif in installed package or inst/extdata/")
}

# Run the pipeline on our single example file, extracting multiple bonding types at once
analysis_results <- analyze_single_cif(
  cif_path,
  bonding_algorithms = c("minimum_distance", "brunner", "econ", "voronoi", "crystal_nn")
)

# Let's inspect the structure of the output table.
# It's a single row containing all our results in nested data.tables.
str(analysis_results, max.level = 2)

## ----access-nested-data, eval=FALSE--------------------------------------
# # The result is a list-column, so we access the element with [[1]]
# final_bonds <- analysis_results$bonds_minimum_distance[[1]]
# 
# print(head(final_bonds))

## ----load-data, eval=TRUE------------------------------------------------
# The path was defined in the previous section:
cif_data_list <- read_cif_files(cif_path)

# We'll work with the content of the first file
cif_content <- cif_data_list[[1]]

# Let's look at the first few lines of the raw data
knitr::kable(
  head(cif_content),
  caption = "First 6 lines of the raw CIF data."
)

## ----extract-metadata, eval=TRUE-----------------------------------------
database_code <- extract_database_code(cif_content)
chemical_formula <- extract_chemical_formula(cif_content)
space_group_name <- extract_space_group_name(cif_content)
space_group_number <- extract_space_group_number(cif_content)

cat("Database Code:", database_code, "\n")
cat("Chemical Formula:", chemical_formula, "\n")
cat("Space Group:", space_group_name, "(No.", space_group_number, ")\n")

## ----extract-metrics, eval=TRUE------------------------------------------
unit_cell_metrics <- extract_unit_cell_metrics(cif_content)
print(unit_cell_metrics)

## ----extract-coords-symm, eval=TRUE--------------------------------------
# Extract the coordinates of the unique atoms in the asymmetric unit
atomic_coordinates <- extract_atomic_coordinates(cif_content)
print("Asymmetric Atomic Coordinates:")
print(atomic_coordinates)

# Extract the symmetry operations
symmetry_operations <- extract_symmetry_operations(cif_content)
print("Symmetry Operations (first 6 of 8):")
print(head(symmetry_operations))

## ----generate-structure, eval=TRUE---------------------------------------
# Apply symmetry to generate all atoms in the primary unit cell
transformed_coords <- apply_symmetry_operations(atomic_coordinates, symmetry_operations, unit_cell_metrics)
print("Unique atoms in full unit cell (first 6):")
print(head(transformed_coords))

# Expand into a supercell for neighbor calculations
expanded_coords <- expand_transformed_coords(transformed_coords)
print("Atoms in supercell (first 6):")
print(head(expanded_coords))

## ----calc-distances, eval=TRUE-------------------------------------------
distances <- calculate_distances(atomic_coordinates, expanded_coords, unit_cell_metrics)
print("Calculated Distances (shortest 6):")
print(head(distances[order(Distance)]))

## ----calc-bonding-neighbors, eval=TRUE-----------------------------------
# 1. Minimum Distance
bonds_min <- minimum_distance(distances, delta = 0.1)

# 2. Brunner's Method
bonds_brunner <- brunner_nn_reciprocal(distances)

# 3. Hoppe's ECoN
bonds_econ <- econ_nn(distances, atomic_coordinates)

# 4. Voronoi Tessellation
bonds_voronoi <- voronoi_nn(atomic_coordinates, expanded_coords, unit_cell_metrics)

# 5. CrystalNN
bonds_crystal_nn <- crystal_nn(distances, atomic_coordinates, expanded_coords, unit_cell_metrics)

cat("Minimum Distance found", nrow(bonds_min), "bonds.\n")
cat("CrystalNN found", nrow(bonds_crystal_nn), "bonds.\n")

# Calculate integer neighbor counts based on the bonded pairs (e.g., using CrystalNN)
neighbor_counts <- calculate_neighbor_counts(bonds_crystal_nn)
print("CrystalNN Neighbor Counts:")
print(neighbor_counts)

## ----calc-angles, eval=TRUE----------------------------------------------
bond_angles <- calculate_angles(
  bonds_min,
  atomic_coordinates,
  expanded_coords,
  unit_cell_metrics
)
print("Calculated Bond Angles (first 6):")
print(head(bond_angles))

## ----propagate-errors, eval=TRUE-----------------------------------------
# Propagate errors for interatomic distances
bonded_pairs_with_error <- propagate_distance_error(
  bonds_min,
  atomic_coordinates,
  unit_cell_metrics
)
print("Bonded Pairs with Distance Error (first 6):")
print(head(bonded_pairs_with_error))

# Propagate errors for bond angles
bond_angles_with_error <- propagate_angle_error(
  bond_angles,
  atomic_coordinates,
  expanded_coords,
  unit_cell_metrics
)
print("Bond Angles with Angle Error (first 6):")
print(head(bond_angles_with_error))

## ----eval=FALSE----------------------------------------------------------
# # In an interactive R session, you would run this:
# filtered_bonds <- filter_atoms_by_symbol(
#   data_table = bonded_pairs_with_error,
#   atom_col = "Atom1" # Filter based on the central atom
# )

## ----filter-by-wyckoff, eval=TRUE----------------------------------------
# 1. In our example, the asymmetric atoms do not have Wyckoff information from the CIF.
# We will mock them as "4c" for this demonstration to show how the function works.
atomic_coordinates[, WyckoffSymbol := c("c", "c", "c")]
atomic_coordinates[, WyckoffMultiplicity := c(4, 4, 4)]

print("Atomic coordinates showing Wyckoff information:")
print(atomic_coordinates[, .(Label, WyckoffSymbol, WyckoffMultiplicity)])
cat("\n")

# 2. Filter bonds where the central atom is on the "4c" Wyckoff site.
bonds_from_4c_site <- filter_by_wyckoff_symbol(
  data_table = bonded_pairs_with_error,
  atomic_coordinates = atomic_coordinates,
  atom_col = "Atom1",
  wyckoff_symbols = "4c"
)

print(paste("Number of rows in original bond table:", nrow(bonded_pairs_with_error)))
print(paste("Number of rows after filtering for site '4c':", nrow(bonds_from_4c_site)))

## ----filter-ghost-distances, eval=TRUE-----------------------------------
# A distance d is kept if: (r1+r2)*(1-margin) <= d <= (r1+r2)*(1+margin)
filtered_result <- filter_ghost_distances(
    distances = distances,
    atomic_coordinates = atomic_coordinates,
    margin = 0.1 # Default margin is 10%
)

kept_distances <- filtered_result$kept
removed_distances <- filtered_result$removed

cat("Total distances calculated:", nrow(distances), "\n")
cat("Distances kept after filtering:", nrow(kept_distances), "\n")
cat("Unlikely / non-bonded distances removed:", nrow(removed_distances), "\n\n")

print("Subset of removed distances (showing physically impossible / too long connections):")
print(head(removed_distances))

## ----filter-by-elements, eval=TRUE---------------------------------------
# Let's filter our bond table to exclude any bonds involving Strontium ("Sr").
# Since all bonds in this structure are Si-Sr, the result should be an empty table.
bonds_without_sr <- filter_by_elements(
    distances = bonded_pairs_with_error,
    atomic_coordinates = atomic_coordinates,
    elements_to_exclude = "Sr"
)

cat("Number of bonds in original table:", nrow(bonded_pairs_with_error), "\n")
cat("Number of bonds after excluding 'Sr':", nrow(bonds_without_sr), "\n")

## ----calculate-weighted-average-distance, eval=TRUE----------------------
# Calculate the weighted average BOND distance for the entire Sr2Si network.
# First, identify the bonds in the structure. We use the `bonds_min` table
# created in section 1.2.6.

# Then, define the Wyckoff sites belonging to the network. Here, it's just "4c".
# (We assigned dummy "4c" and "4a" Wyckoff positions to our coordinates in section 2.2).
network_wyckoff_sites <- "4c"

# Apply the function to the table of identified bonds.
# For this simple, ordered structure, all occupancies are 1.0, but the function
# correctly applies the full formula.
weighted_avg_bond_dist <- calculate_weighted_average_network_distance(
    distances = bonds_min, # Use the bond table as input
    atomic_coordinates = atomic_coordinates,
    wyckoff_symbols = network_wyckoff_sites
)

cat("Weighted average network bond distance for the '4c' sites:", weighted_avg_bond_dist, "Å\n")

## ----custom-radii-and-batch, eval=FALSE----------------------------------
# # --- 1. Define and set the custom radii dictionary ---
# radii_dict <- c(
#   Ac=2.15, Ag=1.45, Al=1.21, Am=1.8, Ar=1.06, As=1.19, At=1.5, Au=1.36, B=0.84,
#   Ba=2.15, Be=0.96, Bi=1.48, Br=1.2, C=0.73, Ca=1.76, Cd=1.44, Ce=2.04, Cl=1.02,
#   Cm=1.69, Co=1.38, Cr=1.39, Cs=2.44, Cu=1.32, Dy=1.92, Er=1.89, Eu=1.98, F=0.57,
#   Fe=1.42, Fr=2.6, Ga=1.22, Gd=1.96, Ge=1.2, H=0.31, He=0.28, Hf=1.75, Hg=1.32,
#   Ho=1.92, I=1.39, In=1.42, Ir=1.41, K=2.03, Kr=1.16, La=2.07, Li=1.28, Lu=1.87,
#   Mg=1.41, Mn=1.5, Mo=1.54, N=0.71, Na=1.66, Nb=1.64, Nd=2.01, Ne=0.58, Ni=1.24,
#   Np=1.9, O=0.66, Os=1.44, P=1.07, Pa=2, Pb=1.46, Pd=1.39, Pm=1.99, Po=1.4, Pr=2.03,
#   Pt=1.36, Pu=1.87, Ra=2.21, Rb=2.2, Re=1.51, Rh=1.42, Rn=1.5, Ru=1.46, S=1.05,
#   Sb=1.39, Sc=1.7, Se=1.2, Si=1.11, Sm=1.98, Sn=1.39, Sr=1.95, Ta=1.7, Tb=1.94,
#   Tc=1.47, Te=1.38, Th=2.06, Ti=1.6, Tl=1.45, Tm=1.9, U=1.96, V=1.53, W=1.62,
#   Xe=1.4, Y=1.9, Yb=1.87, Zn=1.22, Zr=1.75
# )
# 
# custom_radii_table <- data.table(
#   Symbol = names(radii_dict),
#   Radius = as.numeric(radii_dict),
#   Type   = "covalent"
# )
# 
# # Inject the custom table into the current crystract session
# set_radii_data(custom_radii_table)
# 
# # --- 2. Run the batch analysis ---
# cat("Starting batch analysis on CIF files...\n")
# results <- analyze_cif_files(
#   file_paths = "path/to/cif_directory",
#   tolerance = 1e-4,
#   bonding_algorithms = c("minimum_distance", "brunner", "econ", "voronoi", "crystal_nn"),
#   calculate_bond_angles = FALSE,       # Skip angles to speed up extraction
#   perform_error_propagation = FALSE,   # Skip uncertainties
#   workers = 4                          # Adjust based on your available CPU cores
# )
# 
# # --- 3. Extract and Save Coordination Numbers to CSV ---
# # Create a dedicated folder for the outputs to avoid clutter
# output_dir <- "path/to/output_directory"
# if (!dir.exists(output_dir)) dir.create(output_dir)
# 
# # Identify all coordination number columns generated by the pipeline
# cn_columns <- grep("^cn_", names(results), value = TRUE)
# 
# cat("\nExtracting coordination numbers and saving to CSV...\n")
# 
# for (col in cn_columns) {
# 
#   # Extract the list of data.tables for this specific algorithm
#   list_of_dts <- results[[col]]
# 
#   # Associate each table with its CIF file name
#   names(list_of_dts) <- results$file_name
# 
#   # Unnest and bind all tables together into one giant table
#   combined_dt <- rbindlist(list_of_dts, idcol = "file_name", fill = TRUE)
# 
#   if (nrow(combined_dt) > 0) {
#     # Format the file path (e.g., "cn_crystal_nn_crystract.csv")
#     file_path <- file.path(output_dir, paste0(col, "_crystract.csv"))
# 
#     # Save to CSV
#     fwrite(combined_dt, file_path)
# 
#     cat(sprintf("Saved %d rows across %d files to: %s\n",
#                 nrow(combined_dt),
#                 uniqueN(combined_dt$file_name),
#                 basename(file_path)))
#   } else {
#     cat(sprintf("No valid coordination data found for %s, skipping.\n", col))
#   }
# }
# 
# cat("\nAll operations finished successfully! Files are in:", output_dir, "\n")

## ----include = FALSE----------------------------------------------------------
# Restore original options as required by CRAN
options(old_options)

#' crystract: Tools to Extract Data from CIF Files for Crystallography
#'
#' @description
#' Parse Crystallographic Information Files (.cif) and extract essential
#' data such as chemical formulas, unit cell parameters, atomic coordinates,
#' and symmetry operations. Calculate interatomic distances, identify bonded
#' pairs using various algorithms (Minimum Distance, Brunner's, EconNN,
#' CrystalNN, Voronoi), determine nearest neighbor counts, and calculate bond
#' angles. Facilitates the preparation of crystallographic data for further
#' analysis, including machine learning applications in materials science.
#'
#' @references
#' Shannon, R. D. (1976). "Revised effective ionic radii and systematic studies of
#' interatomic distances in halides and chalcogenides." \emph{Acta Crystallographica
#' Section A: Crystal Physics, Diffraction, Theoretical and General Crystallography},
#' 32(5), 751-767.
#'
#' Pauling, L. (1960). \emph{The Nature of the Chemical Bond and the Structure of
#' Molecules and Crystals: An Introduction to Modern Structural Chemistry} (Vol. 18).
#' Cornell University Press.
#'
#' @keywords internal
"_PACKAGE"
#'
#' @importFrom data.table .N := .SD as.data.table copy data.table fread rbindlist set setkey setnames
#' @importFrom dplyr distinct
#' @importFrom stats as.dist cutree hclust na.omit
#' @importFrom geometry delaunayn
#' @importFrom stringr str_match str_split_fixed
#' @importFrom utils combn globalVariables
NULL

#' @title Identify Atomic Bonds using the Minimum Distance Method
#' @description Identifies bonded atoms by finding the nearest neighbor distance (d_min)
#'   for each central atom and defining a cutoff distance (d_cut) as
#'   d_cut = (1 + delta) * d_min.
#' @param distances A `data.table` of interatomic distances from `calculate_distances`.
#' @param delta The relative tolerance parameter (default 0.1).
#' @return A `data.table` of bonded pairs.
#' @family bonding algorithms
#' @export
#' @examples
#' dists <- data.table::data.table(Atom1 = c("A", "A"), Atom2 = c("B", "C"),
#'                                 Distance = c(1.5, 2.5),
#'                                 DeltaX = c(1, 0), DeltaY = c(0, 1), DeltaZ = c(0, 0))
#' minimum_distance(dists, delta = 0.1)
minimum_distance <- function(distances, delta = 0.1) {
  if (nrow(distances) == 0)
    return(distances)

  dmin <- distances[, .(dmin = min(Distance)), by = .(Atom1)]
  dmin[, dcut := dmin * (1 + delta)]

  bonded_pairs <- distances[dmin, on = .(Atom1)]
  bonded_pairs <- bonded_pairs[Distance <= dcut, .(Atom1,
                                                   Atom2,
                                                   Distance,
                                                   DeltaX,
                                                   DeltaY,
                                                   DeltaZ,
                                                   Weight = dmin / Distance)]
  return(bonded_pairs)
}

#' @title Identify Atomic Bonds using Brunner's Method (Reciprocal)
#' @description An alternative bonding detection method using the
#'   largest reciprocal gap. Matches `pymatgen.analysis.local_env.BrunnerNNReciprocal`.
#' @param distances A `data.table` of interatomic distances.
#' @param tol A small distance offset (default 0.0) to add to the cutoff.
#' @return A `data.table` of bonded pairs.
#' @references
#' Brunner, G. O. (1977). "A Definition of Coordination and Its Relevance in the
#' Structure Types AlB2 and NiAs." \emph{Acta Crystallographica Section A}, 33(1),
#' 226–227. \doi{10.1107/S0567739477000461}
#' @family bonding algorithms
#' @export
#' @examples
#' dists <- data.table::data.table(Atom1 = c("A", "A"), Atom2 = c("B", "C"),
#'                                 Distance = c(1.5, 2.5),
#'                                 DeltaX = c(1, 0), DeltaY = c(0, 1), DeltaZ = c(0, 0))
#' brunner_nn_reciprocal(dists)
brunner_nn_reciprocal <- function(distances, tol = 0.0) {
  if (nrow(distances) == 0)
    return(distances)

  dt <- copy(distances)[order(Atom1, Distance)]

  # Calculate Reciprocal Gaps
  # Use fill=NA to prevent misalignment at group boundaries
  dt[, RecipGap := (1 / Distance) - (1 / data.table::shift(Distance, type = "lead", fill = NA)), by = Atom1]

  # Handle single neighbors where gap is NA
  dt[is.na(RecipGap), RecipGap := -Inf]

  dt[, IsMaxGap := (RecipGap == max(RecipGap, na.rm = TRUE)), by = Atom1]

  # Extract cutoff distance (distance of the atom BEFORE the gap)
  cutoff_map <- dt[IsMaxGap == TRUE, .(CutoffDist = Distance[1]), by = Atom1]

  result <- distances[cutoff_map, on = "Atom1"]
  result <- result[Distance <= CutoffDist + tol]

  result[, MinDist := min(Distance), by = Atom1]
  result[, Weight := MinDist / Distance]

  return(result[, .(Atom1, Atom2, Distance, DeltaX, DeltaY, DeltaZ, Weight)])
}

#' @title Identify Atomic Bonds using Hoppe's EconNN Method
#' @description Uses Effective Coordination Numbers (ECoN) and Mean Fictive Ionic Radii (MEFIR).
#' @param distances A `data.table` of interatomic distances.
#' @param atomic_coordinates A `data.table` of atomic coordinates used to map species to radii.
#' @param tol A bond strength cutoff (default 0.2).
#' @param use_fictive_radius Logical. If `TRUE`, calculates Hoppe's fictive ionic radius.
#' @return A `data.table` of bonded pairs.
#' @references
#' Hoppe, R. (1979). "Effective Coordination Numbers (ECoN) and Mean Fictive Ionic
#' Radii (MEFIR)." \emph{Zeitschrift Für Kristallographie}, 150(1–4), 23–52.
#' \doi{10.1524/zkri.1979.150.14.23}
#'
#' Shannon, R. D. (1976). "Revised Effective Ionic Radii and Systematic Studies of
#' Interatomic Distances in Halides and Chalcogenides." \emph{Acta Crystallographica
#' Section A}, 32(5), 751–767. \doi{10.1107/S0567739476001551}
#' @family bonding algorithms
#' @export
#' @examples
#' dists <- data.table::data.table(Atom1 = c("Si1", "Si1"), Atom2 = c("O1", "O2"),
#'                                 Distance = c(1.6, 2.0),
#'                                 DeltaX = c(1, 0), DeltaY = c(0, 1), DeltaZ = c(0, 0))
#' ac <- data.table::data.table(Label = c("Si1", "O1", "O2"),
#'                              OxidationState = c(4, -2, -2))
#' econ_nn(dists, ac)
econ_nn <- function(distances,
                    atomic_coordinates,
                    tol = 0.2,
                    use_fictive_radius = FALSE) {
  if (nrow(distances) == 0)
    return(distances)

  dt <- copy(distances)

  if (use_fictive_radius) {
    atom_info <- copy(atomic_coordinates)
    if (!"OxidationState" %in% names(atom_info))
      atom_info[, OxidationState := NA_real_]

    get_r_ionic <- Vectorize(get_radius_pymatgen_style,
                             vectorize.args = c("symbol", "oxidation_state"))
    get_r_def <- Vectorize(function(sym)
      get_radius_pymatgen_style(sym, NA_real_, 6),
      vectorize.args = "sym")

    dt[, `:=`(
      Sym1 = sub("[0-9].*", "", sub("_.*", "", Atom1)),
      Sym2 = sub("[0-9].*", "", sub("_.*", "", Atom2)),
      Parent1 = sub("_.*", "", Atom1),
      Parent2 = sub("_.*", "", Atom2)
    )]

    dt[atom_info, on = c("Parent1" = "Label"), OS1 := i.OxidationState]
    dt[atom_info, on = c("Parent2" = "Label"), OS2 := i.OxidationState]

    dt[, R1 := get_r_ionic(Sym1, OS1, cn = 6)]
    dt[, R2 := get_r_ionic(Sym2, OS2, cn = 6)]

    # Fallbacks for fictive radius math
    dt[R1 == 0, R1 := get_r_def(Sym1)]
    dt[R2 == 0, R2 := get_r_def(Sym2)]

    dt[, RadiusSum := R1 + R2]
    dt[, FictiveR := ifelse(RadiusSum > 0, Distance * (R1 / RadiusSum), 0)]
  } else {
    dt[, FictiveR := Distance]
  }

  calc_mefir_group <- function(fictive_r) {
    min_v <- min(fictive_r, na.rm = TRUE)
    if (min_v <= 0 || !is.finite(min_v))
      return(0)

    mean_fir <- min_v
    for (i in 1:100) {
      prev_mean_fir <- mean_fir
      ratio <- (fictive_r / prev_mean_fir)^6
      w <- exp(1 - ratio)
      sum_w <- sum(w, na.rm = TRUE)
      if (sum_w == 0)
        return(prev_mean_fir)
      mean_fir <- sum(fictive_r * w, na.rm = TRUE) / sum_w
      if (abs(mean_fir - prev_mean_fir) < 1e-4)
        break
    }
    return(mean_fir)
  }

  mefir_vals <- dt[, .(MEFIR = calc_mefir_group(FictiveR)), by = Atom1]
  dt[mefir_vals, on = "Atom1", MEFIR := i.MEFIR]
  dt[, Weight := ifelse(MEFIR > 0, exp(1 - (FictiveR / MEFIR)^6), 0)]

  return(dt[Weight > tol, .(Atom1, Atom2, Distance, DeltaX, DeltaY, DeltaZ, Weight)])
}

#' @title Identify Atomic Bonds using Voronoi Tessellation
#' @description Performs 3D Voronoi analysis on the supercell.
#' @param atomic_coordinates A `data.table` of the primary (asymmetric) atom set.
#' @param expanded_coords A `data.table` of atoms in the expanded supercell.
#' @param unit_cell_metrics A `data.table` with cell parameters.
#' @param cutoff Distance cutoff (default 13.0).
#' @param tol Tolerance for solid angle weights (default 0).
#' @return A `data.table` of bonded pairs.
#' @references
#' O'Keeffe, M. (1979). "A Proposed Rigorous Definition of Coordination Number."
#' \emph{Acta Crystallographica Section A}, 35(5), 772–775. \doi{10.1107/S0567739479001765}
#'
#' Aurenhammer, F., Klein, R., & Lee, D.-T. (2013). \emph{Voronoi Diagrams and
#' Delaunay Triangulations}. World Scientific Publishing Company.
#' @family bonding algorithms
#' @export
#' @examples
#' \donttest{
#' cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
#' if (file.exists(cif_path)) {
#'   old_threads <- data.table::setDTthreads(2)
#'   res <- analyze_single_cif(cif_path, bonding_algorithms = "voronoi")
#'   print(head(res$bonds_voronoi[[1]]))
#'   data.table::setDTthreads(old_threads)
#' }
#' }
voronoi_nn <- function(atomic_coordinates,
                       expanded_coords,
                       unit_cell_metrics,
                       cutoff = 13.0,
                       tol = 0) {
  if (!requireNamespace("geometry", quietly = TRUE))
    stop("The 'geometry' package is required.")

  # 1. Convert Supercell to Cartesian Coords
  a <- as.numeric(unit_cell_metrics$`_cell_length_a`)
  b <- as.numeric(unit_cell_metrics$`_cell_length_b`)
  c <- as.numeric(unit_cell_metrics$`_cell_length_c`)
  alpha <- as.numeric(unit_cell_metrics$`_cell_angle_alpha`) * pi / 180
  beta <- as.numeric(unit_cell_metrics$`_cell_angle_beta`) * pi / 180
  gamma <- as.numeric(unit_cell_metrics$`_cell_angle_gamma`) * pi / 180
  v_sq <- 1 - cos(alpha)^2 - cos(beta)^2 - cos(gamma)^2 + 2 * cos(alpha) * cos(beta) * cos(gamma)
  v <- sqrt(max(0, v_sq))

  M <- matrix(
    c(
      a,
      b * cos(gamma),
      c * cos(beta),
      0,
      b * sin(gamma),
      c * (cos(alpha) - cos(beta) * cos(gamma)) / sin(gamma),
      0,
      0,
      c * v / sin(gamma)
    ),
    nrow = 3,
    byrow = TRUE
  )

  coords_mat <- as.matrix(expanded_coords[, .(x_a, y_b, z_c)]) %*% t(M)
  if (anyNA(coords_mat))
    return(NULL)

  # 2. Delaunay Tessellation ("QJ" is crucial to prevent crashes on symmetric grids)
  tess <- tryCatch({
    geometry::delaunayn(coords_mat, options = "QJ Pp")
  }, error = function(e)
    NULL)
  if (is.null(tess))
    return(NULL)

  # 3. Calculate Circumcenters
  n_tets <- nrow(tess)
  circumcenters <- matrix(NA_real_, nrow = n_tets, ncol = 3)
  for (i in 1:n_tets) {
    circumcenters[i, ] <- calculate_circumcenter(coords_mat[tess[i, ], ])
  }
  atom_to_tets <- split(rep(1:n_tets, 4), as.vector(tess))

  # 4. Identify Asymmetric Unit Atoms
  asy_frac <- as.matrix(atomic_coordinates[, .(x_a, y_b, z_c)])
  exp_frac <- as.matrix(expanded_coords[, .(x_a, y_b, z_c)])
  target_indices <- integer(nrow(asy_frac))

  for (i in 1:nrow(asy_frac)) {
    diffs <- t(t(exp_frac) - asy_frac[i, ])
    dists_sq <- rowSums(diffs^2)
    match_idx <- which.min(dists_sq)
    if (dists_sq[match_idx] < 1e-5)
      target_indices[i] <- match_idx
    else
      target_indices[i] <- NA
  }
  target_indices <- na.omit(target_indices)

  results_list <- list()

  # 5. Loop over Asymmetric Unit Atoms
  for (center_idx in target_indices) {
    center_coords <- coords_mat[center_idx, ]
    full_label <- expanded_coords$Label[center_idx]
    prim_label <- sub("_[0-9-]+_[0-9-]+_[0-9-]+$", "", full_label)

    my_tets <- atom_to_tets[[as.character(center_idx)]]
    if (is.null(my_tets))
      next

    potential_neighbors <- unique(as.vector(tess[my_tets, ]))
    potential_neighbors <- potential_neighbors[potential_neighbors != center_idx]
    if (length(potential_neighbors) == 0)
      next

    dists <- sqrt(rowSums((t(
      t(coords_mat[potential_neighbors, , drop = FALSE]) - center_coords
    ))^2))
    valid_mask <- dists <= cutoff
    neighbor_indices <- potential_neighbors[valid_mask]

    for (neigh_idx in neighbor_indices) {
      shared_tets <- intersect(my_tets, atom_to_tets[[as.character(neigh_idx)]])
      if (length(shared_tets) < 3)
        next

      raw_verts <- circumcenters[shared_tets, , drop = FALSE]
      raw_verts <- na.omit(raw_verts)
      if (nrow(raw_verts) < 3)
        next

      normal <- coords_mat[neigh_idx, ] - center_coords
      dist_sq <- sum(normal^2)
      if (is.na(dist_sq) || dist_sq < 1e-12)
        next

      unique_verts <- raw_verts[1, , drop = FALSE]
      if (nrow(raw_verts) > 1) {
        for (k in 2:nrow(raw_verts)) {
          pt <- raw_verts[k, ]
          dists_v <- sqrt(rowSums(t(t(
            unique_verts
          ) - pt)^2))
          if (all(dists_v > 1e-5)) {
            unique_verts <- rbind(unique_verts, pt)
          } else {
            match_idx <- which.min(dists_v)
            unique_verts[match_idx, ] <- (unique_verts[match_idx, ] + pt) / 2
          }
        }
      }
      if (nrow(unique_verts) < 3)
        next

      face_centroid <- colMeans(unique_verts)
      u <- if (abs(normal[1]) < 0.9)
        c(1, 0, 0)
      else
        c(0, 1, 0)
      u <- cross_product(normal, u)
      u <- u / sqrt(sum(u^2))
      v <- cross_product(normal, u)
      v <- v / sqrt(sum(v^2))

      rel <- t(t(unique_verts) - face_centroid)
      angles <- atan2(rel %*% v, rel %*% u)
      face_verts_ordered <- unique_verts[order(angles), , drop = FALSE]

      sa <- calculate_solid_angle(center_coords, face_verts_ordered)

      area <- 0
      n_fv <- nrow(face_verts_ordered)
      for (k in 1:n_fv) {
        p1 <- face_centroid
        p2 <- face_verts_ordered[k, ]
        p3 <- face_verts_ordered[if (k == n_fv)
          1
          else
            k + 1, ]
        cp <- cross_product(p2 - p1, p3 - p1)
        area <- area + 0.5 * sqrt(sum(cp^2))
      }

      if (area < 1e-10 || sa < 1e-10)
        next

      results_list[[length(results_list) + 1]] <- list(
        Atom1 = prim_label,
        Atom2 = expanded_coords$Label[neigh_idx],
        Distance = sqrt(dist_sq),
        SolidAngle = sa,
        Area = area,
        DeltaX = normal[1],
        DeltaY = normal[2],
        DeltaZ = normal[3]
      )
    }
  }

  if (length(results_list) == 0)
    return(NULL)
  raw_dt <- rbindlist(results_list)

  # Reconstruct true mathematical faces by summing the shattered shards belonging to the same atom pair.
  dt <- raw_dt[, .(
    Distance = min(Distance),
    SolidAngle = sum(SolidAngle),
    Area = sum(Area),
    DeltaX = mean(DeltaX),
    DeltaY = mean(DeltaY),
    DeltaZ = mean(DeltaZ)
  ), by = .(Atom1, Atom2)]

  # Apply the physical filter now that the faces are reconstructed
  dt <- dt[Area >= 1e-5 & SolidAngle >= 1e-5]
  if (nrow(dt) == 0)
    return(NULL)

  dt[, MaxSA := max(SolidAngle, na.rm = TRUE), by = Atom1]
  dt[, Weight := ifelse(MaxSA > 0, SolidAngle / MaxSA, 0)]

  return(dt[Weight > tol])
}

#' @title Identify Atomic Bonds using CrystalNN
#' @description Port of Pymatgen's `CrystalNN` weighting logic. Includes Porosity,
#' Electronegativity, and Distance penalties.
#' Uses specific Decision Tree logic for radii (Shannon -> Covalent -> Atomic) per
#' pymatgen's fallback logic.
#' @param distances Ignored.
#' @param atomic_coordinates Primary atom set.
#' @param expanded_coords Expanded supercell.
#' @param unit_cell_metrics Cell parameters.
#' @param cutoff_length Voronoi cutoff. Default 7.0 matches Pymatgen CrystalNN default.
#' @param x_diff_weight Electronegativity weight (default 3.0).
#' @param porosity_adjustment Logical (default TRUE).
#' @param distance_cutoffs Vector c(0.5, 1.0).
#' @return A `data.table` of bonded pairs.
#' @references
#' Shannon, R. D. (1976). "Revised effective ionic radii and systematic studies of
#' interatomic distances in halides and chalcogenides." \emph{Acta Crystallographica
#' Section A}, 32(5), 751-767.
#'
#' Pauling, L. (1960). \emph{The Nature of the Chemical Bond and the Structure of
#' Molecules and Crystals: An Introduction to Modern Structural Chemistry}.
#' Cornell University Press.
#'
#' Pan, H., Ganose, A. M., Horton, M., et al. (2021). "Benchmarking Coordination Number
#' Prediction Algorithms on Inorganic Crystal Structures." \emph{Inorganic Chemistry},
#' 60(3), 1590–1603. \doi{10.1021/acs.inorgchem.0c02996}
#' @family bonding algorithms
#' @export
#' @examples
#' \donttest{
#' cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
#' if (file.exists(cif_path)) {
#'   old_threads <- data.table::setDTthreads(2)
#'   res <- analyze_single_cif(cif_path, bonding_algorithms = "crystal_nn")
#'   print(head(res$bonds_crystal_nn[[1]]))
#'   data.table::setDTthreads(old_threads)
#' }
#' }
crystal_nn <- function(distances,
                       atomic_coordinates,
                       expanded_coords,
                       unit_cell_metrics,
                       cutoff_length = 7.0,
                       # Pymatgen default for CrystalNN is 7.0 (search_cutoff)
                       x_diff_weight = 3.0,
                       porosity_adjustment = TRUE,
                       distance_cutoffs = c(0.5, 1.0)) {
  # 1. Get Base Voronoi Neighbors
  base <- voronoi_nn(
    atomic_coordinates,
    expanded_coords,
    unit_cell_metrics,
    cutoff = cutoff_length,
    tol = 0
  )
  if (is.null(base) || nrow(base) == 0)
    return(NULL)

  dt <- copy(base)
  dt[, `:=`(Sym1 = sub("[0-9].*", "", sub("_.*", "", Atom1)),
            Sym2 = sub("[0-9].*", "", sub("_.*", "", Atom2)))]

  # Merge Oxidation States if available (for radii/EN logic)
  atom_info <- copy(atomic_coordinates)
  if (!"OxidationState" %in% names(atom_info))
    atom_info[, OxidationState := NA_real_]

  dt[, Parent1 := sub("_.*", "", Atom1)]
  dt[, Parent2 := sub("_.*", "", Atom2)]
  dt[atom_info, on = c("Parent1" = "Label"), OS1 := i.OxidationState]
  dt[atom_info, on = c("Parent2" = "Label"), OS2 := i.OxidationState]

  # --- Step 1: Porosity Adjustment ---
  # Pymatgen Logic: x['weight'] *= x['poly_info']['solid_angle'] / x['poly_info']['area']
  # Since x['weight'] IS the SolidAngle (from VoronoiNN),
  # result = SolidAngle * (SolidAngle / Area) = SolidAngle^2 / Area.
  if (porosity_adjustment) {
    dt[, RawScore := ifelse(Area > 1e-8, (SolidAngle * SolidAngle) / Area, SolidAngle)]
  } else {
    dt[, RawScore := SolidAngle]
  }

  # --- Step 2: Chemical Adjustment (Electronegativity) ---
  if (x_diff_weight > 0) {
    dt[, EN1 := get_electronegativity(Sym1)]
    dt[, EN2 := get_electronegativity(Sym2)]

    # 3.3 is max deltaX constant from pymatgen
    dt[, ChemMod := 1 + x_diff_weight * sqrt(abs(EN1 - EN2) / 3.3)]
    dt[is.na(ChemMod), ChemMod := 1.0]

    dt[, RawScore := RawScore * ChemMod]
  }

  # --- Step 3: Normalize (Highest weight = 1.0) ---
  dt[, MaxScore := max(RawScore, na.rm = TRUE), by = Atom1]
  dt[, Weight := ifelse(MaxScore > 0, RawScore / MaxScore, 0)]

  # --- Step 4: Distance Cutoffs (Penalty applied AFTER normalization) ---
  if (!is.null(distance_cutoffs)) {
    # DEFINE atomic_rads IN LOCAL SCOPE FOR Vectorize TO FIND
    atomic_rads <- get_radii_data()

    get_r_vec <- Vectorize(get_radius_decision_tree,
                           vectorize.args = c("symbol", "oxidation_state"))

    # Use Decision Tree Logic: Shannon -> Covalent -> Atomic
    dt[, R1 := get_r_vec(Sym1, OS1, cn = 6)]
    dt[, R2 := get_r_vec(Sym2, OS2, cn = 6)]

    # Pymatgen Logic: If specific radii failed (returned 0 by fallback),
    # fallback to default Covalent/Atomic
    get_default_r <- Vectorize(function(sym) {
      r <- atomic_rads[Symbol == sym & Type == "covalent"]$Radius
      if (length(r) == 0)
        r <- atomic_rads[Symbol == sym & Type == "atomic"]$Radius
      if (length(r) == 0)
        return(0)
      return(r[1])
    }, vectorize.args = "sym")

    dt[R1 == 0, R1 := get_default_r(Sym1)]
    dt[R2 == 0, R2 := get_default_r(Sym2)]

    dt[, Diameter := R1 + R2]
    dt[, CutLow := Diameter + distance_cutoffs[1]]
    dt[, CutHigh := Diameter + distance_cutoffs[2]]

    # Calculate penalty
    dt[, DistMod := data.table::fcase(Distance <= CutLow,
                                      1.0,
                                      Distance >= CutHigh,
                                      0.0,
                                      default = 0.5 * (cos((Distance - CutLow) / (CutHigh - CutLow) * pi
                                      ) + 1))]
    dt[is.na(DistMod), DistMod := 0]

    dt[, Weight := Weight * DistMod]
  }

  # --- Step 5: Pymatgen Rounding Logic ---
  # Pymatgen explicitly rounds weights to 3 decimal places.
  # This acts as a filter for very small weights (e.g., 0.0004 -> 0).
  dt[, Weight := round(Weight, 3)]

  return(dt[Weight > 0, .(Atom1, Atom2, Distance, DeltaX, DeltaY, DeltaZ, Weight)])
}

test_that("Chemical bonding algorithms execute and return correct structure",
          {
            dists <- distances_1590946

            # Minimum Distance
            bonds_md <- minimum_distance(dists, delta = 0.1)
            expect_s3_class(bonds_md, "data.table")
            expect_true(nrow(bonds_md) > 0)
            expect_true("Weight" %in% names(bonds_md))

            # Brunner's Reciprocal
            bonds_br <- brunner_nn_reciprocal(dists)
            expect_s3_class(bonds_br, "data.table")
            expect_true(nrow(bonds_br) > 0)

            # Hoppe's EconNN
            bonds_ec <- econ_nn(dists, atoms_1590946, use_fictive_radius = TRUE)
            expect_s3_class(bonds_ec, "data.table")
            expect_true(nrow(bonds_ec) > 0)
          })

test_that("Geometric bonding algorithms (Voronoi/CrystalNN) execute successfully",
          {
            skip_if_not_installed("geometry")

            # Prepare collapsed coordinate data needed for geometric algorithms
            col_atoms <- crystract:::merge_partial_occupancy_sites(atoms_1590946)
            col_full <- apply_symmetry_operations(col_atoms, sym_ops_1590946, metrics_1590946)

            exp_factors <- calculate_expansion_factors(metrics_1590946, 13.0)
            col_super <- expand_transformed_coords(col_full, exp_factors)

            # Voronoi
            bonds_vo <- voronoi_nn(col_atoms, col_super, metrics_1590946)
            expect_s3_class(bonds_vo, "data.table")
            expect_true("SolidAngle" %in% names(bonds_vo))

            # CrystalNN
            bonds_cr <- crystal_nn(distances_1590946, col_atoms, col_super, metrics_1590946)
            expect_s3_class(bonds_cr, "data.table")
            expect_true("Weight" %in% names(bonds_cr))
          })

test_that("Neighbor counting is correct", {
  counts <- results_1590946$cn_minimum_distance[[1]]
  expect_equal(nrow(counts), 3)
  expect_true("WeightedCoordinationNumber" %in% names(counts))
})

test_that("Bond angle calculations run", {
  angles <- results_1590946$angles_minimum_distance[[1]]
  expect_s3_class(angles, "data.table")
  expect_true("Angle" %in% names(angles))
  expect_true(nrow(angles) > 0)
})

## ----include = FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
# Save current options and apply new ones for the vignette rendering
old_options <- options(width = 75, datatable.print.width = 80)

## ----setup, message=FALSE, warning=FALSE---------------------------------
library(crystract)
library(data.table)

# Limit data.table threads to 2 to comply with CRAN policies on CPU time
old_dt_threads <- data.table::setDTthreads(2)

## ----full-pipeline-with-comments, eval=TRUE------------------------------
# IMPORTANT: Update this path to point to your own downloaded CIF file.
# 1. Try to find the file in the installed package
cif_path <- system.file("extdata", "2405219.cif", package = "crystract")

# 2. If that fails, look in the source directory
if (cif_path == "") {
  cif_path <- "../inst/extdata/2405219.cif"
}

# 3. Final check to provide a clear error if both fail
if (!file.exists(cif_path)) {
  stop("Could not find 2405219.cif in installed package or inst/extdata/")
}

# Run the pipeline on our single example file, extracting multiple bonding types at once
analysis_results <- analyze_single_cif(
  cif_path,
  bonding_algorithms = c("minimum_distance", "brunner", "econ", "voronoi", "crystal_nn")
)

# Let's inspect the structure of the output table.
# It's a single row containing all our results in nested data.tables.
str(analysis_results, max.level = 2)

## ----access-nested-data, eval=FALSE--------------------------------------
# # The result is a list-column, so we access the element with [[1]]
# final_bonds <- analysis_results$bonds_minimum_distance[[1]]
# 
# print(head(final_bonds))

## ----load-data, eval=TRUE------------------------------------------------
# The path was defined in the previous section:
cif_data_list <- read_cif_files(cif_path)

# We'll work with the content of the first file
cif_content <- cif_data_list[[1]]

# Let's look at the first few lines of the raw data
knitr::kable(
  head(cif_content),
  caption = "First 6 lines of the raw CIF data."
)

## ----extract-metadata, eval=TRUE-----------------------------------------
database_code <- extract_database_code(cif_content)
chemical_formula <- extract_chemical_formula(cif_content)
space_group_name <- extract_space_group_name(cif_content)
space_group_number <- extract_space_group_number(cif_content)

cat("Database Code:", database_code, "\n")
cat("Chemical Formula:", chemical_formula, "\n")
cat("Space Group:", space_group_name, "(No.", space_group_number, ")\n")

## ----extract-metrics, eval=TRUE------------------------------------------
unit_cell_metrics <- extract_unit_cell_metrics(cif_content)
print(unit_cell_metrics)

## ----extract-coords-symm, eval=TRUE--------------------------------------
# Extract the coordinates of the unique atoms in the asymmetric unit
atomic_coordinates <- extract_atomic_coordinates(cif_content)
print("Asymmetric Atomic Coordinates:")
print(atomic_coordinates)

# Extract the symmetry operations
symmetry_operations <- extract_symmetry_operations(cif_content)
print("Symmetry Operations (first 6 of 8):")
print(head(symmetry_operations))

## ----generate-structure, eval=TRUE---------------------------------------
# Apply symmetry to generate all atoms in the primary unit cell
transformed_coords <- apply_symmetry_operations(atomic_coordinates, symmetry_operations, unit_cell_metrics)
print("Unique atoms in full unit cell (first 6):")
print(head(transformed_coords))

# Expand into a supercell for neighbor calculations
expanded_coords <- expand_transformed_coords(transformed_coords)
print("Atoms in supercell (first 6):")
print(head(expanded_coords))

## ----calc-distances, eval=TRUE-------------------------------------------
distances <- calculate_distances(atomic_coordinates, expanded_coords, unit_cell_metrics)
print("Calculated Distances (shortest 6):")
print(head(distances[order(Distance)]))

## ----calc-bonding-neighbors, eval=TRUE-----------------------------------
# 1. Minimum Distance
bonds_min <- minimum_distance(distances, delta = 0.1)

# 2. Brunner's Method
bonds_brunner <- brunner_nn_reciprocal(distances)

# 3. Hoppe's ECoN
bonds_econ <- econ_nn(distances, atomic_coordinates)

# 4. Voronoi Tessellation
bonds_voronoi <- voronoi_nn(atomic_coordinates, expanded_coords, unit_cell_metrics)

# 5. CrystalNN
bonds_crystal_nn <- crystal_nn(distances, atomic_coordinates, expanded_coords, unit_cell_metrics)

cat("Minimum Distance found", nrow(bonds_min), "bonds.\n")
cat("CrystalNN found", nrow(bonds_crystal_nn), "bonds.\n")

# Calculate integer neighbor counts based on the bonded pairs (e.g., using CrystalNN)
neighbor_counts <- calculate_neighbor_counts(bonds_crystal_nn)
print("CrystalNN Neighbor Counts:")
print(neighbor_counts)

## ----calc-angles, eval=TRUE----------------------------------------------
bond_angles <- calculate_angles(
  bonds_min,
  atomic_coordinates,
  expanded_coords,
  unit_cell_metrics
)
print("Calculated Bond Angles (first 6):")
print(head(bond_angles))

## ----propagate-errors, eval=TRUE-----------------------------------------
# Propagate errors for interatomic distances
bonded_pairs_with_error <- propagate_distance_error(
  bonds_min,
  atomic_coordinates,
  unit_cell_metrics
)
print("Bonded Pairs with Distance Error (first 6):")
print(head(bonded_pairs_with_error))

# Propagate errors for bond angles
bond_angles_with_error <- propagate_angle_error(
  bond_angles,
  atomic_coordinates,
  expanded_coords,
  unit_cell_metrics
)
print("Bond Angles with Angle Error (first 6):")
print(head(bond_angles_with_error))

## ----eval=FALSE----------------------------------------------------------
# # In an interactive R session, you would run this:
# filtered_bonds <- filter_atoms_by_symbol(
#   data_table = bonded_pairs_with_error,
#   atom_col = "Atom1" # Filter based on the central atom
# )

## ----filter-by-wyckoff, eval=TRUE----------------------------------------
# 1. Let's look at the Wyckoff information already extracted from the CIF.
# All atoms in this structure are on the "a" site with a multiplicity of 4.
print("Atomic coordinates showing Wyckoff information:")
print(atomic_coordinates[, .(Label, WyckoffSymbol, WyckoffMultiplicity)])
cat("\n")

# 2. Filter bonds where the central atom is on the "4d" Wyckoff site.
bonds_from_4d_site <- filter_by_wyckoff_symbol(
  data_table = bonded_pairs_with_error,
  atomic_coordinates = atomic_coordinates,
  atom_col = "Atom1",
  wyckoff_symbols = "4d"
)

print(paste("Number of rows in original bond table:", nrow(bonded_pairs_with_error)))
print(paste("Number of rows after filtering for site '4d':", nrow(bonds_from_4d_site)))

## ----filter-ghost-distances, eval=TRUE-----------------------------------
# A distance d is kept if: (r1+r2)*(1-margin) <= d <= (r1+r2)*(1+margin)
filtered_result <- filter_ghost_distances(
    distances = distances,
    atomic_coordinates = atomic_coordinates,
    margin = 0.1 # Default margin is 10%
)

kept_distances <- filtered_result$kept
removed_distances <- filtered_result$removed

cat("Total distances calculated:", nrow(distances), "\n")
cat("Distances kept after filtering:", nrow(kept_distances), "\n")
cat("Unlikely / non-bonded distances removed:", nrow(removed_distances), "\n\n")

print("Subset of removed distances (showing physically impossible / too long connections):")
print(head(removed_distances))

## ----filter-by-elements, eval=TRUE---------------------------------------
# Let's filter our bond table to exclude any bonds involving Tin ("Sn").
# The resulting table will only contain Cu-Se bonds.
bonds_without_sn <- filter_by_elements(
    distances = bonded_pairs_with_error,
    atomic_coordinates = atomic_coordinates,
    elements_to_exclude = "Ge"
)

cat("Number of bonds in original table:", nrow(bonded_pairs_with_error), "\n")
cat("Number of bonds after excluding 'Ge':", nrow(bonds_without_sn), "\n")

## ----calculate-weighted-average-distance, eval=TRUE----------------------
# Calculate the weighted average BOND distance for the network.
# First, identify the bonds in the structure. We use the `bonds_min` table
# created in section 1.2.6.

# Then, define the Wyckoff sites belonging to the network. 
# For this file, the framework atoms are on the "4d" sites.
network_wyckoff_sites <- "4d"

# Apply the function to the table of identified bonds.
weighted_avg_bond_dist <- calculate_weighted_average_network_distance(
    distances = bonds_min, # Use the bond table as input
    atomic_coordinates = atomic_coordinates,
    wyckoff_symbols = network_wyckoff_sites
)

cat("Weighted average network bond distance for the '4d' sites:", weighted_avg_bond_dist, "Å\n")

## ----custom-radii-and-batch, eval=FALSE----------------------------------
# # --- 1. Define and set the custom radii dictionary ---
# radii_dict <- c(
#   Ac=2.15, Ag=1.45, Al=1.21, Am=1.8, Ar=1.06, As=1.19, At=1.5, Au=1.36, B=0.84,
#   Ba=2.15, Be=0.96, Bi=1.48, Br=1.2, C=0.73, Ca=1.76, Cd=1.44, Ce=2.04, Cl=1.02,
#   Cm=1.69, Co=1.38, Cr=1.39, Cs=2.44, Cu=1.32, Dy=1.92, Er=1.89, Eu=1.98, F=0.57,
#   Fe=1.42, Fr=2.6, Ga=1.22, Gd=1.96, Ge=1.2, H=0.31, He=0.28, Hf=1.75, Hg=1.32,
#   Ho=1.92, I=1.39, In=1.42, Ir=1.41, K=2.03, Kr=1.16, La=2.07, Li=1.28, Lu=1.87,
#   Mg=1.41, Mn=1.5, Mo=1.54, N=0.71, Na=1.66, Nb=1.64, Nd=2.01, Ne=0.58, Ni=1.24,
#   Np=1.9, O=0.66, Os=1.44, P=1.07, Pa=2, Pb=1.46, Pd=1.39, Pm=1.99, Po=1.4, Pr=2.03,
#   Pt=1.36, Pu=1.87, Ra=2.21, Rb=2.2, Re=1.51, Rh=1.42, Rn=1.5, Ru=1.46, S=1.05,
#   Sb=1.39, Sc=1.7, Se=1.2, Si=1.11, Sm=1.98, Sn=1.39, Sr=1.95, Ta=1.7, Tb=1.94,
#   Tc=1.47, Te=1.38, Th=2.06, Ti=1.6, Tl=1.45, Tm=1.9, U=1.96, V=1.53, W=1.62,
#   Xe=1.4, Y=1.9, Yb=1.87, Zn=1.22, Zr=1.75
# )
# 
# custom_radii_table <- data.table(
#   Symbol = names(radii_dict),
#   Radius = as.numeric(radii_dict),
#   Type   = "covalent"
# )
# 
# # Inject the custom table into the current crystract session
# set_radii_data(custom_radii_table)
# 
# # --- 2. Run the batch analysis ---
# cat("Starting batch analysis on CIF files...\n")
# results <- analyze_cif_files(
#   file_paths = "path/to/cif_directory",
#   tolerance = 1e-4,
#   bonding_algorithms = c("minimum_distance", "brunner", "econ", "voronoi", "crystal_nn"),
#   calculate_bond_angles = FALSE,       # Skip angles to speed up extraction
#   perform_error_propagation = FALSE,   # Skip uncertainties
#   workers = 4                          # Adjust based on your available CPU cores
# )
# 
# # --- 3. Extract and Save Coordination Numbers to CSV ---
# # Create a dedicated folder for the outputs to avoid clutter
# output_dir <- "path/to/output_directory"
# if (!dir.exists(output_dir)) dir.create(output_dir)
# 
# # Identify all coordination number columns generated by the pipeline
# cn_columns <- grep("^cn_", names(results), value = TRUE)
# 
# cat("\nExtracting coordination numbers and saving to CSV...\n")
# 
# for (col in cn_columns) {
# 
#   # Extract the list of data.tables for this specific algorithm
#   list_of_dts <- results[[col]]
# 
#   # Associate each table with its CIF file name
#   names(list_of_dts) <- results$file_name
# 
#   # Unnest and bind all tables together into one giant table
#   combined_dt <- rbindlist(list_of_dts, idcol = "file_name", fill = TRUE)
# 
#   if (nrow(combined_dt) > 0) {
#     # Format the file path (e.g., "cn_crystal_nn_crystract.csv")
#     file_path <- file.path(output_dir, paste0(col, "_crystract.csv"))
# 
#     # Save to CSV
#     fwrite(combined_dt, file_path)
# 
#     cat(sprintf("Saved %d rows across %d files to: %s\n",
#                 nrow(combined_dt),
#                 uniqueN(combined_dt$file_name),
#                 basename(file_path)))
#   } else {
#     cat(sprintf("No valid coordination data found for %s, skipping.\n", col))
#   }
# }
# 
# cat("\nAll operations finished successfully! Files are in:", output_dir, "\n")

## ----include = FALSE----------------------------------------------------------
# Restore original options as required by CRAN
options(old_options)
data.table::setDTthreads(old_dt_threads)

#' @title Load Internal Data (Internal)
#' @description Loads data from inst/extdata or package data.
#' @param filename String. Name of file in inst/extdata.
#' @noRd
load_ext_data <- function(filename) {
  path <- system.file("extdata", filename, package = "crystract")
  if (path == "") {
    # Fallback for dev mode
    path <- file.path("inst", "extdata", filename)
  }
  if (!file.exists(path))
    return(NULL)

  # fill=TRUE handles the footer warning
  data.table::fread(path, fill = TRUE)
}

#' @title Get Cached Shannon Radii (Internal)
#' @description Loads, cleans, and caches the Shannon Radii table.
#' @noRd
get_shannon_radii <- function() {
  # 1. return cached version if exists
  if (exists("shannon_radii_cache", envir = .crystract_env)) {
    return(get("shannon_radii_cache", envir = .crystract_env))
  }

  # 2. Load from disk
  dt <- load_ext_data("Shannon_Radii.csv")

  if (is.null(dt))
    return(NULL)

  # 3. Robust Column Renaming
  # Remove spaces/punctuation to handle "Ionic Radius" vs "IonicRadius"
  current_names <- names(dt)
  clean_names <- gsub("[^A-Za-z0-9]", "", current_names)

  # Define mapping based on Clean Names -> Standard Names
  target_map <- c("Element" = "Symbol",
                  "Charge" = "OxidationState",
                  "IonicRadius" = "Radius")

  for (clean_n in names(target_map)) {
    idx <- which(clean_names == clean_n)
    if (length(idx) > 0) {
      setnames(dt, current_names[idx[1]], target_map[[clean_n]])
    }
  }

  # 4. Clean Coordination Column (Roman -> Integer)
  # Many Shannon tables use VI, IV, etc.
  if ("Coordination" %in% names(dt)) {
    # Create a mapping for common Roman numerals found in Shannon data
    roman_map <- c(
      "I" = 1,
      "II" = 2,
      "III" = 3,
      "IV" = 4,
      "V" = 5,
      "VI" = 6,
      "VII" = 7,
      "VIII" = 8,
      "IX" = 9,
      "X" = 10,
      "XI" = 11,
      "XII" = 12,
      "XIII" = 13,
      "XIV" = 14
    )

    # Function to convert single value
    clean_coord <- function(x) {
      # Remove distinct markers like 'sq', 'py' sometimes found in Shannon datasets
      x_clean <- gsub("[^A-Za-z0-9]", "", as.character(x))

      # Check if it's already a number
      if (grepl("^[0-9]+$", x_clean))
        return(as.numeric(x_clean))

      # Check Roman map
      if (x_clean %in% names(roman_map))
        return(roman_map[[x_clean]])

      # Fallback: Try removing letters and parsing (e.g. "4sq" -> 4)
      num_part <- gsub("[^0-9]", "", as.character(x))
      if (nchar(num_part) > 0)
        return(as.numeric(num_part))

      return(NA_real_)
    }

    # Apply conversion
    dt[, Coordination := sapply(Coordination, clean_coord)]
  }

  # 4. Cache and return
  assign("shannon_radii_cache", dt, envir = .crystract_env)
  return(dt)
}

#' @title Calculate Cross Product (Internal)
#' @description Calculates the cross product of two 3D vectors.
#' @param v1 Numeric vector of length 3.
#' @param v2 Numeric vector of length 3.
#' @return Numeric vector of length 3.
#' @noRd
cross_product <- function(v1, v2) {
  c(v1[2] * v2[3] - v1[3] * v2[2], v1[3] * v2[1] - v1[1] * v2[3], v1[1] * v2[2] - v1[2] * v2[1])
}

#' @title Calculate Circumcenter (Internal)
#' @description Calculates the circumcenter of a tetrahedron defined by 4 points.
#'   Includes robust checks for coplanar/degenerate vertices which cause
#'   numerical instability in Voronoi tessellation.
#' @param pts A 4x3 matrix of coordinates.
#' @return Numeric vector of length 3. Returns `rep(NA_real_, 3)` if the
#'   tetrahedron is degenerate (volume ~ 0).
#' @noRd
calculate_circumcenter <- function(pts) {
  # Relative vectors from point 1
  r2 <- pts[2, ] - pts[1, ]
  r3 <- pts[3, ] - pts[1, ]
  r4 <- pts[4, ] - pts[1, ]

  # Construct matrix A (rows are difference vectors)
  A <- rbind(r2, r3, r4)

  # Check for degeneracy using determinant
  # The determinant is 6 * Volume of tetrahedron.
  det_A <- det(A)

  # Tolerance matches pymatgen/scipy qhull sensitivity defaults
  if (abs(det_A) < 1e-10) {
    return(rep(NA_real_, 3))
  }

  # Squared magnitudes
  sq2 <- sum(r2^2)
  sq3 <- sum(r3^2)
  sq4 <- sum(r4^2)

  b <- 0.5 * c(sq2, sq3, sq4)

  # Solve linear system Ax = b with fallback
  x <- tryCatch(
    solve(A, b),
    error = function(e)
      rep(NA_real_, 3)
  )

  if (anyNA(x))
    return(rep(NA_real_, 3))

  return(pts[1, ] + x)
}

#' @title Calculate Solid Angle (Internal)
#' @description Calculates the solid angle subtended by a polygon face from a center point,
#' following the method used in pymatgen (Oosterom & Strackee).
#' @param center Numeric vector of length 3 (center atom coords).
#' @param face_coords Matrix of face vertices (Nx3).
#' @return Numeric solid angle in steradians.
#' @noRd
calculate_solid_angle <- function(center, face_coords) {
  # Calculate displacement vectors from center to vertices
  r <- t(t(face_coords) - center)
  r_norms <- sqrt(rowSums(r^2))

  angle <- 0
  n <- nrow(face_coords)

  if (n < 3)
    return(0)

  # Iterate through the triangle fan (0, i, i+1 logic)
  for (i in 2:(n - 1)) {
    r0 <- r[1, ]
    ri <- r[i, ]
    rj <- r[i + 1, ]

    n0 <- r_norms[1]
    ni <- r_norms[i]
    nj <- r_norms[i + 1]

    # Scalar triple product: r0 . (ri x rj)
    cp <- cross_product(ri, rj)
    tp <- abs(sum(r0 * cp))

    # Denominator (Oosterom & Strackee)
    denom <- n0 * ni * nj +
      nj * sum(r0 * ri) +
      ni * sum(r0 * rj) +
      n0 * sum(ri * rj)

    if (denom == 0) {
      term <- if (tp > 0)
        0.5 * pi
      else-0.5 * pi
    } else {
      term <- atan(tp / denom)
    }

    contrib <- if (term > 0)
      term
    else
      term + pi
    angle <- angle + (contrib * 2)
  }
  return(angle)
}

#' @title Check if Elements are Metals
#' @description Checks a vector of symbols against a standard list of metals.
#' Based on Science Notes list.
#' @param symbols Character vector of element symbols.
#' @return Logical vector.
#' @noRd
is_metal <- function(symbols) {
  metals <- c(
    "Li",
    "Be",
    "Na",
    "Mg",
    "Al",
    "K",
    "Ca",
    "Sc",
    "Ti",
    "V",
    "Cr",
    "Mn",
    "Fe",
    "Co",
    "Ni",
    "Cu",
    "Zn",
    "Ga",
    "Rb",
    "Sr",
    "Y",
    "Zr",
    "Nb",
    "Mo",
    "Tc",
    "Ru",
    "Rh",
    "Pd",
    "Ag",
    "Cd",
    "In",
    "Sn",
    "Cs",
    "Ba",
    "La",
    "Ce",
    "Pr",
    "Nd",
    "Pm",
    "Sm",
    "Eu",
    "Gd",
    "Tb",
    "Dy",
    "Ho",
    "Er",
    "Tm",
    "Yb",
    "Lu",
    "Hf",
    "Ta",
    "W",
    "Re",
    "Os",
    "Ir",
    "Pt",
    "Au",
    "Hg",
    "Tl",
    "Pb",
    "Bi",
    "Po",
    "Fr",
    "Ra",
    "Ac",
    "Th",
    "Pa",
    "U",
    "Np",
    "Pu",
    "Am",
    "Cm",
    "Bk",
    "Cf",
    "Es",
    "Fm",
    "Md",
    "No",
    "Lr",
    "Rf",
    "Db",
    "Sg",
    "Bh",
    "Hs",
    "Mt",
    "Ds",
    "Rg",
    "Cn",
    "Nh",
    "Fl",
    "Mc",
    "Lv",
    "Ts",
    "Og"
  )
  return(symbols %in% metals)
}

#' @title Get Electronegativity
#' @description Looks up Pauling electronegativity. Matches pymatgen `Element.X`.
#' @param symbols Character vector of element symbols.
#' @return Numeric vector of electronegativities.
#' @noRd
get_electronegativity <- function(symbols) {
  vals <- pauling_en$PaulingElectronegativity[match(symbols, pauling_en$Symbol)]
  return(vals)
}

#' @title Get Atomic/Ionic Radius (Pymatgen Style - For EconNN)
#' @description Implements Pymatgen's logic:
#' 1. Ionic (Shannon) if Oxi != 0.
#' 2. Covalent (Cordero) if Oxi == 0 or Ionic not found.
#' 3. Atomic if Covalent not found.
#' Uses cached data to prevent disk I/O on every call.
#' @param symbol Character. Element symbol.
#' @param oxidation_state Numeric. Formal charge.
#' @param cn Numeric. Coordination number.
#' @return Numeric radius in Angstroms.
#' @noRd
get_radius_pymatgen_style <- function(symbol,
                                      oxidation_state = NA,
                                      cn = 6) {
  atomic_rads <- get_radii_data()
  # USE CACHED LOADER
  shannon_rads <- get_shannon_radii()

  # --- CHECK 1: Oxidation State ---
  if (!is.na(oxidation_state) &&
      oxidation_state != 0 && !is.null(shannon_rads)) {
    # Exact match for Symbol + OS + CN
    match <- shannon_rads[Symbol == symbol &
                            OxidationState == oxidation_state &
                            Coordination == cn]

    if (nrow(match) > 0)
      return(match$Radius[1])

    # Relax CN: Find matches for Symbol + OS, pick closest CN
    matches <- shannon_rads[Symbol == symbol &
                              OxidationState == oxidation_state]
    if (nrow(matches) > 0) {
      # The coordination column is now numeric, so subtraction is safe
      matches[, Diff := abs(Coordination - cn)]
      return(matches[order(Diff)]$Radius[1])
    }
  }

  # --- CHECK 2: Covalent (Cordero) ---
  r_atom <- atomic_rads[Symbol == symbol &
                          Type == "covalent"]$Radius
  if (length(r_atom) > 0)
    return(r_atom[1])

  # --- CHECK 3: Atomic ---
  radius_fallback <- atomic_rads[Symbol == symbol &
                                   Type == "atomic"]$Radius
  if (length(radius_fallback) > 0)
    return(radius_fallback[1])

  # Pymatgen _get_radius returns 0 if no match found
  return(0)
}

#' @title Get Atomic/Ionic Radius (Decision Tree - For CrystalNN)
#' @description Implements Julia-Maria Huebner's specific Decision Tree logic,
#' aligned with Pymatgen's _get_radius fallback.
#' Uses cached data to prevent disk I/O on every call.
#' @param symbol Character. Element symbol.
#' @param oxidation_state Numeric. Formal charge.
#' @param cn Numeric. Coordination number.
#' @return Numeric radius in Angstroms.
#' @noRd
get_radius_decision_tree <- function(symbol,
                                     oxidation_state = NA,
                                     cn = 6) {
  atomic_rads <- get_radii_data()
  # USE CACHED LOADER
  shannon_rads <- get_shannon_radii()

  # --- CHECK 1: Oxidation State ---
  if (!is.na(oxidation_state) &&
      oxidation_state != 0 && !is.null(shannon_rads)) {
    match <- shannon_rads[Symbol == symbol &
                            OxidationState == oxidation_state &
                            Coordination == cn]
    if (nrow(match) > 0)
      return(match$Radius[1])

    matches <- shannon_rads[Symbol == symbol &
                              OxidationState == oxidation_state]
    if (nrow(matches) > 0) {
      # The coordination column is now numeric, so subtraction is safe
      matches[, Diff := abs(Coordination - cn)]
      return(matches[order(Diff)]$Radius[1])
    }
  }

  # --- CHECK 2: Covalent (Cordero) ---
  # Pymatgen logic falls back to covalent/atomic if ionic not found or oxi=0
  radius_match <- atomic_rads[Symbol == symbol &
                                Type == "covalent"]$Radius
  if (length(radius_match) > 0)
    return(radius_match[1])

  # --- CHECK 3: Atomic ---
  radius_fallback <- atomic_rads[Symbol == symbol &
                                   Type == "atomic"]$Radius
  if (length(radius_fallback) > 0)
    return(radius_fallback[1])

  return(1.0)
}

#' @title Snap to Common Crystallographic Fractions (Internal)
#' @description Vectorized function that snaps decimal values (like 0.3333) to exact
#'   fractions (like 1/3) to avoid floating point precision issues. Common denominators:
#'   1, 2, 3, 4, 6, 8, 12, 24.
#' @param vals Numeric vector of values to snap.
#' @param tolerance Numeric tolerance.
#' @param wrap Logical. If TRUE, wraps values to[0, 1) and safely binds exactly 1.0 to 0.0.
#' @return Numeric vector of snapped values.
#' @noRd
snap_to_fraction <- function(vals,
                             tolerance = 1e-4,
                             wrap = TRUE) {
  if (length(vals) == 0)
    return(vals)

  vals <- as.numeric(vals)

  if (wrap) {
    vals <- vals %% 1.0
  }

  snapped_vals <- vals
  valid_mask <- !is.na(vals)
  snapped_mask <- rep(FALSE, length(vals))

  denominators <- c(1, 2, 3, 4, 6, 8, 12, 24)

  for (d in denominators) {
    to_check <- valid_mask & !snapped_mask
    if (!any(to_check))
      break

    diffs <- abs(vals[to_check] * d - round(vals[to_check] * d))
    matches <- diffs < tolerance

    if (any(matches)) {
      match_idx <- which(to_check)[matches]
      snapped_vals[match_idx] <- round(vals[match_idx] * d) / d
      snapped_mask[match_idx] <- TRUE
    }
  }

  if (wrap) {
    wrap_mask <- valid_mask &
      (abs(snapped_vals - 1.0) < 1e-7 | abs(snapped_vals) < 1e-7)
    snapped_vals[wrap_mask] <- 0.0
  }

  return(snapped_vals)
}

#' @title Merge Partial Occupancy Sites (Internal)
#' @description Identifies atoms in the asymmetric unit that share the same
#'   fractional coordinates and merges them into a single representative site.
#'   This is required for Voronoi and CrystalNN to function with disordered structures.
#' @param atomic_coordinates A `data.table` of atomic coordinates.
#' @param tolerance Numeric. Snap tolerance for identifying overlapping sites.
#' @return A `data.table` with overlapping atoms combined into single labels.
#' @noRd
merge_partial_occupancy_sites <- function(atomic_coordinates, tolerance = 1e-4) {
  if (is.null(atomic_coordinates) ||
      nrow(atomic_coordinates) == 0)
    return(atomic_coordinates)

  # 1. Create grouping keys based on snapped coordinates
  dt <- copy(atomic_coordinates)
  dt[, `:=`(
    gx = snap_to_fraction(x_a, tolerance = tolerance, wrap = TRUE),
    gy = snap_to_fraction(y_b, tolerance = tolerance, wrap = TRUE),
    gz = snap_to_fraction(z_c, tolerance = tolerance, wrap = TRUE)
  )]

  # 2. Check if there are actually any overlapping sites
  overlap_check <- dt[, .N, by = .(gx, gy, gz)]
  if (all(overlap_check$N == 1))
    return(atomic_coordinates)

  # 3. Merge overlapping sites
  collapsed <- dt[, .(
    Label = paste(Label, collapse = "_"),
    TypeSymbol = if (all(is.na(TypeSymbol)))
      NA_character_
    else
      paste(na.omit(TypeSymbol), collapse = "/"),
    WyckoffSymbol = WyckoffSymbol[1],
    WyckoffMultiplicity = WyckoffMultiplicity[1],
    OxidationState = mean(OxidationState, na.rm = TRUE),
    ThermalParam = mean(ThermalParam, na.rm = TRUE),
    Occupancy = sum(Occupancy),
    OccupancyError = if (all(is.na(OccupancyError)))
      NA_real_
    else
      sqrt(sum(OccupancyError^2, na.rm = TRUE)),
    x_a = x_a[1],
    y_b = y_b[1],
    z_c = z_c[1],
    x_error = x_error[1],
    y_error = y_error[1],
    z_error = z_error[1]
  ), by = .(gx, gy, gz)]

  collapsed[, c("gx", "gy", "gz") := NULL]

  return(collapsed)
}

#' @title Merge Close Atoms (Pymatgen Style)
#' @description Merges atomic sites that are within a specified distance tolerance,
#' accounting for Periodic Boundary Conditions (PBC). This mimics pymatgen's
#' `Structure.merge_sites` logic: it performs hierarchical clustering and then
#' averages the coordinates of the merged sites, taking care of PBC wrapping.
#'
#' @param atomic_coordinates A data.table with columns `Label`, `x_a`, `y_b`, `z_c`.
#' @param unit_cell_metrics A data.table containing cell lengths and angles.
#' @param tol Numeric. Distance tolerance in Angstroms. Sites closer than this will be merged.
#' Default is 1e-4.
#' @return A deduplicated data.table with coordinates averaged.
#' @export
#' @examples
#' ac <- data.table::data.table(Label = c("A", "B"),
#'                              x_a = c(0, 0.00001),
#'                              y_b = c(0, 0),
#'                              z_c = c(0, 0))
#' uc <- data.table::data.table(`_cell_length_a` = 10, `_cell_length_b` = 10,
#'                              `_cell_length_c` = 10, `_cell_angle_alpha` = 90,
#'                              `_cell_angle_beta` = 90, `_cell_angle_gamma` = 90)
#' merge_sites_pbc(ac, uc, tol = 1e-4)
merge_sites_pbc <- function(atomic_coordinates,
                            unit_cell_metrics,
                            tol = 1e-4) {
  if (nrow(atomic_coordinates) < 2)
    return(atomic_coordinates)

  # 1. Extract Cell Parameters
  a <- as.numeric(unit_cell_metrics$`_cell_length_a`)
  b <- as.numeric(unit_cell_metrics$`_cell_length_b`)
  c <- as.numeric(unit_cell_metrics$`_cell_length_c`)
  alpha <- as.numeric(unit_cell_metrics$`_cell_angle_alpha`) * pi / 180
  beta  <- as.numeric(unit_cell_metrics$`_cell_angle_beta`) * pi / 180
  gamma <- as.numeric(unit_cell_metrics$`_cell_angle_gamma`) * pi / 180

  # 2. Metric Tensor components
  g11 <- a^2
  g22 <- b^2
  g33 <- c^2
  g12 <- a * b * cos(gamma)
  g13 <- a * c * cos(beta)
  g23 <- b * c * cos(alpha)

  # 3. Calculate Distance Matrix with PBC manually
  coords <- as.matrix(atomic_coordinates[, .(x_a, y_b, z_c)])
  n <- nrow(coords)
  dist_mat <- matrix(0, nrow = n, ncol = n)

  # Store displacements for use in averaging later
  # dx_mat[i, j] = dx from i to j
  dx_mat <- matrix(0, nrow = n, ncol = n)
  dy_mat <- matrix(0, nrow = n, ncol = n)
  dz_mat <- matrix(0, nrow = n, ncol = n)

  for (i in 1:(n - 1)) {
    for (j in (i + 1):n) {
      dx <- coords[i, 1] - coords[j, 1]
      dy <- coords[i, 2] - coords[j, 2]
      dz <- coords[i, 3] - coords[j, 3]

      # Apply Minimum Image Convention
      rx <- round(dx)
      ry <- round(dy)
      rz <- round(dz)

      dx <- dx - rx
      dy <- dy - ry
      dz <- dz - rz

      d2 <- g11 * dx^2 + g22 * dy^2 + g33 * dz^2 +
        2 * g12 * dx * dy + 2 * g13 * dx * dz + 2 * g23 * dy * dz

      d <- sqrt(d2)
      dist_mat[i, j] <- d
      dist_mat[j, i] <- d

      # Store wrapped displacements
      # If dist is small, this helps us average correctly
      if (d < tol + 1e-6) {
        dx_mat[i, j] <- dx
        dy_mat[i, j] <- dy
        dz_mat[i, j] <- dz
        dx_mat[j, i] <- -dx
        dy_mat[j, i] <- -dy
        dz_mat[j, i] <- -dz
      }
    }
  }

  # 4. Hierarchical Clustering
  dist_obj <- as.dist(dist_mat)
  # Single linkage groups chain-linked neighbors (standard for merging)
  hc <- hclust(dist_obj, method = "single")
  clusters <- cutree(hc, h = tol)

  # 5. Merge Sites with Averaging (matching pymatgen structure.py logic)
  # We iterate through unique clusters. If a cluster has size > 1, we average.

  # Add cluster ID to data
  work_dt <- copy(atomic_coordinates)
  work_dt[, ClusterID := clusters]

  # Identify clusters with duplicates
  dup_clusters <- work_dt[, .N, by = ClusterID][N > 1]$ClusterID

  if (length(dup_clusters) > 0) {
    for (cid in dup_clusters) {
      # Get indices of atoms in this cluster
      indices <- which(clusters == cid)

      # Base coordinate is the first atom
      base_idx <- indices[1]
      avg_x <- coords[base_idx, 1]
      avg_y <- coords[base_idx, 2]
      avg_z <- coords[base_idx, 3]

      # For subsequent atoms, add the Displacement from base to current
      # This effectively unwraps them relative to the base atom
      # Matches pymatgen: coords += ((offset - np.round(offset)) / (site_idx + 2))
      # Here we just sum them up then divide by N

      for (k in 2:length(indices)) {
        curr_idx <- indices[k]
        # dx_mat[base, curr] represents (base - curr)
        # We want curr relative to base, so we subtract (base-curr) from base
        # No, wait.
        # base + (curr - base)_wrapped
        # (curr - base)_wrapped is -dx_mat[base, curr]

        avg_x <- avg_x + (coords[base_idx, 1] - dx_mat[base_idx, curr_idx])
        avg_y <- avg_y + (coords[base_idx, 2] - dy_mat[base_idx, curr_idx])
        avg_z <- avg_z + (coords[base_idx, 3] - dz_mat[base_idx, curr_idx])
      }

      # Final average
      n_clus <- length(indices)
      work_dt[ClusterID == cid, `:=`(
        x_a = avg_x / n_clus,
        y_b = avg_y / n_clus,
        z_c = avg_z / n_clus
      )]
    }
  }

  # Deduplicate: take first entry of each cluster, but with the averaged coordinates
  merged_dt <- unique(work_dt, by = "ClusterID")

  # Wrap back to unit cell[0, 1) and snap strictly
  merged_dt[, `:=`(
    x_a = snap_to_fraction(x_a, wrap = TRUE),
    y_b = snap_to_fraction(y_b, wrap = TRUE),
    z_c = snap_to_fraction(z_c, wrap = TRUE)
  )]
  merged_dt[, ClusterID := NULL]

  return(merged_dt)
}

#' @title Apply Symmetry Operations to Generate a Full Unit Cell
#' @description Generates all symmetry-equivalent atomic positions within the unit cell.
#' @param atomic_coordinates A `data.table` of asymmetric atoms.
#' @param symmetry_operations A `data.table` of operations.
#' @param unit_cell_metrics A `data.table` of cell parameters (needed for distance merging).
#' @param tolerance Numeric. The distance tolerance in Angstroms for merging close atoms.
#' Default is 1e-4.
#' @return A `data.table` containing all unique atomic positions in the unit cell.
#' @family coordinate processors
#' @export
#' @examples
#' ac <- data.table::data.table(Label = "A", x_a = 0, y_b = 0, z_c = 0)
#' ops <- data.table::data.table(x = c("x", "-x"), y = c("y", "-y"), z = c("z", "-z"))
#' uc <- data.table::data.table(`_cell_length_a` = 10, `_cell_length_b` = 10,
#'                              `_cell_length_c` = 10, `_cell_angle_alpha` = 90,
#'                              `_cell_angle_beta` = 90, `_cell_angle_gamma` = 90)
#' apply_symmetry_operations(ac, ops, uc)
apply_symmetry_operations <- function(atomic_coordinates,
                                      symmetry_operations,
                                      unit_cell_metrics,
                                      tolerance = 1e-4) {
  if (is.null(atomic_coordinates) || nrow(atomic_coordinates) == 0) {
    return(
      data.table(
        Label = character(),
        SymOp = integer(),
        x_a = numeric(),
        y_b = numeric(),
        z_c = numeric()
      )
    )
  }

  safe_eval <- function(op_str, x, y, z) {
    expr_str <- gsub("x", sprintf("(%.16f)", x), op_str, fixed = TRUE)
    expr_str <- gsub("y", sprintf("(%.16f)", y), expr_str, fixed = TRUE)
    expr_str <- gsub("z", sprintf("(%.16f)", z), expr_str, fixed = TRUE)
    tryCatch(
      eval(parse(text = expr_str)),
      error = function(e)
        stop(paste("SymOp parse error:", op_str))
    )
  }

  n_atoms <- nrow(atomic_coordinates)
  n_ops <- nrow(symmetry_operations)

  expanded_list <- lapply(seq_len(n_atoms), function(i) {
    row <- atomic_coordinates[i, ]
    lbl <- row$Label
    val_x <- as.numeric(row$x_a)
    val_y <- as.numeric(row$y_b)
    val_z <- as.numeric(row$z_c)

    new_coords <- lapply(seq_len(n_ops), function(j) {
      op_x <- symmetry_operations$x[j]
      op_y <- symmetry_operations$y[j]
      op_z <- symmetry_operations$z[j]

      data.table(
        Label = paste(lbl, j, sep = "_"),
        SymOp = j,
        x_a = snap_to_fraction(safe_eval(op_x, val_x, val_y, val_z), wrap = TRUE),
        y_b = snap_to_fraction(safe_eval(op_y, val_x, val_y, val_z), wrap = TRUE),
        z_c = snap_to_fraction(safe_eval(op_z, val_x, val_y, val_z), wrap = TRUE)
      )
    })
    rbindlist(new_coords)
  })

  transformed_coords <- rbindlist(expanded_list)
  if (nrow(transformed_coords) == 0)
    return(transformed_coords)

  # --- DEDUPLICATION LOGIC ---
  # Pymatgen-style: Clean up using distance tolerance and averaging
  transformed_coords <- merge_sites_pbc(transformed_coords, unit_cell_metrics, tol = tolerance)

  return(transformed_coords)
}

#' @title Calculate Supercell Expansion Factors
#' @description Computes the number of unit cell repetitions required in each
#'   direction (a, b, c) to encompass a sphere of a given radius.
#' @param unit_cell_metrics A `data.table` with cell parameters.
#' @param radius Numeric. The cutoff radius (e.g., 13.0 for Voronoi).
#' @return A named numeric vector `c(a=..., b=..., c=...)`.
#' @family coordinate processors
#' @export
#' @examples
#' uc <- data.table::data.table(`_cell_length_a` = 10, `_cell_length_b` = 10,
#'                              `_cell_length_c` = 10, `_cell_angle_alpha` = 90,
#'                              `_cell_angle_beta` = 90, `_cell_angle_gamma` = 90)
#' calculate_expansion_factors(uc, radius = 13.0)
calculate_expansion_factors <- function(unit_cell_metrics, radius) {
  a <- as.numeric(unit_cell_metrics$`_cell_length_a`)
  b <- as.numeric(unit_cell_metrics$`_cell_length_b`)
  c <- as.numeric(unit_cell_metrics$`_cell_length_c`)
  alpha <- as.numeric(unit_cell_metrics$`_cell_angle_alpha`) * pi / 180
  beta <- as.numeric(unit_cell_metrics$`_cell_angle_beta`) * pi / 180
  gamma <- as.numeric(unit_cell_metrics$`_cell_angle_gamma`) * pi / 180

  # Volume calculation
  cos_a <- cos(alpha)
  cos_b <- cos(beta)
  cos_g <- cos(gamma)

  term <- 1 - cos_a^2 - cos_b^2 - cos_g^2 + 2 * cos_a * cos_b * cos_g
  V <- a * b * c * sqrt(max(0, term))

  # Perpendicular widths (height of the parallelepiped along each vector)
  h_a <- V / (b * c * sin(alpha))
  h_b <- V / (a * c * sin(beta))
  h_c <- V / (a * b * sin(gamma))

  # We need radius to be covered by n * height.
  # Note: Pymatgen adds a small buffer (0.15). We match that.
  r_eff <- radius + 0.15

  n_a <- ceiling(r_eff / h_a)
  n_b <- ceiling(r_eff / h_b)
  n_c <- ceiling(r_eff / h_c)

  # Ensure at least 1 (though 3x3x3 means index -1 to 1, so minimum factor is 1)
  # If factors are c(1,1,1), ranges are -1:1.
  return(c(
    a = max(1, n_a),
    b = max(1, n_b),
    c = max(1, n_c)
  ))
}

#' @title Expand Coordinates into a Supercell
#' @description Takes a set of atomic coordinates within a single unit cell and
#'   replicates them into a supercell grid defined by expansion factors.
#' @param transformed_coords A `data.table` of atom positions within one unit cell.
#' @param expansion_factors A numeric vector `c(a, b, c)` specifying the number of
#'   images in each direction (+/-). Default `c(1,1,1)` creates a 3x3x3 grid.
#' @return A `data.table` containing all atomic positions in the expanded supercell.
#' @family coordinate processors
#' @export
#' @examples
#' tc <- data.table::data.table(Label = "A", x_a = 0, y_b = 0, z_c = 0)
#' expand_transformed_coords(tc, expansion_factors = c(1, 1, 1))
expand_transformed_coords <- function(transformed_coords,
                                      expansion_factors = c(1, 1, 1)) {
  if (is.null(transformed_coords) ||
      nrow(transformed_coords) == 0 ||
      !"x_a" %in% names(transformed_coords)) {
    return(
      data.table(
        Label = character(),
        SymOp = integer(),
        Tx = integer(),
        Ty = integer(),
        Tz = integer(),
        x_a = numeric(),
        y_b = numeric(),
        z_c = numeric()
      )
    )
  }

  na <- expansion_factors[1]
  nb <- expansion_factors[2]
  nc <- expansion_factors[3]

  # Generate indices from -n to +n
  cell_indices <- as.data.table(expand.grid(
    x = -na:na,
    y = -nb:nb,
    z = -nc:nc
  ))

  expanded_coords <- rbindlist(lapply(1:nrow(cell_indices), function(i) {
    cell_shift <- cell_indices[i, ]
    dt <- copy(transformed_coords)
    dt[, `:=`(Tx = cell_shift$x,
              Ty = cell_shift$y,
              Tz = cell_shift$z)]
    # Make labels unique for the supercell images
    dt[, Label := paste(Label, Tx, Ty, Tz, sep = "_")]

    # Apply snapping without cell wrap to ensure clean negative / greater than 1 floats
    dt[, `:=`(
      x_a = snap_to_fraction(x_a + Tx, wrap = FALSE),
      y_b = snap_to_fraction(y_b + Ty, wrap = FALSE),
      z_c = snap_to_fraction(z_c + Tz, wrap = FALSE)
    )]

    return(dt)
  }))

  # Simple unique check
  expanded_coords <- unique(expanded_coords, by = c("x_a", "y_b", "z_c"))
  return(expanded_coords)
}

test_that("apply_symmetry_operations generates correct full cell for 1590946",
          {
            # Ground truth is 12 atoms in the full unit cell.
            # The space group is P n m a (general multiplicity 8), but the atoms in this
            # structure are on the 4c Wyckoff site (y=0.25). So 3 atoms * 4 = 12 atoms.
            expect_equal(nrow(full_cell_1590946), 12)

            # Check a specific transformed coordinate against the ground truth data.
            # The atom 'Sr1' (0.6529, 0.25, 0.0769) transformed by the 1st sym op ('x+1/2, y, -z+1/2')
            # becomes (1.1529, 0.25, 0.4231) -> wrapped to (0.1529, 0.25, 0.4231)
            sr1_transformed <- full_cell_1590946[abs(x_a - 0.1529) < 1e-4 &
                                                   abs(z_c - 0.4231) < 1e-4]
            expect_true(nrow(sr1_transformed) >= 1)
          })

test_that("expand_transformed_coords generates correct supercell for 1590946",
          {
            # Ground truth is 12 atoms/cell * (3x3x3 cells = 27) = 324 atoms
            expect_equal(nrow(super_cell_1590946), 324)
          })

test_that(
  "calculate_expansion_factors correctly determines padding for geometric algorithms",
  {
    # For a 13.0 Angstrom Voronoi cutoff on an 8.11x5.15x9.54 unit cell,
    # expansion factors should safely cover it (e.g., cell b is 5.15, 13 / 5.15 > 2.5 => ceiling + buffer => ~3)
    factors <- calculate_expansion_factors(metrics_1590946, radius = 13.0)
    expect_length(factors, 3)
    expect_true(all(factors >= 1))
  }
)

#' Internal, Safe Worker for Parallel CIF Analysis
#'
#' @description A robust wrapper for `analyze_single_cif` that includes error
#'   handling. Designed to be called by `future.apply` without capturing a
#'   large parent environment.
#'
#' @param cif_content A data.table of CIF content.
#' @param file_name The name of the file.
#' @param ... Additional arguments passed directly to `analyze_single_cif`.
#' @return A one-row data.table on success, or NULL on failure.
#' @noRd
.analyze_single_cif_safe <- function(cif_content, file_name, ...) {
  tryCatch({
    analyze_single_cif(cif_content = cif_content, file_name = file_name, ...)
  }, error = function(e) {
    warning(
      sprintf(
        "A critical error occurred while processing '%s': %s",
        file_name,
        e$message
      )
    )
    return(NULL)
  })
}

#' @title Read CIF Files into Memory
#' @description Reads one or more CIF files from disk and loads each into a
#'   `data.table`. It includes encoding sanitization to handle legacy characters.
#' @param file_paths A character vector of paths to the CIF files.
#' @return A list of `data.table` objects.
#' @export
#' @examples
#' cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
#' if (file.exists(cif_path)) {
#'   cifs <- read_cif_files(cif_path)
#' }
read_cif_files <- function(file_paths) {
  # Set the names of the list to the base filenames
  cif_list <- lapply(file_paths, function(fp) {
    # Read the file
    dt <- data.table::fread(
      fp,
      sep = "\n",
      header = FALSE,
      strip.white = FALSE,
      encoding = "UTF-8" # Attempt UTF-8 first
    )

    # Sanitize: Convert strings to valid UTF-8, replacing invalid bytes with empty strings
    if (nrow(dt) > 0) {
      dt[, V1 := iconv(V1, to = "UTF-8", sub = "")]
    }

    return(dt)
  })

  names(cif_list) <- basename(file_paths)
  return(cif_list)
}

#' @title Analyze the Content of a Single CIF File
#' @description The core worker function that orchestrates the analysis pipeline
#'   for a single crystal structure's data. It dynamically adjusts supercell
#'   size and coordinate sets based on the requested bonding algorithms.
#'   For geometric algorithms (Voronoi/CrystalNN), it automatically merges atoms
#'   occupying the same site to ensure mathematical validity.
#' @param cif_content Either a `data.table` containing the lines of a CIF file,
#'   OR a character string specifying the file path to a CIF file.
#' @param file_name The name of the original CIF file, used for labeling output.
#' @param perform_extraction Logical. If `TRUE`, extracts all metadata.
#' @param perform_calcs_and_transforms Logical. If `TRUE`, generates unit cell
#'   and supercell.
#' @param bonding_algorithms Character vector. Options: `"minimum_distance"`,
#'   `"brunner"`, `"econ"`, `"crystal_nn"`, `"voronoi"`.
#' @param calculate_bond_angles Logical.
#' @param perform_error_propagation Logical.
#' @param tolerance Numeric. Cutoff for floating-point noise and atom merging (default 1e-4).
#' @param minimum_distance_delta Numeric. Tolerance for min-dist algorithm.
#' @return A one-row `data.table` with results.
#' @export
#' @examples
#' cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
#' if (file.exists(cif_path)) {
#'   old_threads <- data.table::setDTthreads(2)
#'   res <- analyze_single_cif(cif_path, perform_error_propagation = FALSE)
#'   data.table::setDTthreads(old_threads)
#' }
analyze_single_cif <- function(cif_content,
                               file_name = NULL,
                               perform_extraction = TRUE,
                               perform_calcs_and_transforms = TRUE,
                               bonding_algorithms = c("minimum_distance"),
                               calculate_bond_angles = TRUE,
                               perform_error_propagation = TRUE,
                               tolerance = 1e-4,
                               minimum_distance_delta = 0.1) {
  # --- Step 0: Handle Input Type ---
  if (is.character(cif_content)) {
    if (length(cif_content) != 1)
      stop("`analyze_single_cif` accepts exactly one string.")
    if (!file.exists(cif_content))
      stop(paste("File not found:", cif_content))
    if (is.null(file_name))
      file_name <- basename(cif_content)
    cif_content <- read_cif_files(cif_content)[[1]]
  } else if (inherits(cif_content, "data.table")) {
    if (is.null(file_name))
      file_name <- "unknown"
  } else {
    stop("`cif_content` must be a file path or loaded CIF data.")
  }

  # --- Step 1: Data Extraction ---
  if (!perform_extraction)
    return(data.table::data.table(file_name = file_name))

  database_code <- extract_database_code(cif_content)
  chemical_formula <- extract_chemical_formula(cif_content)
  structure_type <- extract_structure_type(cif_content)
  space_group_name <- extract_space_group_name(cif_content)
  space_group_number <- extract_space_group_number(cif_content)
  unit_cell_metrics <- extract_unit_cell_metrics(cif_content)
  atomic_coordinates <- extract_atomic_coordinates(cif_content, chemical_formula)
  symmetry_operations <- extract_symmetry_operations(cif_content)

  # --- Step 2: Validate Essential Data ---
  if (is.null(atomic_coordinates) ||
      is.null(symmetry_operations) ||
      is.null(unit_cell_metrics)) {
    warning(paste(
      "Skipping '",
      file_name,
      "': missing essential structural data.",
      sep = ""
    ))
    return(NULL)
  }

  # --- Step 3: Initialize Result Variables ---
  transformed_coords <- NULL
  expanded_coords <- NULL
  distances <- NULL
  algo_results <- list()

  # --- Step 4: Calculations, Transformations, and Expansions ---
  if (perform_calcs_and_transforms) {
    # 4a. Standard Transformation (Keeps partially occupied atoms separate)
    transformed_coords <- apply_symmetry_operations(atomic_coordinates,
                                                    symmetry_operations,
                                                    unit_cell_metrics,
                                                    tolerance = tolerance)

    # 4b. Identify needs for Geometric (Collapsed) Algorithms
    needs_voronoi <- any(c("voronoi", "crystal_nn") %in% bonding_algorithms)

    if (needs_voronoi) {
      # Merge partial occupancy sites in the asymmetric unit for tessellation algorithms
      collapsed_atomic_coords <- merge_partial_occupancy_sites(atomic_coordinates, tolerance)
      collapsed_transformed_coords <- apply_symmetry_operations(collapsed_atomic_coords,
                                                                symmetry_operations,
                                                                unit_cell_metrics,
                                                                tolerance = tolerance)
      # Geometric algorithms need a larger coverage (13.0 Angstroms)
      voronoi_factors <- calculate_expansion_factors(unit_cell_metrics, 13.0)
      collapsed_expanded_coords <- expand_transformed_coords(collapsed_transformed_coords, voronoi_factors)
    }

    # 4c. Standard Expansion (Used for MinDist, Brunner, EconNN)
    # Default 3x3x3 image grid (-1 to 1)
    expansion_factors <- c(1, 1, 1)
    expanded_coords <- expand_transformed_coords(transformed_coords, expansion_factors)

    # Distance calculation for algorithms using standard coordinates
    distances <- calculate_distances(atomic_coordinates,
                                     expanded_coords,
                                     unit_cell_metrics,
                                     tolerance = tolerance)
  }

  # --- Step 5: Bonding Algorithms & Step 6: Angles & Errors ---
  if (!is.null(distances) &&
      length(bonding_algorithms) > 0 &&
      !"none" %in% bonding_algorithms) {
    for (algo in unique(bonding_algorithms)) {
      # Determine which coordinate set and algorithm logic to apply
      if (algo %in% c("voronoi", "crystal_nn")) {
        # GEOMETRIC ALGORITHMS: Use Collapsed Data
        current_bonds <- switch(
          algo,
          "voronoi" = voronoi_nn(
            collapsed_atomic_coords,
            collapsed_expanded_coords,
            unit_cell_metrics
          ),
          "crystal_nn" = crystal_nn(
            distances,
            # Note: crystal_nn uses distances for EN math, but primary coords for tessellation
            collapsed_atomic_coords,
            collapsed_expanded_coords,
            unit_cell_metrics
          )
        )
        ref_coords <- collapsed_atomic_coords
        ref_expanded <- collapsed_expanded_coords
      } else {
        # CHEMICAL ALGORITHMS: Use Standard Separated Data
        current_bonds <- switch(
          algo,
          "minimum_distance" = minimum_distance(distances, delta = minimum_distance_delta),
          "brunner" = brunner_nn_reciprocal(distances),
          "econ" = econ_nn(distances, atomic_coordinates),
          NULL
        )
        ref_coords <- atomic_coordinates
        ref_expanded <- expanded_coords
      }

      if (!is.null(current_bonds) && nrow(current_bonds) > 0) {
        out_algo_name <- algo

        if (perform_error_propagation) {
          current_bonds <- propagate_distance_error(current_bonds, ref_coords, unit_cell_metrics)
        }

        # Coordination counts use the respective coordinate table (standard vs collapsed)
        cn_table <- calculate_weighted_neighbor_counts(current_bonds, ref_coords)

        if (is.null(cn_table)) {
          warning(
            paste(
              "Discarding '",
              file_name,
              "': fractional occupancies sum to > 1.0.",
              sep = ""
            )
          )
          return(NULL)
        }

        algo_results[[paste0("bonds_", out_algo_name)]] <- list(current_bonds)
        algo_results[[paste0("cn_", out_algo_name)]] <- list(cn_table)

        if (calculate_bond_angles) {
          current_angles <- calculate_angles(current_bonds,
                                             ref_coords,
                                             ref_expanded,
                                             unit_cell_metrics)

          if (perform_error_propagation &&
              !is.null(current_angles) &&
              nrow(current_angles) > 0) {
            current_angles <- propagate_angle_error(current_angles,
                                                    ref_coords,
                                                    ref_expanded,
                                                    unit_cell_metrics)
          }
          algo_results[[paste0("angles_", out_algo_name)]] <- list(current_angles)
        }
      }
    }
  }

  # --- Step 7: Construct Return Data Table ---
  final_dt <- data.table::data.table(
    file_name = file_name,
    database_code = database_code,
    chemical_formula = chemical_formula,
    structure_type = structure_type,
    space_group_name = space_group_name,
    space_group_number = space_group_number,
    unit_cell_metrics = list(unit_cell_metrics),
    atomic_coordinates = list(atomic_coordinates),
    symmetry_operations = list(symmetry_operations),
    transformed_coords = list(transformed_coords),
    expanded_coords = list(expanded_coords),
    distances = list(distances)
  )

  if (length(algo_results) > 0) {
    for (res_name in names(algo_results)) {
      set(final_dt, j = res_name, value = algo_results[[res_name]])
    }
  }

  return(final_dt)
}

#' @title Analyze a Batch of CIF Files
#' @description A high-level wrapper that analyzes CIF files in batch, supporting
#'   parallel processing and batch-to-disk operations.
#' @param file_paths Character vector of paths or list of data.tables.
#' @param workers Integer. Number of parallel workers.
#' @param output_dir Path to output directory (optional).
#' @param batch_size Integer.
#' @param ... Args passed to `analyze_single_cif`.
#' @return Data.table or output directory path.
#' @export
#' @examples
#' \donttest{
#' cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
#' if (file.exists(cif_path)) {
#'   old_threads <- data.table::setDTthreads(2)
#'   out_dir <- file.path(tempdir(), "cif_analysis")
#'   res <- analyze_cif_files(cif_path, output_dir = out_dir, workers = 1)
#'
#'   # Clean up
#'   unlink(out_dir, recursive = TRUE)
#'   data.table::setDTthreads(old_threads)
#' }
#' }
analyze_cif_files <- function(file_paths,
                              workers = 1,
                              output_dir = NULL,
                              batch_size = 1000,
                              ...) {
  if (is.character(file_paths)) {
    if (length(file_paths) == 1 && dir.exists(file_paths)) {
      message(sprintf(
        "Input is a directory. Searching for .cif files in '%s'...",
        file_paths
      ))
      found_files <- list.files(
        file_paths,
        pattern = "\\.cif$",
        full.names = TRUE,
        ignore.case = TRUE
      )
      if (length(found_files) == 0) {
        warning(sprintf("No .cif files found in directory '%s'.", file_paths))
        return(data.table::data.table())
      }
      file_paths <- found_files
    }
    cif_contents_list <- read_cif_files(file_paths)
  } else if (is.list(file_paths) &&
             (length(file_paths) == 0 ||
              inherits(file_paths[[1]], "data.table"))) {
    cif_contents_list <- file_paths
    if (is.null(names(cif_contents_list)))
      names(cif_contents_list) <- paste0("unnamed_", seq_along(cif_contents_list))
  } else {
    stop(
      "`file_paths` must be a character vector of paths, a directory path, or a list of data.tables."
    )
  }

  if (workers > 1) {
    if (!requireNamespace("future", quietly = TRUE) ||
        !requireNamespace("future.apply", quietly = TRUE)) {
      stop(
        "Packages 'future' and 'future.apply' are required for parallel processing.",
        call. = FALSE
      )
    }
    old_plan <- future::plan()
    on.exit(future::plan(old_plan), add = TRUE)
    message(paste0("Setting parallel plan to use ", workers, " workers."))
    future::plan(future::multisession, workers = workers)
  }

  total_files <- length(cif_contents_list)
  if (total_files == 0) {
    message("No files to process.")
    return(data.table::data.table())
  }

  batch_starts <- seq(1, total_files, by = batch_size)
  num_batches <- length(batch_starts)
  all_results_list <- if (is.null(output_dir))
    list()
  else
    NULL
  message(sprintf(
    "Starting analysis of %d files in %d batches.",
    total_files,
    num_batches
  ))
  more_args <- list(...)

  for (i in seq_along(batch_starts)) {
    start_index <- batch_starts[i]
    end_index <- min(start_index + batch_size - 1, total_files)
    cif_batch <- cif_contents_list[start_index:end_index]
    message(
      sprintf(
        "\n--- Processing Batch %d of %d (Files %d to %d) ---",
        i,
        num_batches,
        start_index,
        end_index
      )
    )

    results_list <- if (workers > 1) {
      future.apply::future_mapply(
        FUN = .analyze_single_cif_safe,
        cif_content = cif_batch,
        file_name = names(cif_batch),
        MoreArgs = more_args,
        SIMPLIFY = FALSE,
        future.seed = TRUE
      )
    } else {
      mapply(
        FUN = .analyze_single_cif_safe,
        cif_content = cif_batch,
        file_name = names(cif_batch),
        MoreArgs = more_args,
        SIMPLIFY = FALSE
      )
    }

    successful_results <- results_list[!sapply(results_list, is.null)]
    if (length(successful_results) > 0) {
      batch_dt <- data.table::rbindlist(successful_results, fill = TRUE)
      if (is.null(output_dir)) {
        all_results_list[[i]] <- batch_dt
      } else {
        if (!dir.exists(output_dir))
          dir.create(output_dir, recursive = TRUE)
        batch_filename <- file.path(output_dir, paste0("batch_", i, ".rds"))
        saveRDS(batch_dt, batch_filename)
        message(
          sprintf(
            "Batch %d complete. Saved %d results to '%s'.",
            i,
            nrow(batch_dt),
            batch_filename
          )
        )
      }
    } else {
      message(sprintf(
        "Batch %d complete. No structures processed successfully.\n",
        i
      ))
    }
    rm(cif_batch, results_list, successful_results)
    gc()
  }

  message("----------------------------------\nAnalysis Complete!")
  if (is.null(output_dir)) {
    if (length(all_results_list) > 0) {
      final_dt <- data.table::rbindlist(all_results_list, fill = TRUE)
      message(sprintf(
        "Successfully processed a total of %d structures.",
        nrow(final_dt)
      ))
      return(final_dt)
    } else {
      warning("No structures were processed successfully.")
      return(data.table::data.table())
    }
  } else {
    message(sprintf(
      "Batch results have been saved in the '%s/' directory.",
      normalizePath(output_dir)
    ))
    return(invisible(normalizePath(output_dir)))
  }
}

#' @title Aggregate Batched Analysis Results
#' @description Reads all `batch_*.rds` files from a directory and combines them.
#' @param input_dir The directory containing RDS files from `analyze_cif_files`.
#' @param cols_to_keep Optional character vector of column names to keep.
#' @return A single `data.table` containing the aggregated results.
#' @export
#' @examples
#' \donttest{
#' cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
#' if (file.exists(cif_path)) {
#'   old_threads <- data.table::setDTthreads(2)
#'   out_dir <- file.path(tempdir(), "cif_batches")
#'   analyze_cif_files(cif_path, output_dir = out_dir)
#'
#'   aggr <- aggregate_batch_results(out_dir)
#'
#'   unlink(out_dir, recursive = TRUE)
#'   data.table::setDTthreads(old_threads)
#' }
#' }
aggregate_batch_results <- function(input_dir, cols_to_keep = NULL) {
  batch_files <- list.files(path = input_dir,
                            pattern = "^batch_\\d+\\.rds$",
                            full.names = TRUE)
  if (length(batch_files) == 0) {
    stop("No batch result files (e.g., 'batch_1.rds') found in the directory.")
  }

  message(sprintf("Found %d batch files to aggregate.", length(batch_files)))

  all_results_list <- lapply(batch_files, function(file) {
    batch_dt <- readRDS(file)
    if (!is.null(cols_to_keep)) {
      valid_cols <- intersect(cols_to_keep, names(batch_dt))
      if (length(valid_cols) > 0) {
        return(batch_dt[, ..valid_cols])
      }
    }
    return(batch_dt)
  })

  aggregated_results <- data.table::rbindlist(all_results_list, fill = TRUE)
  message(sprintf(
    "Aggregation complete. Total rows: %d.",
    nrow(aggregated_results)
  ))
  return(aggregated_results)
}

#' @title Export Analysis Results to a Directory of CSVs
#' @description Exports analysis results to CSV structure.
#' @param analysis_results A `data.table` object.
#' @param output_dir Path to main output directory.
#' @param overwrite Logical.
#' @return Invisibly returns output_dir.
#' @export
#' @examples
#' \donttest{
#' cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
#' if (file.exists(cif_path)) {
#'   old_threads <- data.table::setDTthreads(2)
#'   res <- analyze_single_cif(cif_path)
#'
#'   out_dir <- file.path(tempdir(), "cif_csvs")
#'   export_analysis_to_csv(res, output_dir = out_dir)
#'
#'   unlink(out_dir, recursive = TRUE)
#'   data.table::setDTthreads(old_threads)
#' }
#' }
export_analysis_to_csv <- function(analysis_results, output_dir, overwrite = FALSE) {
  if (!inherits(analysis_results, "data.table") ||
      nrow(analysis_results) == 0) {
    stop("`analysis_results` must be a non-empty data.table.")
  }
  if (!"file_name" %in% names(analysis_results)) {
    stop("`analysis_results` must contain a 'file_name' column.")
  }

  if (dir.exists(output_dir)) {
    if (overwrite) {
      unlink(output_dir, recursive = TRUE, force = TRUE)
      dir.create(output_dir, recursive = TRUE)
    } else {
      stop(paste(
        "Output directory '",
        output_dir,
        "' already exists. Use overwrite = TRUE."
      ))
    }
  } else {
    dir.create(output_dir, recursive = TRUE)
  }

  is_nested_col <- sapply(analysis_results, is.list)
  nested_col_names <- names(is_nested_col)[is_nested_col]
  meta_col_names <- names(is_nested_col)[!is_nested_col]

  meta_dir <- file.path(output_dir, "meta")
  dir.create(meta_dir)
  meta_data <- analysis_results[, ..meta_col_names]
  data.table::fwrite(meta_data, file.path(meta_dir, "meta_summary.csv"))

  for (col_name in nested_col_names) {
    if (all(sapply(analysis_results[[col_name]], is.null)))
      next
    nested_dir <- file.path(output_dir, col_name)
    dir.create(nested_dir)
    for (i in 1:nrow(analysis_results)) {
      nested_table <- analysis_results[[col_name]][[i]]
      if (!is.null(nested_table) &&
          inherits(nested_table, "data.table") &&
          nrow(nested_table) > 0) {
        cif_name <- analysis_results$file_name[i]
        csv_name <- sub("(?i)\\.cif$", ".csv", cif_name, perl = TRUE)
        output_path <- file.path(nested_dir, csv_name)
        data.table::fwrite(nested_table, output_path)
      }
    }
  }
  message(paste(
    "Analysis successfully exported to:",
    normalizePath(output_dir)
  ))
  invisible(output_dir)
}

test_that(
  "analyze_cif_files produces a valid, self-consistent result with multiple algorithms",
  {
    expect_s3_class(results_1590946, "data.table")
    expect_equal(nrow(results_1590946), 1)

    # Check that requested algorithm columns were generated
    expect_true("bonds_minimum_distance" %in% names(results_1590946))
    expect_true("bonds_crystal_nn" %in% names(results_1590946))
    expect_true("bonds_voronoi" %in% names(results_1590946))
    expect_true("bonds_econ" %in% names(results_1590946))

    # Validate structure of a specific result table
    bonded_pairs_table <- results_1590946$bonds_minimum_distance[[1]]
    expect_s3_class(bonded_pairs_table, "data.frame")
    expect_true("DistanceError" %in% names(bonded_pairs_table))
  }
)

test_that("analyze_cif_files handles multiple files and batching", {
  temp_dir <- tempfile("test-multi-")
  dir.create(temp_dir)
  on.exit(unlink(temp_dir, recursive = TRUE))

  # Copy file twice to simulate batch
  file.copy(cif_path_1590946, file.path(temp_dir, "file1.cif"))
  file.copy(cif_path_1590946, file.path(temp_dir, "file2.cif"))

  multi_results <- analyze_cif_files(list.files(temp_dir, full.names = TRUE))
  expect_equal(nrow(multi_results), 2)
})

test_that("analyze_cif_files handles errors gracefully", {
  bad_cif_path <- tempfile(fileext = ".cif")
  on.exit(unlink(bad_cif_path), add = TRUE)
  writeLines("this is not a valid cif file", bad_cif_path)

  # Expect a warning that processing failed.
  # The function should return an EMPTY data.table if no files succeed.
  expect_warning(results <- analyze_cif_files(bad_cif_path))
  expect_s3_class(results, "data.table")
  expect_equal(nrow(results), 0)
})

#' @title Set or Reset a Custom Atomic Radii Table
#' @description Allows the user to provide their own table of atomic radii for
#' the current R session. This custom table will be used by functions like
#' `filter_ghost_distances`.
#'
#' @details
#' The provided `data.table` or `data.frame` must contain at least three columns:
#' \itemize{
#'   \item `Symbol` (character): The chemical symbol of the element (e.g., "Si").
#'   \item `Radius` (numeric): The atomic radius in Angstroms (Å).
#'   \item `Type` (character): A descriptor for the radius type (e.g., "covalent", "ionic").
#' }
#' This allows for storing multiple types of radii in the same table, which can
#' be selected using the `radii_type` argument in relevant functions.
#'
#' To revert to using the package's default radii table, call the function
#' with `NULL` or without any arguments.
#'
#' @param radii_data A `data.table` or `data.frame` containing the custom radii
#' data, or `NULL` to reset to the package default.
#' @return Invisibly returns `NULL`. A message is printed to the console
#' confirming the action.
#' @export
#' @family post-processing
#' @examples
#' # 1. Create a custom radii table with both covalent and ionic radii
#' my_radii <- data.table::data.table(
#'   Symbol = c("Si", "O", "O"),
#'   Radius = c(1.11, 0.66, 1.40),
#'   Type   = c("covalent", "covalent", "ionic")
#' )
#'
#' # 2. Set this table for the current session
#' set_radii_data(my_radii)
#'
#' # 3. Reset to the package's default covalent radii table
#' set_radii_data(NULL)
#'
set_radii_data <- function(radii_data = NULL) {
  if (is.null(radii_data)) {
    # If NULL, remove the custom data from the environment
    if (exists("user_radii_data", envir = .crystract_env)) {
      rm("user_radii_data", envir = .crystract_env)
    }
    message("Session radii data reset to package default.")
  } else {
    # --- Input Validation ---
    if (!is.data.frame(radii_data)) {
      stop("`radii_data` must be a data.frame or data.table.")
    }

    required_cols <- c("Symbol", "Radius", "Type")
    if (!all(required_cols %in% names(radii_data))) {
      stop(paste(
        "`radii_data` must contain the columns:",
        paste(required_cols, collapse = ", ")
      ))
    }

    # Coerce to data.table and store in the package environment
    custom_dt <- data.table::as.data.table(radii_data)

    if (!is.character(custom_dt$Symbol) ||
        !is.numeric(custom_dt$Radius) ||
        !is.character(custom_dt$Type)) {
      stop(
        "Please check column types: Symbol (character), Radius (numeric), Type (character)."
      )
    }

    assign("user_radii_data", custom_dt, envir = .crystract_env)
    message("Custom radii table has been set for the current R session.")
  }
  invisible(NULL)
}

#' @title Get Radii Data (Internal)
#' @description An internal helper function that retrieves the active radii table.
#' It prioritizes a user-defined table set via `set_radii_data`, falling
#' back to the package's default `covalent_radii` data if none is set.
#' @return A `data.table` of atomic radii.
#' @noRd
get_radii_data <- function() {
  if (exists("user_radii_data", envir = .crystract_env)) {
    return(get("user_radii_data", envir = .crystract_env))
  } else {
    # Returns the package's default data
    return(crystract::covalent_radii)
  }
}

# Ensure default data is available
utils::data("covalent_radii", package = "crystract", envir = environment())

test_that("filter_ghost_distances identifies plausible and implausible distances",
          {
            # Inject dummy radii for Si and Sr to ensure test is robust regardless of default table
            my_radii <- data.table::data.table(
              Symbol = c("Si", "Sr"),
              Radius = c(1.11, 1.92),
              Type = c("covalent", "covalent")
            )
            set_radii_data(my_radii)
            on.exit(set_radii_data(NULL), add = TRUE)

            distances <- distances_1590946

            # Extract a real distance from the structure to test the filter
            real_bond <- distances[Atom1 == "Si1" &
                                     grepl("Sr", Atom2)][order(Distance)][1, ]

            filtered_plausible <- filter_ghost_distances(real_bond, atoms_1590946, margin = 0.5)
            expect_equal(nrow(filtered_plausible$kept), 1)
            expect_equal(nrow(filtered_plausible$removed), 0)

            # --- Test Case: Known short, implausible distance ---
            fake_ghost <- data.table::data.table(Atom1 = "Si1",
                                                 Atom2 = "Si1_1_0_0_0",
                                                 Distance = 0.5)
            filtered_short <- filter_ghost_distances(fake_ghost, atoms_1590946, margin = 0.2)
            expect_equal(nrow(filtered_short$removed), 1)
            expect_equal(filtered_short$removed$Reason, "Distance is TOO SHORT")
          })

test_that("calculate_weighted_average_network_distance is correct", {
  bonds_md <- results_1590946$bonds_minimum_distance[[1]]

  # The Wyckoff positions for this file are missing in the CIF, so mock them
  test_atoms <- data.table::copy(atoms_1590946)
  test_atoms[, WyckoffSymbol := "c"]
  test_atoms[, WyckoffMultiplicity := 4]

  # Calculate distance based on the mocked 4c site
  avg_dist <- calculate_weighted_average_network_distance(bonds_md, test_atoms, "4c")

  expect_type(avg_dist, "double")
  expect_true(!is.na(avg_dist))
  expect_true(avg_dist > 1.0 &&
                avg_dist < 5.0) # Reasonable bond length sanity check
})

test_that("set_radii_data custom atomic radii management works", {
  # Create a fake user radii table
  my_radii <- data.table::data.table(
    Symbol = c("Si", "Sr"),
    Radius = c(1.0, 2.0),
    Type = c("covalent", "covalent")
  )

  # Set the custom radii
  set_radii_data(my_radii)

  # Fetch internal active radii
  active_radii <- crystract:::get_radii_data()
  expect_equal(nrow(active_radii), 2)
  expect_equal(active_radii[Symbol == "Si", Radius], 1.0)

  # Reset to default
  set_radii_data(NULL)

  # Verify reset
  default_radii <- crystract:::get_radii_data()
  expect_true(nrow(default_radii) > 2)
})

# This helper script prepares data for running tests locally.
# NOTE: Due to licensing restrictions, the CCDC/ICSD CIF files are not included in this package.
# To run these tests, you must:
#   1. Download the file for CCDC entry 1590946 (ICSD 422).
#   2. Save it as "1590946.cif" inside the 'tests/testthat/' directory.

cif_path_1590946 <- "1590946.cif"

# Only run the analysis if the test file has been provided by the user.
if (file.exists(cif_path_1590946)) {
  # Load the single CIF file content
  cif_content_1590946 <- read_cif_files(cif_path_1590946)[[1]]

  # Run the full analysis pipeline with multiple algorithms to get the final results object
  results_1590946 <- analyze_cif_files(
    cif_path_1590946,
    bonding_algorithms = c(
      "minimum_distance",
      "brunner",
      "econ",
      "crystal_nn",
      "voronoi"
    )
  )

  # For more granular testing, create the intermediate building blocks
  atoms_1590946 <- extract_atomic_coordinates(cif_content_1590946)
  metrics_1590946 <- extract_unit_cell_metrics(cif_content_1590946)
  sym_ops_1590946 <- extract_symmetry_operations(cif_content_1590946)

  # Coordinate processing results
  # Note: apply_symmetry_operations now requires unit_cell_metrics for PBC distance calculation
  full_cell_1590946 <- apply_symmetry_operations(atoms_1590946, sym_ops_1590946, metrics_1590946)
  super_cell_1590946 <- expand_transformed_coords(full_cell_1590946)
  distances_1590946 <- calculate_distances(atoms_1590946, super_cell_1590946, metrics_1590946)

} else {
  # Skip all tests that require the data if the file is not present.
  # A message will be printed during testing to inform the developer.
  testthat::skip("Skipping tests: '1590946.cif' not found. See 'helper-setup.R' for instructions.")
}

utils::globalVariables(
  c(
    # Core data.table symbols
    ".",
    ".N",
    ".I",

    # Column names used across multiple files
    "Angle",
    "AngleError",
    "Area",
    "Atom1",
    "Atom2",
    "BondStrength",
    "CentralAtom",
    "DeltaX",
    "DeltaY",
    "DeltaZ",
    "Distance",
    "DistanceError",
    "Element",
    "Element1",
    "Element2",
    "file_name",
    "FullWyckoff",
    "Label",
    "Neighbor1",
    "Neighbor2",
    "NeighborCount",
    "Occupancy",
    "Occupancy2",
    "Parent1",
    "Parent2",
    "ParentLabel",
    "Radius",
    "Radius1",
    "Radius2",
    "SolidAngle",
    "Symbol",
    "Type",
    "V1",
    "Weight",
    "WyckoffMultiplicity",
    "WyckoffSymbol",
    "x",
    "y",
    "z",
    "x_a",
    "x_error",
    "y_b",
    "y_error",
    "z_c",
    "z_error",

    # Variables for bonding algorithms
    "dcut",
    "dmin",
    "CutoffLow",
    "CutoffHigh",
    "Target",
    "R1",
    "R2",
    "RecipGap",
    "IsMaxGap",
    "CutoffDist",
    "MinDist",
    "FictiveR",
    "MEFIR",
    "Sym1",
    "Sym2",
    "EN1",
    "EN2",
    "RawWeight",
    "ChemMod",
    "MaxW",
    "Diameter",
    "CutLow",
    "CutHigh",
    "DistMod",
    "MaxSA",
    "Coordination",
    "OxidationState",
    "Diff",
    "Atom1_Index",
    "Atom2_Index",
    "Atom1_Label",
    "Atom2_Label",
    "Tx",
    "Ty",
    "Tz",
    "OS1",
    "i.OxidationState",
    "OS2",
    "i.MEFIR",
    "MaxScore",
    "RadiusSum",
    "RawScore",

    # Variables for calculate_angles
    "cos_theta",
    "cx",
    "cy",
    "cz",
    "dot_product",
    "i.x_a",
    "i.y_b",
    "i.z_c",
    "mag1",
    "mag2",
    "mag_sq1",
    "mag_sq2",
    "n1x",
    "n1y",
    "n1z",
    "n2x",
    "n2y",
    "n2z",
    "Neighbor1_sorted",
    "Neighbor2_sorted",
    "xf1",
    "xf2",
    "yf1",
    "yf2",
    "zf1",
    "zf2",

    # Variables for error propagation
    "a_vx",
    "a_vy",
    "a_vz",
    "b_vx",
    "b_vy",
    "b_vz",
    "C_val",
    "dx",
    "dy",
    "dz",
    "i.s_xc_sq",
    "i.s_yc_sq",
    "i.s_zc_sq",
    "i.xc",
    "i.yc",
    "i.zc",
    "mag_a",
    "mag_b",
    "Original_Atom2",
    "p_C_xc1",
    "p_C_xc2",
    "p_C_xc3",
    "p_C_yc1",
    "p_C_yc2",
    "p_C_yc3",
    "p_C_zc1",
    "p_C_zc2",
    "p_C_zc3",
    "p_xc_a",
    "p_xc_alpha",
    "p_xc_b",
    "p_xc_beta",
    "p_xc_c",
    "p_xc_gamma",
    "p_xc_xf",
    "p_xc_yf",
    "p_xc_zf",
    "p_yc_a",
    "p_yc_alpha",
    "p_yc_b",
    "p_yc_beta",
    "p_yc_c",
    "p_yc_gamma",
    "p_yc_xf",
    "p_yc_yf",
    "p_yc_zf",
    "p_zc_a",
    "p_zc_alpha",
    "p_zc_b",
    "p_zc_beta",
    "p_zc_c",
    "p_zc_gamma",
    "p_zc_xf",
    "p_zc_yf",
    "p_zc_zf",
    "pd_a",
    "pd_alpha",
    "pd_b",
    "pd_beta",
    "pd_c",
    "pd_gamma",
    "pd_xf1",
    "pd_xf2",
    "pd_yf1",
    "pd_yf2",
    "pd_zf1",
    "pd_zf2",
    "Parent_C",
    "Parent_N1",
    "Parent_N2",
    "s_C_sq",
    "s_theta_sq",
    "s_xc1_sq",
    "s_xc2_sq",
    "s_xc3_sq",
    "s_xc_sq",
    "s_yc1_sq",
    "s_yc2_sq",
    "s_yc3_sq",
    "s_yc_sq",
    "s_zc1_sq",
    "s_zc2_sq",
    "s_zc3_sq",
    "s_zc_sq",
    "s_xf",
    "s_xf1",
    "s_xf2",
    "s_yf",
    "s_yf1",
    "s_yf2",
    "s_zf",
    "s_zf1",
    "s_zf2",
    "term_common",
    "variance",
    "xc",
    "xc1",
    "xc2",
    "xc3",
    "yc",
    "yc1",
    "yc2",
    "yc3",
    "zc",
    "zc1",
    "zc2",
    "zc3",

    # Variables for post-processing and analysis
    "..meta_col_names",
    "..required_metrics",
    "expected_dist",
    "lower_bound",
    "n",
    "sum_d",
    "upper_bound",
    "null",
    "..valid_cols",

    # Voronoi NN specific
    "idx",
    "idx1",
    "idx2",
    "temp",
    "x1",
    "x2",
    "y1",
    "y2",
    "z1",
    "z2",
    "X",
    "center_coords",
    "my_tets",
    "neighbor_indices",
    "prim_label",

    # CrystalNN specific
    "atomic_rads",

    # Merge Sites
    "ClusterID",
    "N",

    # Unit cell metrics used in apply_symmetry_operations
    "unit_cell_metrics",

    # Internal
    "pauling_en",
    "shannon_radii",
    "NeighborSite",
    "get_default_radius",
    "get_ionic_radius",
    "m_i",
    "OccupancyError",
    "ThermalParam",
    "TypeSymbol",
    "gx",
    "gy",
    "gz"

  )
)

#' @title Generic Value Extractor (Internal)
#' @description Extracts a single value from CIF content based on a matching text pattern.
#' @param cif_content A data.table containing the lines of a CIF file.
#' @param patterns A character vector of text patterns to search for. The first matching pattern is returned.
#' @param remove_pattern A boolean indicating whether to remove the search pattern from the result.
#' @return A character string of the cleaned value, or NA if not found.
#' @noRd
extract_value <- function(cif_content, patterns, remove_pattern = TRUE) {
  for (pattern in patterns) {
    matching_lines <- cif_content[grepl(pattern, V1, fixed = TRUE)]
    if (nrow(matching_lines) > 0) {
      value <- matching_lines$V1[1]
      if (remove_pattern) {
        value <- gsub(pattern, "", value, fixed = TRUE)
      }
      value <- gsub("'", "", value)
      value <- trimws(value)
      return(value)
    }
  }
  return(NA)
}

#' @title Extract Database Code from CIF Content
#' @description Extracts the database code identifier (e.g., from `_database_code_`) from the CIF.
#' @param cif_content A `data.table` containing the lines of a CIF file.
#' @return A character string of the database code, or `NA` if not found.
#' @family extractors
#' @export
#' @examples
#' cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
#' if (file.exists(cif_path)) {
#'   cif_content <- read_cif_files(cif_path)[[1]]
#'   extract_database_code(cif_content)
#' }
extract_database_code <- function (cif_content) {
  extract_value(cif_content, "_database_code_")
}

#' @title Extract Chemical Formula from CIF Content
#' @description Extracts the chemical sum formula (e.g., from `_chemical_formula_sum`).
#' @param cif_content A `data.table` containing the lines of a CIF file.
#' @return A character string of the chemical formula, or `NA` if not found.
#' @family extractors
#' @export
#' @examples
#' cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
#' if (file.exists(cif_path)) {
#'   cif_content <- read_cif_files(cif_path)[[1]]
#'   extract_chemical_formula(cif_content)
#' }
extract_chemical_formula <- function(cif_content) {
  extract_value(
    cif_content,
    c(
      "_chemical_formula_sum",
      "_chemical_formula_structural",
      "_chemical_formula_moiety",
      "_chemical_formula_analytical"
    )
  )
}

#' @title Extract Structure Type from CIF Content
#' @description Extracts the structure type name (e.g., from `_chemical_name_structure_type`).
#' @param cif_content A `data.table` containing the lines of a CIF file.
#' @return A character string of the structure type, or `NA` if not found.
#' @family extractors
#' @export
#' @examples
#' cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
#' if (file.exists(cif_path)) {
#'   cif_content <- read_cif_files(cif_path)[[1]]
#'   extract_structure_type(cif_content)
#' }
extract_structure_type <- function(cif_content) {
  extract_value(cif_content, "_chemical_name_structure_type")
}

#' @title Extract Space Group Name from CIF Content
#' @description Extracts the Hermann-Mauguin space group name (e.g., from `_space_group_name_H-M_alt`).
#' @param cif_content A `data.table` containing the lines of a CIF file.
#' @return A character string of the space group name, or `NA` if not found.
#' @family extractors
#' @export
#' @examples
#' cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
#' if (file.exists(cif_path)) {
#'   cif_content <- read_cif_files(cif_path)[[1]]
#'   extract_space_group_name(cif_content)
#' }
extract_space_group_name <- function(cif_content) {
  extract_value(
    cif_content,
    c(
      "_space_group_name_H-M_alt",
      "_symmetry_space_group_name_H-M",
      "_space_group_name_H-M"
    )
  )
}

#' @title Extract Space Group Number from CIF Content
#' @description Extracts the International Tables space group number (e.g., from `_space_group_IT_number`).
#' @param cif_content A `data.table` containing the lines of a CIF file.
#' @return A character string of the space group number, or `NA` if not found.
#' @family extractors
#' @export
#' @examples
#' cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
#' if (file.exists(cif_path)) {
#'   cif_content <- read_cif_files(cif_path)[[1]]
#'   extract_space_group_number(cif_content)
#' }
extract_space_group_number <- function(cif_content) {
  extract_value(cif_content,
                c("_space_group_IT_number", "_symmetry_Int_Tables_number"))
}

#' @title Extract Unit Cell Metrics
#' @description Parses the unit cell parameters (lengths a, b, c and angles
#'   alpha, beta, gamma) and their associated standard uncertainties from CIF
#'   content.
#' @param cif_content A `data.table` containing the lines of a CIF file.
#' @return A one-row `data.table` with columns for each cell parameter and its
#'   error.
#' @family extractors
#' @export
#' @examples
#' cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
#' if (file.exists(cif_path)) {
#'   cif_content <- read_cif_files(cif_path)[[1]]
#'   extract_unit_cell_metrics(cif_content)
#' }
extract_unit_cell_metrics <- function(cif_content) {
  cell_parameters <- c(
    "_cell_length_a",
    "_cell_length_b",
    "_cell_length_c",
    "_cell_angle_alpha",
    "_cell_angle_beta",
    "_cell_angle_gamma"
  )
  values <- list()
  errors <- list()

  scale_error <- function(value_str, error_str) {
    if (is.na(error_str) || error_str == "")
      return(NA_real_)
    decimal_pos <- regexpr("\\.", value_str)
    if (decimal_pos == -1) {
      as.numeric(error_str)
    } else {
      as.numeric(error_str) * 10^-(nchar(value_str) - decimal_pos)
    }
  }

  for (param in cell_parameters) {
    line <- cif_content[grepl(param, V1, fixed = TRUE)]$V1
    if (length(line) > 0) {
      match <- stringr::str_match(line[1], "\\s+([0-9\\.]+)(?:\\(([0-9]+)\\))?")
      if (!is.na(match[1, 1])) {
        values[[param]] <- as.numeric(match[1, 2])
        errors[[paste0(param, "_error")]] <- scale_error(match[1, 2], match[1, 3])
      } else {
        values[[param]] <- NA_real_
        errors[[paste0(param, "_error")]] <- NA_real_
      }
    } else {
      values[[param]] <- NA_real_
      errors[[paste0(param, "_error")]] <- NA_real_
    }
  }
  return(data.table::as.data.table(c(values, errors)))
}

#' @title Extract Atom Type Information (Internal)
#' @description Parses the `_atom_type_` loop to get oxidation states and other
#' species-specific metadata defined globally in the CIF.
#' @param cif_content A `data.table` containing the lines of a CIF file.
#' @return A `data.table` mapping Symbol to OxidationState, or NULL.
#' @noRd
extract_atom_type_info <- function(cif_content) {
  # 1. Identify Loop
  header_pattern <- "^\\s*_atom_type_symbol"
  idx <- grep(header_pattern, cif_content$V1)

  if (length(idx) == 0)
    return(NULL)
  first_header_idx <- idx[1]

  # Find start of loop
  loop_start <- max(grep("^\\s*loop_", cif_content$V1[1:first_header_idx]))
  if (is.infinite(loop_start))
    return(NULL)

  # 2. Parse Headers
  line_indices <- (loop_start + 1):nrow(cif_content)
  headers <- character()
  first_data_line <- 0

  for (i in line_indices) {
    line <- trimws(cif_content$V1[i])
    if (line == "" || startsWith(line, "#"))
      next
    if (startsWith(line, "_")) {
      headers <- c(headers, line)
    } else {
      first_data_line <- i
      break
    }
  }

  # 3. Check for specific columns
  sym_idx <- which(headers == "_atom_type_symbol")
  ox_idx  <- which(headers == "_atom_type_oxidation_number")

  if (length(sym_idx) == 0)
    return(NULL)

  # 4. Extract Data Block
  end_candidates <- c(
    grep("^\\s*loop_|^\\s*_|^\\s*#", cif_content$V1[first_data_line:nrow(cif_content)]),
    grep("^\\s*$", cif_content$V1[first_data_line:nrow(cif_content)])
  )

  last_data_line <- if (length(end_candidates) > 0)
    first_data_line + min(end_candidates) - 2
  else
    nrow(cif_content)

  if (first_data_line > last_data_line)
    return(NULL)

  data_lines <- cif_content$V1[first_data_line:last_data_line]

  dt_raw <- data.table::fread(
    text = paste(data_lines, collapse = "\n"),
    header = FALSE,
    sep = "auto",
    quote = "",
    fill = TRUE
  )

  clean_val <- function(x)
    gsub("^'|'$|^\"|\"$", "", as.character(x))

  if (sym_idx > ncol(dt_raw))
    return(NULL)
  symbols <- clean_val(dt_raw[[sym_idx]])

  oxi_states <- rep(NA_real_, nrow(dt_raw))
  if (length(ox_idx) > 0 && ox_idx <= ncol(dt_raw)) {
    raw_ox <- clean_val(dt_raw[[ox_idx]])
    raw_ox <- gsub("\\(.*\\)", "", raw_ox)
    oxi_states <- as.numeric(raw_ox)
  }

  return(data.table::data.table(Symbol = symbols, OxidationState = oxi_states))
}

#' @title Extract Atomic Coordinates
#' @description Parses atomic site info, including labels, fractional coordinates,
#'   occupancies, oxidation states, and thermal parameters. It includes logic
#'   to extract oxidation states from both site tags and type loops.
#' @param cif_content A `data.table` containing the lines of a CIF file.
#' @param chemical_formula The chemical formula string for validation.
#' @return A `data.table` with atomic coordinate data, or `NULL` if not found.
#' @family extractors
#' @export
#' @examples
#' cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
#' if (file.exists(cif_path)) {
#'   cif_content <- read_cif_files(cif_path)[[1]]
#'   extract_atomic_coordinates(cif_content)
#' }
extract_atomic_coordinates <- function(cif_content, chemical_formula = NA) {
  # --- 0. Pre-fetch Atom Type Info (for Oxidation State Lookup) ---
  atom_type_info <- extract_atom_type_info(cif_content)

  # --- 1. Find Header Block ---
  first_header_line_idx <- grep("^\\s*_atom_site_fract_x", cif_content$V1)
  if (length(first_header_line_idx) == 0) {
    first_header_line_idx <- grep("^\\s*_atom_site_label", cif_content$V1)
    if (length(first_header_line_idx) == 0)
      return(NULL)
  }
  first_header_line_idx <- first_header_line_idx[1]

  # --- 2. Find Loop Start ---
  loop_start_line_idx <- max(grep("^\\s*loop_", cif_content$V1[1:first_header_line_idx]))
  if (is.infinite(loop_start_line_idx))
    return(NULL)

  # --- 3. Parse Headers ---
  line_indices <- (loop_start_line_idx + 1):nrow(cif_content)
  headers <- character()
  first_data_line_idx <- 0

  for (i in line_indices) {
    line <- trimws(cif_content$V1[i])
    if (line == "" || startsWith(line, "#"))
      next
    if (startsWith(line, "_")) {
      headers <- c(headers, line)
    } else {
      first_data_line_idx <- i
      break
    }
  }

  tags_to_find <- list(
    label = c("_atom_site_label"),
    type_symbol = c("_atom_site_type_symbol", "_atom_site_type_symbol_"),
    x = c("_atom_site_fract_x"),
    y = c("_atom_site_fract_y"),
    z = c("_atom_site_fract_z"),
    occupancy = c("_atom_site_occupancy"),
    wyckoff = c(
      "_atom_site_Wyckoff_symbol",
      "_atom_site_wyckoff_symbol",
      "_atom_site_Wychoff_symbol",
      "_atom_site_calc_flag"
    ),
    multiplicity = c(
      "_atom_site_symmetry_multiplicity",
      "_atom_site_site_symmetry_multiplicity"
    ),
    oxidation = c("_atom_site_oxidation_number"),
    u_iso = c("_atom_site_U_iso_or_equiv"),
    b_iso = c("_atom_site_B_iso_or_equiv")
  )

  # Grabs the index of the first matched potential alias for a given metric
  col_indices <- sapply(tags_to_find, function(tags) {
    for (tag in tags) {
      idx <- which(headers == tag)
      if (length(idx) > 0)
        return(idx[1])
    }
    return(NA)
  })

  if (anyNA(col_indices[c("label", "x", "y", "z")]))
    return(NULL)

  # --- 4. Determine Data Block Range ---
  end_candidates <- c(
    grep("^\\s*loop_|^\\s*_|^\\s*#", cif_content$V1[first_data_line_idx:nrow(cif_content)]),
    grep("^\\s*$", cif_content$V1[first_data_line_idx:nrow(cif_content)])
  )

  last_data_line_idx <- if (length(end_candidates) > 0)
    first_data_line_idx + min(end_candidates) - 2
  else
    nrow(cif_content)

  if (first_data_line_idx > last_data_line_idx)
    return(NULL)

  # --- 5. Read Data ---
  data_lines <- cif_content$V1[first_data_line_idx:last_data_line_idx]

  # fill=TRUE handles the footer warning ("Discarded single-line footer: <<END>>")
  atom_data <- data.table::fread(
    text = paste(data_lines, collapse = "\n"),
    header = FALSE,
    sep = "auto",
    quote = "",
    fill = TRUE
  )

  # --- 6. Helper Functions ---
  clean_str <- function(col_data) {
    gsub("^'|'$|^\"|\"$", "", as.character(col_data))
  }

  parse_vector_with_error <- function(coord_vector) {
    coord_vector <- clean_str(coord_vector)
    matches <- stringr::str_match(coord_vector,
                                  "([0-9\\.\\-]+(?:[eE][+\\-]?[0-9]+)?)(?:\\(([0-9]+)\\))?")
    value_str <- matches[, 2]
    error_str <- matches[, 3]

    values <- as.numeric(value_str)
    errors <- mapply(function(val_s, err_s) {
      if (is.na(err_s) || err_s == "")
        return(0)
      if (is.na(val_s))
        return(NA_real_)
      if (grepl("[eE]", val_s))
        return(as.numeric(err_s))
      decimal_pos <- regexpr("\\.", val_s)
      decimal_places <- ifelse(decimal_pos == -1, 0, nchar(val_s) - decimal_pos)
      as.numeric(err_s) * 10^(-decimal_places)
    }, value_str, error_str)

    return(list(value = values, error = errors))
  }

  safe_extract <- function(idx) {
    if (is.na(idx) ||
        idx > ncol(atom_data))
      return(rep(NA, nrow(atom_data)))
    return(atom_data[[idx]])
  }

  # --- 7. Extract and Parse Core Columns ---
  x_data <- parse_vector_with_error(safe_extract(col_indices["x"]))
  y_data <- parse_vector_with_error(safe_extract(col_indices["y"]))
  z_data <- parse_vector_with_error(safe_extract(col_indices["z"]))

  occ_data <- if (!is.na(col_indices["occupancy"]))
    parse_vector_with_error(safe_extract(col_indices["occupancy"]))
  else
    list(value = rep(1.0, nrow(atom_data)),
         error = rep(NA_real_, nrow(atom_data)))

  thermal_val <- rep(NA_real_, nrow(atom_data))
  if (!is.na(col_indices["u_iso"])) {
    thermal_val <- parse_vector_with_error(safe_extract(col_indices["u_iso"]))$value
  } else if (!is.na(col_indices["b_iso"])) {
    thermal_val <- parse_vector_with_error(safe_extract(col_indices["b_iso"]))$value
  }

  # --- 8. Advanced Oxidation State Logic ---
  site_ox_vals <- rep(NA_real_, nrow(atom_data))

  # Method A: Direct Site Oxidation
  if (!is.na(col_indices["oxidation"])) {
    raw_ox <- clean_str(safe_extract(col_indices["oxidation"]))
    raw_ox <- gsub("\\(.*\\)", "", raw_ox)
    site_ox_vals <- as.numeric(raw_ox)
  }

  # Method B: Lookup via Type Symbol if direct method fails
  if (all(is.na(site_ox_vals)) && !is.null(atom_type_info)) {
    site_symbols <- if (!is.na(col_indices["type_symbol"])) {
      clean_str(safe_extract(col_indices["type_symbol"]))
    } else {
      lbls <- clean_str(safe_extract(col_indices["label"]))
      gsub("[0-9_\\+\\-].*", "", lbls)
    }
    match_idx <- match(site_symbols, atom_type_info$Symbol)
    site_ox_vals <- atom_type_info$OxidationState[match_idx]
  }

  # --- 8b. Snap Coordinates and Occupancies to Ideal Fractions ---
  # Ensures exact boundaries to prevent downstream float-comparison issues
  x_vals <- snap_to_fraction(x_data$value, wrap = TRUE)
  y_vals <- snap_to_fraction(y_data$value, wrap = TRUE)
  z_vals <- snap_to_fraction(z_data$value, wrap = TRUE)
  occ_vals <- snap_to_fraction(occ_data$value, wrap = FALSE)

  # --- 9. Build Initial DataTable ---
  atomic_coordinates <- data.table::data.table(
    Label = clean_str(safe_extract(col_indices["label"])),
    TypeSymbol = if (!is.na(col_indices["type_symbol"]))
      clean_str(safe_extract(col_indices["type_symbol"]))
    else
      NA_character_,
    WyckoffSymbol = if (!is.na(col_indices["wyckoff"]))
      clean_str(safe_extract(col_indices["wyckoff"]))
    else
      NA_character_,
    WyckoffMultiplicity = if (!is.na(col_indices["multiplicity"]))
      as.numeric(clean_str(safe_extract(col_indices["multiplicity"])))
    else
      NA_real_,
    OxidationState = site_ox_vals,
    ThermalParam = thermal_val,
    Occupancy = occ_vals,
    OccupancyError = occ_data$error,
    x_a = x_vals,
    y_b = y_vals,
    z_c = z_vals,
    x_error = x_data$error,
    y_error = y_data$error,
    z_error = z_data$error
  )

  atomic_coordinates <- atomic_coordinates[!is.na(x_a) &
                                             !is.na(y_b) &
                                             !is.na(z_c)]
  if (nrow(atomic_coordinates) == 0)
    return(NULL)

  # --- 10. Heuristic Label Correction ---
  valid_elements <- c(
    "H",
    "He",
    "Li",
    "Be",
    "B",
    "C",
    "N",
    "O",
    "F",
    "Ne",
    "Na",
    "Mg",
    "Al",
    "Si",
    "P",
    "S",
    "Cl",
    "Ar",
    "K",
    "Ca",
    "Sc",
    "Ti",
    "V",
    "Cr",
    "Mn",
    "Fe",
    "Co",
    "Ni",
    "Cu",
    "Zn",
    "Ga",
    "Ge",
    "As",
    "Se",
    "Br",
    "Kr",
    "Rb",
    "Sr",
    "Y",
    "Zr",
    "Nb",
    "Mo",
    "Tc",
    "Ru",
    "Rh",
    "Pd",
    "Ag",
    "Cd",
    "In",
    "Sn",
    "Sb",
    "Te",
    "I",
    "Xe",
    "Cs",
    "Ba",
    "La",
    "Ce",
    "Pr",
    "Nd",
    "Pm",
    "Sm",
    "Eu",
    "Gd",
    "Tb",
    "Dy",
    "Ho",
    "Er",
    "Tm",
    "Yb",
    "Lu",
    "Hf",
    "Ta",
    "W",
    "Re",
    "Os",
    "Ir",
    "Pt",
    "Au",
    "Hg",
    "Tl",
    "Pb",
    "Bi",
    "Po",
    "At",
    "Rn",
    "Fr",
    "Ra",
    "Ac",
    "Th",
    "Pa",
    "U",
    "Np",
    "Pu",
    "Am",
    "Cm",
    "Bk",
    "Cf",
    "Es",
    "Fm",
    "Md",
    "No",
    "Lr",
    "Rf",
    "Db",
    "Sg",
    "Bh",
    "Hs",
    "Mt",
    "Ds",
    "Rg",
    "Cn",
    "Nh",
    "Fl",
    "Mc",
    "Lv",
    "Ts",
    "Og"
  )

  base_symbols <- if (!all(is.na(atomic_coordinates$TypeSymbol))) {
    stringr::str_extract(atomic_coordinates$TypeSymbol, "^[A-Za-z]+")
  } else {
    stringr::str_extract(atomic_coordinates$Label, "^[A-Za-z]+")
  }

  unique_base_symbols <- unique(base_symbols)
  non_standard_symbols <- setdiff(unique_base_symbols, valid_elements)
  corrections_made <- list()

  if (length(non_standard_symbols) > 0) {
    elements_in_formula <- if (!is.na(chemical_formula))
      unique(stringr::str_extract_all(chemical_formula, "[A-Z][a-z]?")[[1]])
    else
      character(0)
    for (sym in non_standard_symbols) {
      corrected_sym <- NA_character_
      if (toupper(sym) %in% c("OW", "WAT", "OH")) {
        if ("O" %in% elements_in_formula ||
            length(elements_in_formula) == 0)
          corrected_sym <- "O"
      } else {
        prefix2 <- substr(sym, 1, 2)
        prefix1 <- substr(sym, 1, 1)
        if (prefix2 %in% valid_elements &&
            (prefix2 %in% elements_in_formula ||
             length(elements_in_formula) == 0)) {
          corrected_sym <- prefix2
        } else if (prefix1 %in% valid_elements &&
                   (prefix1 %in% elements_in_formula ||
                    length(elements_in_formula) == 0)) {
          corrected_sym <- prefix1
        }
      }
      if (!is.na(corrected_sym)) {
        indices_to_fix <- which(base_symbols == sym)
        atomic_coordinates[indices_to_fix, Label := sub("^[A-Za-z]+", corrected_sym, Label)]
        corrections_made[[sym]] <- corrected_sym
      }
    }
  }

  if (length(corrections_made) > 0) {
    correction_messages <- sapply(names(corrections_made), function(key)
      paste0("'", key, "' -> '", corrections_made[[key]], "'"))
    warning(paste0(
      "Corrected non-standard atom labels: ",
      paste(correction_messages, collapse = ", ")
    ))
  }

  # --- 11. Normalize Labels and Ensure Uniqueness ---
  symbols <- sub("^([A-Z][a-z]?).*", "\\1", atomic_coordinates$Label, perl = TRUE)
  numbers <- stringr::str_remove_all(atomic_coordinates$Label, "[^0-9]")
  atomic_coordinates[, Label := paste0(symbols, numbers)]

  if (any(duplicated(atomic_coordinates$Label))) {
    atomic_coordinates[, Label := make.unique(Label, sep = "")]
  }

  return(atomic_coordinates)
}

#' @title Extract Symmetry Operations
#' @description Parses the symmetry operation definitions from the CIF content.
#' @param cif_content A `data.table` containing the lines of a CIF file.
#' @return A `data.table` with symmetry operations. Returns `NULL` if not found.
#' @family extractors
#' @export
#' @examples
#' cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
#' if (file.exists(cif_path)) {
#'   cif_content <- read_cif_files(cif_path)[[1]]
#'   extract_symmetry_operations(cif_content)
#' }
extract_symmetry_operations <- function(cif_content) {
  target_tags <- c("_space_group_symop_operation_xyz",
                   "_symmetry_equiv_pos_as_xyz")
  header_matches <- grep(paste(target_tags, collapse = "|"), cif_content$V1)
  if (length(header_matches) == 0)
    return(NULL)
  first_header_line_idx <- header_matches[1]

  loop_start_line_idx <- max(grep("^\\s*loop_", cif_content$V1[1:first_header_line_idx]))
  if (is.infinite(loop_start_line_idx))
    return(NULL)

  line_indices <- (loop_start_line_idx + 1):nrow(cif_content)
  headers <- character()
  first_data_line_idx <- 0
  for (i in line_indices) {
    line <- trimws(cif_content$V1[i])
    if (line == "" || startsWith(line, "#"))
      next
    if (startsWith(line, "_")) {
      headers <- c(headers, line)
    } else {
      first_data_line_idx <- i
      break
    }
  }

  xyz_col_idx <- which(headers %in% target_tags)
  if (length(xyz_col_idx) == 0)
    return(NULL)
  xyz_col_idx <- xyz_col_idx[1]

  end_candidates <- c(
    grep("^\\s*loop_|^\\s*_|^\\s*#", cif_content$V1[first_data_line_idx:nrow(cif_content)]),
    grep("^\\s*$", cif_content$V1[first_data_line_idx:nrow(cif_content)])
  )
  last_data_line_idx <- if (length(end_candidates) > 0)
    first_data_line_idx + min(end_candidates) - 2
  else
    nrow(cif_content)
  if (first_data_line_idx > last_data_line_idx)
    return(NULL)

  data_lines <- cif_content$V1[first_data_line_idx:last_data_line_idx]

  extract_nth_token <- function(line, n) {
    line <- trimws(line)
    pattern <- "(?:'[^']*'|\"[^\"]*\"|\\S+)"
    matches <- gregexpr(pattern, line)
    tokens <- regmatches(line, matches)[[1]]
    if (length(tokens) < n)
      return(NA)
    return(tokens[n])
  }

  raw_strings <- sapply(data_lines, extract_nth_token, n = xyz_col_idx)
  raw_strings <- raw_strings[!is.na(raw_strings)]
  if (length(raw_strings) == 0)
    return(NULL)

  cleaned_strings <- gsub("\\s+", "", raw_strings)
  cleaned_strings <- gsub("['\"]", "", cleaned_strings)
  cleaned_strings <- tolower(cleaned_strings)

  symmetry_matrix <- stringr::str_split_fixed(cleaned_strings, ",", n = 3)
  symmetry_dt <- data.table::data.table(x = symmetry_matrix[, 1], y = symmetry_matrix[, 2], z = symmetry_matrix[, 3])
  symmetry_dt <- symmetry_dt[x != "" & y != "" & z != ""]

  if (nrow(symmetry_dt) == 0)
    return(NULL)
  return(symmetry_dt)
}

test_that("Metadata extraction works for 1590946", {
  # Check that the extracted database code string contains 1590946
  db_code <- extract_database_code(cif_content_1590946)
  expect_true(grepl("1590946", db_code))

  expect_equal(extract_chemical_formula(cif_content_1590946), "Si1 Sr2")
  expect_equal(extract_space_group_name(cif_content_1590946), "P n m a")
})

test_that("Unit cell metrics extraction works for 1590946", {
  expect_s3_class(metrics_1590946, "data.table")
  expect_equal(nrow(metrics_1590946), 1)
  expect_equal(metrics_1590946$`_cell_length_a`, 8.11)
  expect_equal(metrics_1590946$`_cell_length_b`, 5.15)
  expect_equal(metrics_1590946$`_cell_length_c`, 9.54)
})

test_that("Atomic coordinate extraction works for 1590946", {
  expect_s3_class(atoms_1590946, "data.table")
  expect_equal(nrow(atoms_1590946), 3)

  sr1 <- atoms_1590946[Label == "Sr1"]
  expect_equal(sr1$x_a, 0.6529)
  expect_equal(sr1$y_b, 0.25)

  # Ensure parsing didn't fail on missing error parameters (should default to 0)
  expect_equal(sr1$x_error, 0)
})

test_that("Symmetry operations extraction works for 1590946", {
  expect_s3_class(sym_ops_1590946, "data.table")
  expect_equal(nrow(sym_ops_1590946), 8)
  # The first operation in this specific file is 'x+1/2, y, -z+1/2'
  expect_equal(sym_ops_1590946$x[1], "x+1/2")
  expect_equal(sym_ops_1590946$y[1], "y")
  expect_equal(sym_ops_1590946$z[1], "-z+1/2")
  # The identity operation is last in this file
  expect_equal(sym_ops_1590946$x[8], "x")
})

#' Atomic Radii Data for Bond-Length Estimation
#'
#' A data.table containing atomic radii for elements. This data is used for
#' estimating plausible bond lengths to filter out non-physical "ghost" distances
#' that can occur in disordered structures. Radii are in Angstroms (Å).
#'
#' @format A data table with three columns:
#' \describe{
#' \item{Symbol}{The chemical symbol of the element.}
#' \item{Radius}{The atomic radius in Angstroms.}
#' \item{Type}{The type of radius, e.g., "covalent". The default table only contains covalent radii.}
#' }
#' @source J. Emsley. The Elements. Third edition 1998, Oxford University Press.
#' As provided by Julia-Maria Huebner.
#' @export
#' @examples
#' head(covalent_radii)
"covalent_radii"

#' @keywords internal
.onLoad <- function(libname, pkgname) {
  # This makes the data available for use within the package's internal functions.
  utils::data("covalent_radii",
              package = pkgname,
              envir = parent.env(environment()))
}

#' @title Calculate Interatomic Distances
#' @description Computes the distances between a central set of atoms and an
#'   expanded set, using the metric tensor for accuracy.
#' @param atomic_coordinates A `data.table` of the primary (asymmetric) atom set.
#' @param expanded_coords A `data.table` of atoms in the expanded supercell.
#' @param unit_cell_metrics A `data.table` with cell parameters.
#' @param tolerance A numeric cutoff (default 1e-6). Distances smaller than this
#'   value are considered floating-point noise (overlapping atoms) and are
#'   filtered out.
#' @return A `data.table` of all non-zero distances.
#' @family property calculators
#' @export
#' @examples
#' ac <- data.table::data.table(Label = "Si1", x_a = 0, y_b = 0, z_c = 0)
#' ec <- data.table::data.table(Label = "O1_1", x_a = 0.5, y_b = 0.5, z_c = 0.5)
#' uc <- data.table::data.table(`_cell_length_a` = 10, `_cell_length_b` = 10,
#'                              `_cell_length_c` = 10, `_cell_angle_alpha` = 90,
#'                              `_cell_angle_beta` = 90, `_cell_angle_gamma` = 90)
#' calculate_distances(ac, ec, uc)
calculate_distances <- function(atomic_coordinates,
                                expanded_coords,
                                unit_cell_metrics,
                                tolerance = 1e-6) {
  # Safe extraction of cell metrics
  a <- as.numeric(unit_cell_metrics$`_cell_length_a`)
  b <- as.numeric(unit_cell_metrics$`_cell_length_b`)
  c <- as.numeric(unit_cell_metrics$`_cell_length_c`)
  alpha <- as.numeric(unit_cell_metrics$`_cell_angle_alpha`) * pi / 180
  beta <- as.numeric(unit_cell_metrics$`_cell_angle_beta`) * pi / 180
  gamma <- as.numeric(unit_cell_metrics$`_cell_angle_gamma`) * pi / 180

  coords_atomic <- as.matrix(atomic_coordinates[, .(x_a, y_b, z_c)])
  coords_expanded <- as.matrix(expanded_coords[, .(x_a, y_b, z_c)])
  labels_atomic <- atomic_coordinates$Label
  labels_expanded <- expanded_coords$Label

  delta_x <- outer(coords_atomic[, 1], coords_expanded[, 1], "-")
  delta_y <- outer(coords_atomic[, 2], coords_expanded[, 2], "-")
  delta_z <- outer(coords_atomic[, 3], coords_expanded[, 3], "-")

  cos_alpha <- cos(alpha)
  cos_beta <- cos(beta)
  cos_gamma <- cos(gamma)

  r2 <- (a^2 * delta_x^2) + (b^2 * delta_y^2) + (c^2 * delta_z^2) +
    (2 * b * c * cos_alpha * delta_y * delta_z) +
    (2 * c * a * cos_beta * delta_z * delta_x) +
    (2 * a * b * cos_gamma * delta_x * delta_y)

  r <- sqrt(r2)

  # FIX: stringsAsFactors = FALSE prevents "Ops.factor" warnings downstream
  atom_pairs <- expand.grid(
    Atom1 = labels_atomic,
    Atom2 = labels_expanded,
    KEEP.OUT.ATTRS = TRUE,
    stringsAsFactors = FALSE
  )

  distances <- data.table(
    Atom1 = atom_pairs$Atom1,
    Atom2 = atom_pairs$Atom2,
    Distance = as.vector(r),
    DeltaX = as.vector(delta_x),
    DeltaY = as.vector(delta_y),
    DeltaZ = as.vector(delta_z)
  )

  # Filter out self-interactions or near-zero distances based on tolerance
  distances <- distances[Distance > tolerance]

  return(distances)
}

#' @title Calculate Coordination Numbers
#' @description Counts the number of nearest neighbors for each central atom based on
#'   a table of bonded pairs.
#' @param bonded_pairs_table A `data.table` of bonded pairs.
#' @return A `data.table` with columns 'Atom' and 'CoordinationNumber'.
#' @family property calculators
#' @export
#' @examples
#' bp <- data.table::data.table(Atom1 = c("Si1", "Si1", "O1"),
#'                              Atom2 = c("O1", "O2", "Si1"))
#' calculate_neighbor_counts(bp)
calculate_neighbor_counts <- function(bonded_pairs_table) {
  if (is.null(bonded_pairs_table) || nrow(bonded_pairs_table) == 0) {
    return(data.table(Atom = character(), CoordinationNumber = integer()))
  }

  neighbor_counts <- bonded_pairs_table[, .(CoordinationNumber = .N), by = .(Atom1)]

  setnames(neighbor_counts, "Atom1", "Atom")

  return(neighbor_counts)
}

#' @title Calculate Weighted Coordination Numbers
#' @description Counts the weighted number of nearest neighbors for each central atom based on
#'   a table of bonded pairs, accounting for the fractional occupancy of the neighbor sites.
#'   This is particularly useful for analyzing disordered crystal structures.
#'
#'   If all occupancies are 1.0, this efficiently returns the standard coordination number.
#'   If the fractional occupancies on any single crystallographic site sum to greater than 1.0,
#'   this indicates an invalid or physically impossible structure; the function will throw a
#'   warning and return `NULL` so the file can be discarded by the analysis pipeline.
#'
#' @param bonded_pairs_table A `data.table` of bonded pairs (e.g., from `minimum_distance`).
#' @param atomic_coordinates A `data.table` of asymmetric atoms from `extract_atomic_coordinates`.
#' @return A `data.table` with columns 'Atom', 'CoordinationNumber', and 'WeightedCoordinationNumber'.
#'   Returns `NULL` if the occupancies sum to > 1 on any shared site.
#' @family property calculators
#' @export
#' @examples
#' bp <- data.table::data.table(Atom1 = c("Si1", "Si1"),
#'                              Atom2 = c("O1_1_0_0", "O2_1_0_0"))
#' ac <- data.table::data.table(Label = c("Si1", "O1", "O2"),
#'                              Occupancy = c(1.0, 0.5, 0.5),
#'                              x_a=c(0,0,0), y_b=c(0,0,0), z_c=c(0,0,0))
#' calculate_weighted_neighbor_counts(bp, ac)
calculate_weighted_neighbor_counts <- function(bonded_pairs_table, atomic_coordinates) {
  if (is.null(bonded_pairs_table) || nrow(bonded_pairs_table) == 0) {
    return(
      data.table(
        Atom = character(),
        CoordinationNumber = integer(),
        WeightedCoordinationNumber = numeric()
      )
    )
  }

  coords <- copy(atomic_coordinates)

  # Ensure Occupancy column exists and handle NAs
  if (!"Occupancy" %in% names(coords)) {
    coords[, Occupancy := 1.0]
  } else {
    coords[is.na(Occupancy), Occupancy := 1.0]
  }

  # --- 1. VALIDATION ---
  # Check if fractional occupancies sum to > 1 at any physical crystallographic site.
  # Grouping by exact snapped coordinates identifies atoms sharing the same site.
  site_occ <- coords[, .(SumOccupancy = sum(Occupancy)), by = .(x_a, y_b, z_c)]

  # We use 1.01 to allow for minor floating-point rounding errors (e.g., 1/3 + 2/3)
  if (any(site_occ$SumOccupancy > 1.01)) {
    warning(
      "Fractional occupancies add up to greater than 1.0 on one or more sites. Discarding file."
    )
    return(NULL)
  }

  # --- 2. OPTIMIZATION ---
  # If all occupancies are fully occupied (1.0), skip the fractional math (WCN == CN)
  if (all(abs(coords$Occupancy - 1.0) < 1e-5)) {
    neighbor_counts <- bonded_pairs_table[, .(CoordinationNumber = .N,
                                              WeightedCoordinationNumber = as.numeric(.N)), by = .(Atom1)]

    setnames(neighbor_counts, "Atom1", "Atom")
    return(neighbor_counts)
  }

  # --- 3. CALCULATE WEIGHTED CN ---
  work_dt <- copy(bonded_pairs_table)

  # Extract the parent site label (e.g., "Fe1" from "Fe1_1_0_0")
  work_dt[, NeighborSite := sub("_.*", "", Atom2)]

  # Count how many times a specific asymmetric site appears as a neighbor (multiplicity m_i)
  site_counts <- work_dt[, .(m_i = .N), by = .(Atom1, NeighborSite)]

  # Merge occupancy info from the asymmetric unit coordinates
  site_counts <- merge(
    site_counts,
    coords[, .(Label, Occupancy)],
    by.x = "NeighborSite",
    by.y = "Label",
    all.x = TRUE
  )

  # Handle any missing occupancies resulting from the merge just in case
  site_counts[is.na(Occupancy), Occupancy := 1.0]

  # Weighted Sum: Sum(Multiplicity * Occupancy)
  w_cn <- site_counts[, .(
    CoordinationNumber = sum(m_i),
    WeightedCoordinationNumber = sum(m_i * Occupancy)
  ), by = .(Atom1)]

  setnames(w_cn, "Atom1", "Atom")

  return(w_cn)
}

#' @title Calculate Bond Angles
#' @description Calculates all bond angles centered on each atom, formed by pairs
#'   of its bonded neighbors.
#' @param bonded_pairs Data.table of bonded atoms.
#' @param atomic_coordinates Data.table of asymmetric atom coordinates.
#' @param expanded_coords Data.table of supercell atom coordinates.
#' @param unit_cell_metrics Data.table with unit cell parameters.
#' @return A `data.table` of all unique bond angles.
#' @family property calculators
#' @export
#' @examples
#' bp <- data.table::data.table(Atom1 = c("Si1", "Si1"), Atom2 = c("O1", "O2"))
#' ac <- data.table::data.table(Label = "Si1", x_a = 0, y_b = 0, z_c = 0)
#' ec <- data.table::data.table(Label = c("O1", "O2"),
#'                              x_a = c(1, 0), y_b = c(0, 1), z_c = c(0, 0))
#' uc <- data.table::data.table(`_cell_length_a` = 10, `_cell_length_b` = 10,
#'                              `_cell_length_c` = 10, `_cell_angle_alpha` = 90,
#'                              `_cell_angle_beta` = 90, `_cell_angle_gamma` = 90)
#' calculate_angles(bp, ac, ec, uc)
calculate_angles <- function(bonded_pairs,
                             atomic_coordinates,
                             expanded_coords,
                             unit_cell_metrics) {
  if (is.null(bonded_pairs) || nrow(bonded_pairs) == 0) {
    return(
      data.table(
        CentralAtom = character(),
        Neighbor1 = character(),
        Neighbor2 = character(),
        Angle = numeric()
      )
    )
  }

  # Extract unit cell parameters
  a <- unit_cell_metrics$`_cell_length_a`
  b <- unit_cell_metrics$`_cell_length_b`
  c <- unit_cell_metrics$`_cell_length_c`
  alpha_rad <- unit_cell_metrics$`_cell_angle_alpha` * pi / 180
  beta_rad <- unit_cell_metrics$`_cell_angle_beta` * pi / 180
  gamma_rad <- unit_cell_metrics$`_cell_angle_gamma` * pi / 180

  cos_alpha <- cos(alpha_rad)
  cos_beta <- cos(beta_rad)
  cos_gamma <- cos(gamma_rad)

  # Combine all unique atomic coordinates and set key for fast lookup
  # (Including expanded_coords ensures all possible neighbors are available)
  all_coords <- unique(rbind(atomic_coordinates[, .(Label, x_a, y_b, z_c)], expanded_coords[, .(Label, x_a, y_b, z_c)], fill = TRUE),
                       by = "Label")
  setkey(all_coords, Label)

  # Generate all unique (CentralAtom, Neighbor1, Neighbor2) triplets
  # Start with central atoms and their first neighbors
  angle_triplets <- bonded_pairs[, .(CentralAtom = Atom1, Neighbor1 = Atom2)]

  # Self-join to find all pairs of neighbors for each CentralAtom
  angle_triplets <- merge(
    angle_triplets,
    bonded_pairs[, .(CentralAtom = Atom1, Neighbor2 = Atom2)],
    by = "CentralAtom",
    allow.cartesian = TRUE # Needed to get all combinations
  )

  # Filter out cases where Neighbor1 is the same as Neighbor2
  angle_triplets <- angle_triplets[Neighbor1 != Neighbor2]

  # Ensure consistent ordering of Neighbor1 and Neighbor2 to avoid duplicate angles
  # (e.g., A-B-C is the same as C-B-A when B is central)
  angle_triplets[, `:=`(
    Neighbor1_sorted = pmin(Neighbor1, Neighbor2),
    Neighbor2_sorted = pmax(Neighbor1, Neighbor2)
  )]
  angle_triplets <- unique(angle_triplets,
                           by = c("CentralAtom", "Neighbor1_sorted", "Neighbor2_sorted"))

  # Merge fractional coordinates for all three atoms involved in each angle
  angle_triplets[all_coords, on = .(CentralAtom = Label), `:=`(cx = i.x_a, cy = i.y_b, cz = i.z_c)]
  angle_triplets[all_coords, on = .(Neighbor1_sorted = Label), `:=`(n1x = i.x_a, n1y = i.y_b, n1z = i.z_c)]
  angle_triplets[all_coords, on = .(Neighbor2_sorted = Label), `:=`(n2x = i.x_a, n2y = i.y_b, n2z = i.z_c)]

  # Calculate vectors v1 (CentralAtom -> Neighbor1) and v2 (CentralAtom -> Neighbor2)
  angle_triplets[, `:=`(
    xf1 = n1x - cx,
    yf1 = n1y - cy,
    zf1 = n1z - cz,
    xf2 = n2x - cx,
    yf2 = n2y - cy,
    zf2 = n2z - cz
  )]

  # Vectorized calculation of dot product using metric tensor
  angle_triplets[, dot_product := (
    xf1 * xf2 * a^2 + yf1 * yf2 * b^2 + zf1 * zf2 * c^2 +
      (xf1 * yf2 + yf1 * xf2) * a * b * cos_gamma +
      (xf1 * zf2 + zf1 * xf2) * a * c * cos_beta +
      (yf1 * zf2 + zf1 * yf2) * b * c * cos_alpha
  )]

  # Vectorized calculation of squared magnitudes for both vectors
  angle_triplets[, mag_sq1 := (
    xf1^2 * a^2 + yf1^2 * b^2 + zf1^2 * c^2 +
      2 * xf1 * yf1 * a * b * cos_gamma +
      2 * xf1 * zf1 * a * c * cos_beta +
      2 * yf1 * zf1 * b * c * cos_alpha
  )]
  angle_triplets[, mag_sq2 := (
    xf2^2 * a^2 + yf2^2 * b^2 + zf2^2 * c^2 +
      2 * xf2 * yf2 * a * b * cos_gamma +
      2 * xf2 * zf2 * a * c * cos_beta +
      2 * yf2 * zf2 * b * c * cos_alpha
  )]

  # Calculate magnitudes, guarding against division by zero for very small lengths
  angle_triplets[, `:=`(
    mag1 = sqrt(pmax(mag_sq1, 0)),
    # pmax(...,0) to handle potential negative due to FP error
    mag2 = sqrt(pmax(mag_sq2, 0))
  )]

  # Calculate cos_theta, guarding against division by zero or invalid acos input range [-1, 1]
  angle_triplets[, cos_theta := pmin(pmax(dot_product / (mag1 * mag2), -1.0), 1.0)]
  angle_triplets[mag1 <= 1e-10 |
                   mag2 <= 1e-10, cos_theta := NA_real_] # Set to NA if magnitudes are too small

  # Calculate angle in degrees
  angle_triplets[, Angle := acos(cos_theta) * 180 / pi]

  # Filter out NA angles (if any, due to impossible geometry) and select final columns
  result_angles <- angle_triplets[!is.na(Angle), .(
    CentralAtom = CentralAtom,
    Neighbor1 = Neighbor1_sorted,
    Neighbor2 = Neighbor2_sorted,
    Angle
  )]

  return(result_angles[order(CentralAtom, Neighbor1, Neighbor2)])
}

#' @title Propagate Distance Error
#' @description Calculates the standard uncertainty for each interatomic distance.
#' @param bonded_pairs Data.table of bonded atoms with their distances.
#' @param atomic_coordinates Data.table with fractional coordinates and errors.
#' @param unit_cell_metrics Data.table with unit cell parameters and errors.
#' @return The input `bonded_pairs` data.table with a new 'DistanceError' column.
#' @family error propagators
#' @export
#' @examples
#' bp <- data.table::data.table(Atom1 = "Si1", Atom2 = "O1_1", Distance = 1.6,
#'                              DeltaX = 0.1, DeltaY = 0.1, DeltaZ = 0)
#' ac <- data.table::data.table(Label = c("Si1", "O1"),
#'                              x_error = c(0.01, 0.01),
#'                              y_error = c(0.01, 0.01),
#'                              z_error = c(0.01, 0.01))
#' uc <- data.table::data.table(`_cell_length_a` = 10, `_cell_length_a_error` = 0.1,
#'                              `_cell_length_b` = 10, `_cell_length_b_error` = 0.1,
#'                              `_cell_length_c` = 10, `_cell_length_c_error` = 0.1,
#'                              `_cell_angle_alpha` = 90, `_cell_angle_alpha_error` = 0,
#'                              `_cell_angle_beta` = 90, `_cell_angle_beta_error` = 0,
#'                              `_cell_angle_gamma` = 90, `_cell_angle_gamma_error` = 0)
#' propagate_distance_error(bp, ac, uc)
propagate_distance_error <- function(bonded_pairs,
                                     atomic_coordinates,
                                     unit_cell_metrics) {
  if (is.null(bonded_pairs) ||
      nrow(bonded_pairs) == 0) {
    return(bonded_pairs[, DistanceError := NA_real_])
  }

  a <- unit_cell_metrics$`_cell_length_a`
  s_a <- unit_cell_metrics$`_cell_length_a_error`
  b <- unit_cell_metrics$`_cell_length_b`
  s_b <- unit_cell_metrics$`_cell_length_b_error`
  c <- unit_cell_metrics$`_cell_length_c`
  s_c <- unit_cell_metrics$`_cell_length_c_error`
  alpha_rad <- unit_cell_metrics$`_cell_angle_alpha` * pi / 180
  s_alpha_rad <- unit_cell_metrics$`_cell_angle_alpha_error` * pi / 180
  beta_rad <- unit_cell_metrics$`_cell_angle_beta` * pi / 180
  s_beta_rad <- unit_cell_metrics$`_cell_angle_beta_error` * pi / 180
  gamma_rad <- unit_cell_metrics$`_cell_angle_gamma` * pi / 180
  s_gamma_rad <- unit_cell_metrics$`_cell_angle_gamma_error` * pi / 180

  cos_a <- cos(alpha_rad)
  sin_a <- sin(alpha_rad)
  cos_b <- cos(beta_rad)
  sin_b <- sin(beta_rad)
  cos_g <- cos(gamma_rad)
  sin_g <- sin(gamma_rad)

  s_a <- ifelse(is.na(s_a), 0, s_a)
  s_b <- ifelse(is.na(s_b), 0, s_b)
  s_c <- ifelse(is.na(s_c), 0, s_c)
  s_alpha_rad <- ifelse(is.na(s_alpha_rad), 0, s_alpha_rad)
  s_beta_rad <- ifelse(is.na(s_beta_rad), 0, s_beta_rad)
  s_gamma_rad <- ifelse(is.na(s_gamma_rad), 0, s_gamma_rad)

  work_dt <- copy(bonded_pairs)
  setnames(work_dt, c("DeltaX", "DeltaY", "DeltaZ"), c("dx", "dy", "dz"))
  atom_errors <- atomic_coordinates[, .(Label,
                                        s_xf = x_error,
                                        s_yf = y_error,
                                        s_zf = z_error)]

  work_dt <- merge(
    work_dt,
    atom_errors,
    by.x = "Atom1",
    by.y = "Label",
    all.x = TRUE
  )
  setnames(work_dt,
           c("s_xf", "s_yf", "s_zf"),
           c("s_xf1", "s_yf1", "s_zf1"))
  work_dt[, Original_Atom2 := sub("_.*", "", Atom2)]
  work_dt <- merge(
    work_dt,
    atom_errors,
    by.x = "Original_Atom2",
    by.y = "Label",
    all.x = TRUE
  )
  setnames(work_dt,
           c("s_xf", "s_yf", "s_zf"),
           c("s_xf2", "s_yf2", "s_zf2"))
  work_dt[, Original_Atom2 := NULL]

  coord_err_cols <- c("s_xf1", "s_yf1", "s_zf1", "s_xf2", "s_yf2", "s_zf2")
  for (col in coord_err_cols)
    set(work_dt, which(is.na(work_dt[[col]])), col, 0)

  work_dt[, term_common := 1 / (2 * Distance)]
  work_dt[, `:=`(
    pd_a = (term_common * (
      2 * a * dx^2 + 2 * b * dx * dy * cos_g + 2 * c * dx * dz * cos_b
    )),
    pd_b = (term_common * (
      2 * b * dy^2 + 2 * a * dx * dy * cos_g + 2 * c * dy * dz * cos_a
    )),
    pd_c = (term_common * (
      2 * c * dz^2 + 2 * a * dx * dz * cos_b + 2 * b * dy * dz * cos_a
    )),
    pd_alpha = (term_common * (-2 * b * c * dy * dz * sin_a)),
    pd_beta = (term_common * (-2 * a * c * dx * dz * sin_b)),
    pd_gamma = (term_common * (-2 * a * b * dx * dy * sin_g))
  )]
  work_dt[, `:=`(
    pd_xf1 = (term_common * (
      2 * a^2 * dx + 2 * a * b * dy * cos_g + 2 * a * c * dz * cos_b
    )),
    pd_yf1 = (term_common * (
      2 * b^2 * dy + 2 * a * b * dx * cos_g + 2 * b * c * dz * cos_a
    )),
    pd_zf1 = (term_common * (
      2 * c^2 * dz + 2 * a * c * dx * cos_b + 2 * b * c * dy * cos_a
    ))
  )][, `:=`(
    pd_xf2 = -pd_xf1,
    pd_yf2 = -pd_yf1,
    pd_zf2 = -pd_zf1
  )]
  work_dt[, variance := (pd_a * s_a)^2 + (pd_b * s_b)^2 + (pd_c * s_c)^2 + (pd_alpha *
                                                                              s_alpha_rad)^2 + (pd_beta * s_beta_rad)^2 + (pd_gamma * s_gamma_rad)^2 + (pd_xf1 *
                                                                                                                                                          s_xf1)^2 + (pd_yf1 * s_yf1)^2 + (pd_zf1 * s_zf1)^2 + (pd_xf2 * s_xf2)^2 + (pd_yf2 *
                                                                                                                                                                                                                                       s_yf2)^2 + (pd_zf2 * s_zf2)^2]
  work_dt[, DistanceError := sqrt(variance)]
  return(merge(
    bonded_pairs,
    work_dt[, .(Atom1, Atom2, Distance, DistanceError)],
    by = c("Atom1", "Atom2", "Distance"),
    all.x = TRUE
  ))
}

#' @title Propagate Angle Error
#' @description Calculates the standard uncertainty for each bond angle.
#' @param bond_angles Data.table of calculated bond angles.
#' @param atomic_coordinates Data.table with fractional coordinates and errors.
#' @param expanded_coords Data.table of supercell atom coordinates.
#' @param unit_cell_metrics Data.table with unit cell parameters and errors.
#' @return The input `bond_angles` data.table with a new 'AngleError' column.
#' @family error propagators
#' @export
#' @examples
#' # 1. Create dummy bond angles
#' ba <- data.table::data.table(
#'   CentralAtom = "Si1", Neighbor1 = "O1", Neighbor2 = "O2", Angle = 109.5
#' )
#'
#' # 2. Create dummy atomic coordinates with errors
#' ac <- data.table::data.table(
#'   Label = c("Si1", "O1", "O2"),
#'   x_a = c(0, 0.1, 0), y_b = c(0, 0, 0.1), z_c = c(0, 0, 0),
#'   x_error = c(0.01, 0.01, 0.01),
#'   y_error = c(0.01, 0.01, 0.01),
#'   z_error = c(0.01, 0.01, 0.01)
#' )
#'
#' # 3. Create dummy expanded coordinates
#' ec <- data.table::data.table(
#'   Label = c("O1", "O2"),
#'   x_a = c(0.1, 0), y_b = c(0, 0.1), z_c = c(0, 0)
#' )
#'
#' # 4. Create dummy unit cell metrics
#' uc <- data.table::data.table(
#'   `_cell_length_a` = 10, `_cell_length_a_error` = 0.1,
#'   `_cell_length_b` = 10, `_cell_length_b_error` = 0.1,
#'   `_cell_length_c` = 10, `_cell_length_c_error` = 0.1,
#'   `_cell_angle_alpha` = 90, `_cell_angle_alpha_error` = 0,
#'   `_cell_angle_beta` = 90, `_cell_angle_beta_error` = 0,
#'   `_cell_angle_gamma` = 90, `_cell_angle_gamma_error` = 0
#' )
#'
#' # 5. Run the error propagation
#' propagate_angle_error(ba, ac, ec, uc)
propagate_angle_error <- function(bond_angles,
                                  atomic_coordinates,
                                  expanded_coords,
                                  unit_cell_metrics) {
  if (is.null(bond_angles) ||
      nrow(bond_angles) == 0) {
    return(bond_angles[, AngleError := NA_real_])
  }
  a <- unit_cell_metrics$`_cell_length_a`
  b <- unit_cell_metrics$`_cell_length_b`
  c <- unit_cell_metrics$`_cell_length_c`
  s_a <- ifelse(
    is.na(unit_cell_metrics$`_cell_length_a_error`),
    0,
    unit_cell_metrics$`_cell_length_a_error`
  )
  s_b <- ifelse(
    is.na(unit_cell_metrics$`_cell_length_b_error`),
    0,
    unit_cell_metrics$`_cell_length_b_error`
  )
  s_c <- ifelse(
    is.na(unit_cell_metrics$`_cell_length_c_error`),
    0,
    unit_cell_metrics$`_cell_length_c_error`
  )
  alpha_rad <- unit_cell_metrics$`_cell_angle_alpha` * pi / 180
  beta_rad <- unit_cell_metrics$`_cell_angle_beta` * pi / 180
  gamma_rad <- unit_cell_metrics$`_cell_angle_gamma` * pi / 180
  s_alpha_rad <- ifelse(
    is.na(unit_cell_metrics$`_cell_angle_alpha_error`),
    0,
    unit_cell_metrics$`_cell_angle_alpha_error` * pi / 180
  )
  s_beta_rad <- ifelse(
    is.na(unit_cell_metrics$`_cell_angle_beta_error`),
    0,
    unit_cell_metrics$`_cell_angle_beta_error` * pi / 180
  )
  s_gamma_rad <- ifelse(
    is.na(unit_cell_metrics$`_cell_angle_gamma_error`),
    0,
    unit_cell_metrics$`_cell_angle_gamma_error` * pi / 180
  )
  cos_a <- cos(alpha_rad)
  sin_a <- sin(alpha_rad)
  cos_b <- cos(beta_rad)
  sin_b <- sin(beta_rad)
  cos_g <- cos(gamma_rad)
  sin_g <- sin(gamma_rad)
  v_sq <- 1 - cos_a^2 - cos_b^2 - cos_g^2 + 2 * cos_a * cos_b * cos_g
  v <- sqrt(v_sq)
  cart_errors <- copy(atomic_coordinates)
  for (col in c("x_error", "y_error", "z_error"))
    set(cart_errors, which(is.na(cart_errors[[col]])), col, 0)
  cart_errors[, `:=`(
    p_xc_a = x_a,
    p_xc_b = y_b * cos_g,
    p_xc_c = z_c * cos_b,
    p_xc_alpha = 0,
    p_xc_beta = -c * z_c * sin_b,
    p_xc_gamma = -b * y_b * sin_g,
    p_xc_xf = a,
    p_xc_yf = b * cos_g,
    p_xc_zf = c * cos_b
  )]
  cart_errors[, `:=`(
    p_yc_a = 0,
    p_yc_b = y_b * sin_g,
    p_yc_c = z_c * (cos_a - cos_b * cos_g) / sin_g,
    p_yc_alpha = -c * z_c * sin_a / sin_g,
    p_yc_beta = c * z_c * sin_b * cos_g / sin_g,
    p_yc_gamma = b * y_b * cos_g + c * z_c * (cos_b - cos_a * cos_g) / sin_g^2,
    p_yc_xf = 0,
    p_yc_yf = b * sin_g,
    p_yc_zf = c * (cos_a - cos_b * cos_g) / sin_g
  )]
  cart_errors[, `:=`(
    p_zc_a = 0,
    p_zc_b = 0,
    p_zc_c = z_c * v / sin_g,
    p_zc_alpha = c * z_c * (cos_b * cos_g - cos_a) / (sin_g * v),
    p_zc_beta = c * z_c * (cos_a * cos_g - cos_b) / (sin_g * v),
    p_zc_gamma = -c * z_c * (v_sq * cos_g + sin_g^2 * (cos_a * cos_b - cos_g)) /
      (sin_g^2 * v),
    p_zc_xf = 0,
    p_zc_yf = 0,
    p_zc_zf = c * v / sin_g
  )]
  cart_errors[, s_xc_sq := (p_xc_a * s_a)^2 + (p_xc_b * s_b)^2 + (p_xc_c *
                                                                    s_c)^2 + (p_xc_alpha * s_alpha_rad)^2 + (p_xc_beta * s_beta_rad)^2 + (p_xc_gamma *
                                                                                                                                            s_gamma_rad)^2 + (p_xc_xf * x_error)^2 + (p_xc_yf * y_error)^2 + (p_xc_zf *
                                                                                                                                                                                                                z_error)^2]
  cart_errors[, s_yc_sq := (p_yc_a * s_a)^2 + (p_yc_b * s_b)^2 + (p_yc_c *
                                                                    s_c)^2 + (p_yc_alpha * s_alpha_rad)^2 + (p_yc_beta * s_beta_rad)^2 + (p_yc_gamma *
                                                                                                                                            s_gamma_rad)^2 + (p_yc_xf * x_error)^2 + (p_yc_yf * y_error)^2 + (p_yc_zf *
                                                                                                                                                                                                                z_error)^2]

  cart_errors[, s_zc_sq := (p_zc_a * s_a)^2 + (p_zc_b * s_b)^2 + (p_zc_c *
                                                                    s_c)^2 + (p_zc_alpha * s_alpha_rad)^2 + (p_zc_beta * s_beta_rad)^2 + (p_zc_gamma *
                                                                                                                                            s_gamma_rad)^2 + (p_zc_xf * x_error)^2 + (p_zc_yf * y_error)^2 + (p_zc_zf *
                                                                                                                                                                                                                z_error)^2]

  # Select only compatible columns from expanded_coords to avoid column mismatch errors
  all_frac_coords <- unique(rbind(atomic_coordinates[, .(Label, x_a, y_b, z_c)], expanded_coords[, .(Label, x_a, y_b, z_c)]), by = "Label")

  all_cart_coords <- all_frac_coords[, `:=`(
    xc = a * x_a + b * y_b * cos_g + c * z_c * cos_b,
    yc = b * y_b * sin_g + c * z_c * (cos_a - cos_b * cos_g) / sin_g,
    zc = c * z_c * v / sin_g
  )][, .(Label, xc, yc, zc)]
  setkey(all_cart_coords, Label)
  error_subset <- cart_errors[, .(Label, s_xc_sq, s_yc_sq, s_zc_sq)]
  setkey(error_subset, Label)
  work_dt <- copy(bond_angles)
  work_dt[all_cart_coords, on = .(CentralAtom = Label), `:=`(xc1 = i.xc, yc1 =
                                                               i.yc, zc1 = i.zc)]
  work_dt[all_cart_coords, on = .(Neighbor1 = Label), `:=`(xc2 = i.xc, yc2 =
                                                             i.yc, zc2 = i.zc)]
  work_dt[all_cart_coords, on = .(Neighbor2 = Label), `:=`(xc3 = i.xc, yc3 =
                                                             i.yc, zc3 = i.zc)]
  work_dt[, Parent_C := sub("_.*", "", CentralAtom)][, Parent_N1 := sub("_.*", "", Neighbor1)][, Parent_N2 :=
                                                                                                 sub("_.*", "", Neighbor2)]
  work_dt[error_subset, on = .(Parent_C = Label), `:=`(s_xc1_sq = i.s_xc_sq,
                                                       s_yc1_sq = i.s_yc_sq,
                                                       s_zc1_sq = i.s_zc_sq)]
  work_dt[error_subset, on = .(Parent_N1 = Label), `:=`(s_xc2_sq = i.s_xc_sq,
                                                        s_yc2_sq = i.s_yc_sq,
                                                        s_zc2_sq = i.s_zc_sq)]
  work_dt[error_subset, on = .(Parent_N2 = Label), `:=`(s_xc3_sq = i.s_xc_sq,
                                                        s_yc3_sq = i.s_yc_sq,
                                                        s_zc3_sq = i.s_zc_sq)]
  work_dt[, `:=`(
    a_vx = xc2 - xc1,
    a_vy = yc2 - yc1,
    a_vz = zc2 - zc1,
    b_vx = xc3 - xc1,
    b_vy = yc3 - yc1,
    b_vz = zc3 - zc1
  )]
  work_dt[, `:=`(
    mag_a = sqrt(a_vx^2 + a_vy^2 + a_vz^2),
    mag_b = sqrt(b_vx^2 + b_vy^2 + b_vz^2)
  )]
  work_dt <- work_dt[mag_a > 1e-9 & mag_b > 1e-9]
  work_dt[, C_val := pmin(pmax((a_vx * b_vx + a_vy * b_vy + a_vz * b_vz) /
                                 (mag_a * mag_b), -1.0), 1.0)]
  work_dt[, `:=`(
    p_C_xc1 = -(((
      b_vx / mag_b - C_val * a_vx / mag_a
    ) / mag_a) + ((
      a_vx / mag_a - C_val * b_vx / mag_b
    ) / mag_b)),
    p_C_yc1 = -(((
      b_vy / mag_b - C_val * a_vy / mag_a
    ) / mag_a) + ((
      a_vy / mag_a - C_val * b_vy / mag_b
    ) / mag_b)),
    p_C_zc1 = -(((
      b_vz / mag_b - C_val * a_vz / mag_a
    ) / mag_a) + ((
      a_vz / mag_a - C_val * b_vz / mag_b
    ) / mag_b))
  )]
  work_dt[, `:=`(
    p_C_xc2 = (b_vx / mag_b - C_val * a_vx / mag_a) / mag_a,
    p_C_yc2 = (b_vy / mag_b - C_val * a_vy / mag_a) / mag_a,
    p_C_zc2 = (b_vz / mag_b - C_val * a_vz / mag_a) / mag_a
  )]
  work_dt[, `:=`(
    p_C_xc3 = (a_vx / mag_a - C_val * b_vx / mag_b) / mag_b,
    p_C_yc3 = (a_vy / mag_a - C_val * b_vy / mag_b) / mag_b,
    p_C_zc3 = (a_vz / mag_a - C_val * b_vz / mag_b) / mag_b
  )]
  work_dt[, s_C_sq := (p_C_xc1^2 * s_xc1_sq) + (p_C_yc1^2 * s_yc1_sq) +
            (p_C_zc1^2 * s_zc1_sq) + (p_C_xc2^2 * s_xc2_sq) + (p_C_yc2^2 * s_yc2_sq) +
            (p_C_zc2^2 * s_zc2_sq) + (p_C_xc3^2 * s_xc3_sq) + (p_C_yc3^2 * s_yc3_sq) +
            (p_C_zc3^2 * s_zc3_sq)]
  work_dt[C_val^2 >= 1.0, s_theta_sq := 0][C_val^2 < 1.0, s_theta_sq := s_C_sq /
                                             (1 - C_val^2)]
  work_dt[, AngleError := sqrt(pmax(0, s_theta_sq)) * 180 / pi]
  return(merge(
    bond_angles,
    work_dt[, .(CentralAtom, Neighbor1, Neighbor2, AngleError)],
    by = c("CentralAtom", "Neighbor1", "Neighbor2"),
    all.x = TRUE
  ))
}

#' @title Filter Data by Atom Symbol Interactively
#' @description Prompts the user to select chemical elements to keep in a data table
#'   of bonds or angles. Filtering is based on matching the base chemical symbol
#'   in a specified column (e.g., "CentralAtom").
#'
#' @details
#' The function first identifies all unique base chemical symbols from the atom
#' labels in the specified column (e.g., it extracts 'C' from 'C1', 'Si' from 'Si2_1').
#' It then presents these symbols to the user and asks for a comma-separated list
#' of the symbols they wish to retain.
#'
#' The matching logic is designed to be specific to avoid ambiguity between elements.
#' For example, if the user enters 'C', the function will match labels like 'C1',
#' 'C2', 'C_10', or a lone 'C'. However, it will *not* match labels for different
#' elements that start with C, such as 'Cr1' or 'Ca2'. This is achieved by
#' constructing a regular expression that ensures the character(s) immediately
#' following the selected symbol are not alphabetical letters.
#'
#' This function is intended for interactive use.
#'
#' @param data_table A `data.table` object containing atomic information, such as
#'   the output from `calculate_angles` or `minimum_distance`.
#' @param atom_col A character string specifying the name of the column in
#'   `data_table` that contains the atom labels to filter by. Defaults to "CentralAtom".
#' @return A `data.table` filtered to include only the rows where the atom
#'   label in `atom_col` corresponds to one of the user-selected chemical symbols.
#'   If the user provides no input, an empty `data.table` is returned.
#' @family property calculators
#' @export
#' @examples
#' # 1. Create a sample data.table of bond angles
#' sample_angles <- data.table::data.table(
#'   CentralAtom = c("C1", "C2", "Si1", "Cr1", "O1", "O2", "C"),
#'   Neighbor1 = c("O1", "O2", "O1", "N1", "C1", "C2", "H1"),
#'   Neighbor2 = c("H1", "H2", "O2", "N2", "H3", "H4", "H2"),
#'   Angle = c(109.5, 109.5, 120.0, 90.0, 104.5, 104.5, 120)
#' )
#'
#' # 2. In an interactive R session, the function would prompt the user.
#' if (interactive()) {
#'   filtered_data <- filter_atoms_by_symbol(sample_angles, atom_col = "CentralAtom")
#'   print(filtered_data)
#' }
#'
filter_atoms_by_symbol <- function(data_table, atom_col = "CentralAtom") {
  # Check if the column exists in the data.table
  if (!atom_col %in% names(data_table)) {
    stop(paste("Column '", atom_col, "' not found in the data table.", sep = ""))
  }

  # Extract unique base chemical symbols from the specified column to show the user
  unique_labels <- unique(data_table[[atom_col]])
  # This regex extracts the initial sequence of one or more letters
  base_symbols <- sort(unique(stringr::str_extract(unique_labels, "^[A-Za-z]+")))

  if (length(base_symbols) == 0) {
    message("No recognizable atom symbols found in the specified column.")
    return(data_table) # Return original table if no symbols are found
  }

  # Interactively prompt the user for their choice of atoms
  message("Available base atom symbols found: ",
          paste(base_symbols, collapse = ", "))

  user_input <- readline(prompt = "Please enter the chemical symbols you want to filter for, separated by commas (e.g., C,Si,O):\n")

  # Process the user's input string
  symbols_to_keep <- trimws(strsplit(user_input, ",")[[1]])
  # Remove any empty strings that might result from trailing commas or empty input
  symbols_to_keep <- symbols_to_keep[symbols_to_keep != ""]

  # If the user did not provide any valid symbols, return an empty data.table
  if (length(symbols_to_keep) == 0) {
    message("No symbols entered. Returning an empty data table.")
    return(data_table[0, ])
  }

  # Construct the regular expression pattern for grepl
  # For each symbol (e.g., "C"), the pattern is (^C$)|(^C[^A-Za-z])
  # This matches the exact symbol OR the symbol followed by a non-alphabetic character.
  # This correctly distinguishes 'C' from 'Cr', for example.
  patterns <- sapply(symbols_to_keep, function(sym) {
    paste0("(^", sym, "$)|(^", sym, "[^A-Za-z])")
  })
  full_pattern <- paste(patterns, collapse = "|")

  # Filter the data.table using the generated regex pattern on the specified column
  filtered_dt <- data_table[grepl(full_pattern, get(atom_col))]

  return(filtered_dt)
}

#' @title Filter Data by Wyckoff Symbol
#' @description Filters a data table (e.g., bonds or angles) to include only
#'   entries where a specified atom occupies one of the given Wyckoff sites.
#'
#' @details
#' This function is designed to facilitate analysis based on crystallographic
#' site symmetry. It is particularly useful for analyzing complex structures like
#' clathrates where different atoms perform distinct structural roles based
#' on their site symmetry.
#'
#' The function works by:
#' 1.  Creating a full Wyckoff label (e.g., "4c", "24k") by combining the
#'     `WyckoffMultiplicity` and `WyckoffSymbol` columns from the
#'     `atomic_coordinates` table.
#' 2.  Identifying the parent atom for each entry in the `data_table`.
#' 3.  Merging this Wyckoff information into the results table.
#' 4.  Filtering to keep only rows where the atom's full Wyckoff label matches
#'     one of the symbols provided in the `wyckoff_symbols` vector.
#'
#' @param data_table A `data.table` object, such as one produced by
#'   `minimum_distance` or
#'   `calculate_angles`.
#' @param atomic_coordinates A `data.table` from `extract_atomic_coordinates`
#'   containing `WyckoffMultiplicity` and `WyckoffSymbol` columns.
#' @param atom_col A character string specifying the column in `data_table`
#'   that contains the atom labels to filter by (e.g., "Atom1", "CentralAtom").
#' @param wyckoff_symbols A character vector of the full Wyckoff symbols to
#'   keep (e.g., `c("4c")` or `c("6c", "16i", "24k")`).
#' @return A `data.table` filtered to include only rows where the specified
#'   atom occupies one of the desired Wyckoff sites.
#' @family property calculators
#' @export
#' @examples
#' cif_file <- system.file("extdata", "1590946.cif", package = "crystract")
#' if (file.exists(cif_file)) {
#'   old_threads <- data.table::setDTthreads(2)
#'   # 1. Perform a standard analysis to get bond and coordinate tables
#'   cif_content <- read_cif_files(cif_file)[[1]]
#'   atoms <- extract_atomic_coordinates(cif_content)
#'   metrics <- extract_unit_cell_metrics(cif_content)
#'   sym_ops <- extract_symmetry_operations(cif_content)
#'   full_cell <- apply_symmetry_operations(atoms, sym_ops, metrics)
#'   super_cell <- expand_transformed_coords(full_cell)
#'   dists <- calculate_distances(atoms, super_cell, metrics)
#'   bonds <- minimum_distance(dists)
#'
#'   # 2. Mock Wyckoff sites since 1590946 doesn't explicitly declare them
#'   atoms[, WyckoffSymbol := "c"]
#'   atoms[, WyckoffMultiplicity := 4]
#'
#'   print("Original atomic coordinates showing Wyckoff sites:")
#'   print(atoms[, .(Label, WyckoffSymbol, WyckoffMultiplicity)])
#'
#'   filtered_bonds <- filter_by_wyckoff_symbol(
#'     data_table = bonds,
#'     atomic_coordinates = atoms,
#'     atom_col = "Atom1",
#'     wyckoff_symbols = "4c"
#'   )
#'
#'   cat("\nNumber of bonds in original table:", nrow(bonds), "\n")
#'   cat("Number of bonds after filtering for '4c' site:", nrow(filtered_bonds), "\n")
#'   data.table::setDTthreads(old_threads)
#' }
filter_by_wyckoff_symbol <- function(data_table,
                                     atomic_coordinates,
                                     atom_col,
                                     wyckoff_symbols) {
  # --- Input Validation ---
  if (!atom_col %in% names(data_table)) {
    stop(paste("Column '", atom_col, "' not found in the data table.", sep = ""))
  }
  required_cols <- c("WyckoffSymbol", "WyckoffMultiplicity", "Label")
  if (!all(required_cols %in% names(atomic_coordinates))) {
    stop(paste(
      "`atomic_coordinates` must contain the columns:",
      paste(required_cols, collapse = ", ")
    ))
  }
  if (!is.character(wyckoff_symbols)) {
    stop("`wyckoff_symbols` must be a character vector.")
  }

  # --- Data Preparation ---
  work_dt <- copy(data_table)

  # Extract parent labels to merge with atomic_coordinates
  work_dt[, ParentLabel := sub("_.*", "", get(atom_col))]

  # Prepare the Wyckoff info lookup table
  wyckoff_info <- atomic_coordinates[, .(Label, FullWyckoff = paste0(WyckoffMultiplicity, WyckoffSymbol))]

  # --- Merging and Filtering ---
  merged_dt <- merge(
    work_dt,
    wyckoff_info,
    by.x = "ParentLabel",
    by.y = "Label",
    all.x = TRUE
  )

  # Filter based on the provided list of full Wyckoff symbols
  filtered_dt <- merged_dt[FullWyckoff %in% wyckoff_symbols, ]

  # --- Clean and Return ---
  # Remove the temporary columns used for the merge
  filtered_dt[, `:=`(ParentLabel = NULL, FullWyckoff = NULL)]

  return(filtered_dt)
}

#' @title Filter Distances by Element Symbols
#' @description Removes any distance pair where at least one of the atoms
#'   corresponds to a specified list of element symbols.
#' @param distances A `data.table` of interatomic distances.
#' @param atomic_coordinates A `data.table` of atomic coordinates.
#' @param elements_to_exclude A character vector of element symbols to exclude.
#' @return A `data.table` of distances with the specified elements removed.
#' @family property calculators
#' @export
#' @examples
#' dists <- data.table::data.table(Atom1 = "Si1_1_0_0", Atom2 = "O1_1_0_0", Distance = 1.6)
#' ac <- data.table::data.table(Label = c("Si1", "O1"))
#' filter_by_elements(dists, ac, elements_to_exclude = "O")
filter_by_elements <- function(distances,
                               atomic_coordinates,
                               elements_to_exclude) {
  if (is.null(elements_to_exclude) ||
      length(elements_to_exclude) == 0) {
    return(distances)
  }
  element_info <- atomic_coordinates[, .(ParentLabel = Label,
                                         Element = sub("[0-9].*", "", Label))]
  dist_with_elements <- copy(distances)
  dist_with_elements[, Parent1 := sub("_.*", "", Atom1)]
  dist_with_elements[, Parent2 := sub("_.*", "", Atom2)]
  merged <- merge(
    dist_with_elements,
    element_info,
    by.x = "Parent1",
    by.y = "ParentLabel",
    all.x = TRUE
  )
  setnames(merged, "Element", "Element1")
  merged <- merge(
    merged,
    element_info,
    by.x = "Parent2",
    by.y = "ParentLabel",
    all.x = TRUE
  )
  setnames(merged, "Element", "Element2")
  filtered <- merged[!(Element1 %in% elements_to_exclude) &
                       !(Element2 %in% elements_to_exclude)]
  return(filtered[, names(distances), with = FALSE])
}

#' @title Filter Ghost Distances Using Atomic Radii
#' @description Cleans a distance table by removing physically implausible distances.
#' It uses a table of atomic radii to establish a plausible bond length
#' range for each atom pair. Any calculated distance falling outside this range
#' (defined by a `margin`) is considered a "ghost" distance and is removed. This
#' is particularly useful for cleaning data from disordered crystal structures.
#' @param distances A `data.table` of interatomic distances, typically from
#' `calculate_distances`. Must contain `Atom1`, `Atom2`, and `Distance`.
#' @param atomic_coordinates A `data.table` of asymmetric atoms from
#' `extract_atomic_coordinates`. Used to link atom labels to element types.
#' @param margin A numeric value (default 0.1) specifying the tolerance. A
#' distance `d` between atoms with radii `r1` and `r2` is kept if
#' `(r1+r2)*(1-margin) <= d <= (r1+r2)*(1+margin)`.
#' @param radii_type A character string specifying the type of radius to use for
#' the calculation (e.g., `"covalent"`, `"ionic"`). This value must correspond
#' to an entry in the `Type` column of the active radii table. Defaults to `"covalent"`.
#' The radii table can be customized for the session using `set_radii_data()`.
#' @return A list containing two `data.table`s:
#' \item{kept}{The distances considered physically plausible.}
#' \item{removed}{The "ghost" distances that were filtered out, with columns
#' explaining the reason for removal.}
#' @family post-processing
#' @export
#' @examples
#' # Create minimal dummy data for demonstration
#' distances <- data.table::data.table(
#'   Atom1 = c("Si1_1_0_0", "O1_1_0_0"),
#'   Atom2 = c("O1_1_0_0", "Si1_1_0_0"),
#'   Distance = c(1.6, 0.5) # 0.5 is implausibly short
#' )
#'
#' atomic_coords <- data.table::data.table(
#'   Label = c("Si1", "O1")
#' )
#'
#' # Run the filter
#' result <- filter_ghost_distances(distances, atomic_coords, margin = 0.1)
#'
#' print(result$kept)
#' print(result$removed)
filter_ghost_distances <- function(distances,
                                   atomic_coordinates,
                                   margin = 0.1,
                                   radii_type = "covalent") {
  # --- 1. Augment data for calculation ---
  # Create a lookup for element type from the parent atom label
  element_info <- atomic_coordinates[, .(ParentLabel = Label,
                                         Element = sub("[0-9].*", "", Label))]

  # Use the new helper to get the active radii table (user-defined or default)
  all_radii_data <- get_radii_data()

  # Filter the radii table for the specific type requested by the user
  radii_lookup <- all_radii_data[Type == radii_type]

  if (nrow(radii_lookup) == 0) {
    stop(
      paste0(
        "Radii type '",
        radii_type,
        "' not found in the active radii table. ",
        "Available types are: ",
        paste(unique(all_radii_data$Type), collapse = ", ")
      )
    )
  }

  # Keep only Symbol and Radius for the merge
  radii_lookup <- radii_lookup[, .(Symbol, Radius)]

  merged <- copy(distances)
  # Create Parent labels in the distances table to match element_info
  merged[, `:=`(Parent1 = sub("_.*", "", Atom1),
                Parent2 = sub("_.*", "", Atom2))]
  # Merge to get Element and Radius for Atom1
  merged <- merge(
    merged,
    element_info,
    by.x = "Parent1",
    by.y = "ParentLabel",
    all.x = TRUE
  )
  setnames(merged, "Element", "Element1")
  merged <- merge(
    merged,
    radii_lookup,
    by.x = "Element1",
    by.y = "Symbol",
    all.x = TRUE
  )
  setnames(merged, "Radius", "Radius1")
  # Merge to get Element and Radius for Atom2
  merged <- merge(
    merged,
    element_info,
    by.x = "Parent2",
    by.y = "ParentLabel",
    all.x = TRUE
  )
  setnames(merged, "Element", "Element2")
  merged <- merge(
    merged,
    radii_lookup,
    by.x = "Element2",
    by.y = "Symbol",
    all.x = TRUE
  )
  setnames(merged, "Radius", "Radius2")
  # Handle cases where radii are not found (e.g., vacancies) by setting them to 0
  merged[is.na(Radius1), Radius1 := 0][is.na(Radius2), Radius2 := 0]

  # --- 2. Calculate the plausible distance range ---
  merged[, expected_dist := Radius1 + Radius2]
  merged[, lower_bound := expected_dist * (1 - margin)]
  merged[, upper_bound := expected_dist * (1 + margin)]

  # --- 3. Apply the filter ---
  kept_mask <- merged$Distance >= merged$lower_bound &
    merged$Distance <= merged$upper_bound &
    merged$expected_dist > 0
  kept_dt <- merged[kept_mask, names(distances), with = FALSE]
  # Create a detailed table of removed rows for quality control
  removed_dt <- merged[!kept_mask, .(
    Atom1,
    Atom2,
    Distance,
    expected_dist,
    lower_bound,
    upper_bound,
    Reason = ifelse(
      Distance < lower_bound,
      "Distance is TOO SHORT",
      "Distance is TOO LONG"
    )
  )]
  return(list(kept = kept_dt, removed = removed_dt))
}

#' @title Calculate Weighted Average Network Bond Distance
#' @description Computes a single, representative bond length for a specified atomic
#'   network. This function precisely implements the validated logic that accounts
#'   for site multiplicity and occupancy of the central atoms.
#' @param distances A `data.table` of interatomic distances filtered to include
#'   **only bonded pairs** (e.g., from `minimum_distance`).
#' @param atomic_coordinates A `data.table` of asymmetric atoms from
#'   `extract_atomic_coordinates`.
#' @param wyckoff_symbols A character vector of Wyckoff symbols defining the
#'   atomic network (e.g., `c("6c", "16i", "24k")`). Must be the full symbol.
#' @return A single numeric value representing the weighted average bond distance.
#' @family post-processing
#' @export
#' @examples
#' dists <- data.table::data.table(Atom1 = "Si1", Atom2 = "O1", Distance = 1.6)
#' ac <- data.table::data.table(Label = c("Si1", "O1"),
#'                              WyckoffMultiplicity = c(4, 4),
#'                              WyckoffSymbol = c("c", "c"),
#'                              Occupancy = c(1, 1))
#' calculate_weighted_average_network_distance(dists, ac, wyckoff_symbols = "4c")
calculate_weighted_average_network_distance <- function(distances,
                                                        atomic_coordinates,
                                                        wyckoff_symbols) {
  # --- 1. Prepare atomic info and identify network atoms ---
  atom_info <- copy(atomic_coordinates)
  atom_info[, FullWyckoff := paste0(WyckoffMultiplicity, WyckoffSymbol)]
  network_atom_labels <- atom_info[FullWyckoff %in% wyckoff_symbols, Label]

  if (length(network_atom_labels) == 0) {
    return(NA_real_)
  }

  # --- 2. Filter bonds originating from network atoms ---
  network_distances <- distances[sub("_.*", "", Atom1) %in% network_atom_labels]

  if (nrow(network_distances) == 0) {
    return(NA_real_)
  }

  # This block explicitly merges neighbor occupancy information, which is a
  # crucial data preparation step from the validated workflow.
  network_distances[, Parent2 := sub("_.*", "", Atom2)]
  network_distances <- merge(
    network_distances,
    atom_info[, .(Label, Occupancy2 = Occupancy)],
    by.x = "Parent2",
    by.y = "Label",
    all.x = TRUE
  )
  network_distances[is.na(Occupancy2), Occupancy2 := 1]

  # --- 3. Calculate SIMPLE sum of distances (sum_d) and coordination (n) ---
  atom_dist_summary <- network_distances[, .(
    sum_d = sum(Distance),
    # Simple, unweighted sum of bond lengths
    n = .N               # Simple, unweighted coordination number
  ), by = Atom1]

  atom_dist_summary[, ParentLabel := sub("_.*", "", Atom1)]

  # --- 4. Merge summary with main atomic info ---
  unique_summary <- unique(atom_dist_summary, by = "ParentLabel")
  merged_data <- merge(
    atom_info[Label %in% network_atom_labels],
    unique_summary[, .(ParentLabel, sum_d, n)],
    by.x = "Label",
    by.y = "ParentLabel",
    all.x = TRUE
  )

  merged_data <- merged_data[!is.na(sum_d)]
  if (nrow(merged_data) == 0)
    return(NA_real_)

  # --- 5. Calculate the final weighted average ---
  numerator <- sum(
    merged_data$WyckoffMultiplicity * merged_data$Occupancy * merged_data$sum_d,
    na.rm = TRUE
  )
  denominator <- sum(merged_data$WyckoffMultiplicity * merged_data$Occupancy * merged_data$n,
                     na.rm = TRUE)

  if (is.na(denominator) || denominator == 0) {
    return(NA_real_)
  }

  final_result <- numerator / denominator
  return(final_result)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calculate_properties.R
\name{calculate_weighted_neighbor_counts}
\alias{calculate_weighted_neighbor_counts}
\title{Calculate Weighted Coordination Numbers}
\usage{
calculate_weighted_neighbor_counts(bonded_pairs_table, atomic_coordinates)
}
\arguments{
\item{bonded_pairs_table}{A \code{data.table} of bonded pairs (e.g., from \code{minimum_distance}).}

\item{atomic_coordinates}{A \code{data.table} of asymmetric atoms from \code{extract_atomic_coordinates}.}
}
\value{
A \code{data.table} with columns 'Atom', 'CoordinationNumber', and 'WeightedCoordinationNumber'.
Returns \code{NULL} if the occupancies sum to > 1 on any shared site.
}
\description{
Counts the weighted number of nearest neighbors for each central atom based on
a table of bonded pairs, accounting for the fractional occupancy of the neighbor sites.
This is particularly useful for analyzing disordered crystal structures.

If all occupancies are 1.0, this efficiently returns the standard coordination number.
If the fractional occupancies on any single crystallographic site sum to greater than 1.0,
this indicates an invalid or physically impossible structure; the function will throw a
warning and return \code{NULL} so the file can be discarded by the analysis pipeline.
}
\examples{
bp <- data.table::data.table(Atom1 = c("Si1", "Si1"),
                             Atom2 = c("O1_1_0_0", "O2_1_0_0"))
ac <- data.table::data.table(Label = c("Si1", "O1", "O2"),
                             Occupancy = c(1.0, 0.5, 0.5),
                             x_a=c(0,0,0), y_b=c(0,0,0), z_c=c(0,0,0))
calculate_weighted_neighbor_counts(bp, ac)
}
\seealso{
Other property calculators: 
\code{\link{calculate_angles}()},
\code{\link{calculate_distances}()},
\code{\link{calculate_neighbor_counts}()},
\code{\link{filter_atoms_by_symbol}()},
\code{\link{filter_by_elements}()},
\code{\link{filter_by_wyckoff_symbol}()}
}
\concept{property calculators}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict_bonds.R
\name{crystal_nn}
\alias{crystal_nn}
\title{Identify Atomic Bonds using CrystalNN}
\usage{
crystal_nn(
  distances,
  atomic_coordinates,
  expanded_coords,
  unit_cell_metrics,
  cutoff_length = 7,
  x_diff_weight = 3,
  porosity_adjustment = TRUE,
  distance_cutoffs = c(0.5, 1)
)
}
\arguments{
\item{distances}{Ignored.}

\item{atomic_coordinates}{Primary atom set.}

\item{expanded_coords}{Expanded supercell.}

\item{unit_cell_metrics}{Cell parameters.}

\item{cutoff_length}{Voronoi cutoff. Default 7.0 matches Pymatgen CrystalNN default.}

\item{x_diff_weight}{Electronegativity weight (default 3.0).}

\item{porosity_adjustment}{Logical (default TRUE).}

\item{distance_cutoffs}{Vector c(0.5, 1.0).}
}
\value{
A \code{data.table} of bonded pairs.
}
\description{
Port of Pymatgen's \code{CrystalNN} weighting logic. Includes Porosity,
Electronegativity, and Distance penalties.
Uses specific Decision Tree logic for radii (Shannon -> Covalent -> Atomic) per
pymatgen's fallback logic.
}
\examples{
\donttest{
cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
if (file.exists(cif_path)) {
  old_threads <- data.table::setDTthreads(2)
  res <- analyze_single_cif(cif_path, bonding_algorithms = "crystal_nn")
  print(head(res$bonds_crystal_nn[[1]]))
  data.table::setDTthreads(old_threads)
}
}
}
\references{
Shannon, R. D. (1976). "Revised effective ionic radii and systematic studies of
interatomic distances in halides and chalcogenides." \emph{Acta Crystallographica
Section A}, 32(5), 751-767.

Pauling, L. (1960). \emph{The Nature of the Chemical Bond and the Structure of
Molecules and Crystals: An Introduction to Modern Structural Chemistry}.
Cornell University Press.

Pan, H., Ganose, A. M., Horton, M., et al. (2021). "Benchmarking Coordination Number
Prediction Algorithms on Inorganic Crystal Structures." \emph{Inorganic Chemistry},
60(3), 1590–1603. \doi{10.1021/acs.inorgchem.0c02996}
}
\seealso{
Other bonding algorithms: 
\code{\link{brunner_nn_reciprocal}()},
\code{\link{econ_nn}()},
\code{\link{minimum_distance}()},
\code{\link{voronoi_nn}()}
}
\concept{bonding algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calculate_properties.R
\name{calculate_neighbor_counts}
\alias{calculate_neighbor_counts}
\title{Calculate Coordination Numbers}
\usage{
calculate_neighbor_counts(bonded_pairs_table)
}
\arguments{
\item{bonded_pairs_table}{A \code{data.table} of bonded pairs.}
}
\value{
A \code{data.table} with columns 'Atom' and 'CoordinationNumber'.
}
\description{
Counts the number of nearest neighbors for each central atom based on
a table of bonded pairs.
}
\examples{
bp <- data.table::data.table(Atom1 = c("Si1", "Si1", "O1"),
                             Atom2 = c("O1", "O2", "Si1"))
calculate_neighbor_counts(bp)
}
\seealso{
Other property calculators: 
\code{\link{calculate_angles}()},
\code{\link{calculate_distances}()},
\code{\link{calculate_weighted_neighbor_counts}()},
\code{\link{filter_atoms_by_symbol}()},
\code{\link{filter_by_elements}()},
\code{\link{filter_by_wyckoff_symbol}()}
}
\concept{property calculators}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_cif_files.R
\name{export_analysis_to_csv}
\alias{export_analysis_to_csv}
\title{Export Analysis Results to a Directory of CSVs}
\usage{
export_analysis_to_csv(analysis_results, output_dir, overwrite = FALSE)
}
\arguments{
\item{analysis_results}{A \code{data.table} object.}

\item{output_dir}{Path to main output directory.}

\item{overwrite}{Logical.}
}
\value{
Invisibly returns output_dir.
}
\description{
Exports analysis results to CSV structure.
}
\examples{
\donttest{
cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
if (file.exists(cif_path)) {
  old_threads <- data.table::setDTthreads(2)
  res <- analyze_single_cif(cif_path)

  out_dir <- file.path(tempdir(), "cif_csvs")
  export_analysis_to_csv(res, output_dir = out_dir)

  unlink(out_dir, recursive = TRUE)
  data.table::setDTthreads(old_threads)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict_bonds.R
\name{brunner_nn_reciprocal}
\alias{brunner_nn_reciprocal}
\title{Identify Atomic Bonds using Brunner's Method (Reciprocal)}
\usage{
brunner_nn_reciprocal(distances, tol = 0)
}
\arguments{
\item{distances}{A \code{data.table} of interatomic distances.}

\item{tol}{A small distance offset (default 0.0) to add to the cutoff.}
}
\value{
A \code{data.table} of bonded pairs.
}
\description{
An alternative bonding detection method using the
largest reciprocal gap. Matches \code{pymatgen.analysis.local_env.BrunnerNNReciprocal}.
}
\examples{
dists <- data.table::data.table(Atom1 = c("A", "A"), Atom2 = c("B", "C"),
                                Distance = c(1.5, 2.5),
                                DeltaX = c(1, 0), DeltaY = c(0, 1), DeltaZ = c(0, 0))
brunner_nn_reciprocal(dists)
}
\references{
Brunner, G. O. (1977). "A Definition of Coordination and Its Relevance in the
Structure Types AlB2 and NiAs." \emph{Acta Crystallographica Section A}, 33(1),
226–227. \doi{10.1107/S0567739477000461}
}
\seealso{
Other bonding algorithms: 
\code{\link{crystal_nn}()},
\code{\link{econ_nn}()},
\code{\link{minimum_distance}()},
\code{\link{voronoi_nn}()}
}
\concept{bonding algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{covalent_radii}
\alias{covalent_radii}
\title{Atomic Radii Data for Bond-Length Estimation}
\format{
A data table with three columns:
\describe{
\item{Symbol}{The chemical symbol of the element.}
\item{Radius}{The atomic radius in Angstroms.}
\item{Type}{The type of radius, e.g., "covalent". The default table only contains covalent radii.}
}
}
\source{
J. Emsley. The Elements. Third edition 1998, Oxford University Press.
As provided by Julia-Maria Huebner.
}
\usage{
covalent_radii
}
\description{
A data.table containing atomic radii for elements. This data is used for
estimating plausible bond lengths to filter out non-physical "ghost" distances
that can occur in disordered structures. Radii are in Angstroms (Å).
}
\examples{
head(covalent_radii)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict_bonds.R
\name{econ_nn}
\alias{econ_nn}
\title{Identify Atomic Bonds using Hoppe's EconNN Method}
\usage{
econ_nn(distances, atomic_coordinates, tol = 0.2, use_fictive_radius = FALSE)
}
\arguments{
\item{distances}{A \code{data.table} of interatomic distances.}

\item{atomic_coordinates}{A \code{data.table} of atomic coordinates used to map species to radii.}

\item{tol}{A bond strength cutoff (default 0.2).}

\item{use_fictive_radius}{Logical. If \code{TRUE}, calculates Hoppe's fictive ionic radius.}
}
\value{
A \code{data.table} of bonded pairs.
}
\description{
Uses Effective Coordination Numbers (ECoN) and Mean Fictive Ionic Radii (MEFIR).
}
\examples{
dists <- data.table::data.table(Atom1 = c("Si1", "Si1"), Atom2 = c("O1", "O2"),
                                Distance = c(1.6, 2.0),
                                DeltaX = c(1, 0), DeltaY = c(0, 1), DeltaZ = c(0, 0))
ac <- data.table::data.table(Label = c("Si1", "O1", "O2"),
                             OxidationState = c(4, -2, -2))
econ_nn(dists, ac)
}
\references{
Hoppe, R. (1979). "Effective Coordination Numbers (ECoN) and Mean Fictive Ionic
Radii (MEFIR)." \emph{Zeitschrift Für Kristallographie}, 150(1–4), 23–52.
\doi{10.1524/zkri.1979.150.14.23}

Shannon, R. D. (1976). "Revised Effective Ionic Radii and Systematic Studies of
Interatomic Distances in Halides and Chalcogenides." \emph{Acta Crystallographica
Section A}, 32(5), 751–767. \doi{10.1107/S0567739476001551}
}
\seealso{
Other bonding algorithms: 
\code{\link{brunner_nn_reciprocal}()},
\code{\link{crystal_nn}()},
\code{\link{minimum_distance}()},
\code{\link{voronoi_nn}()}
}
\concept{bonding algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calculate_properties.R
\name{calculate_distances}
\alias{calculate_distances}
\title{Calculate Interatomic Distances}
\usage{
calculate_distances(
  atomic_coordinates,
  expanded_coords,
  unit_cell_metrics,
  tolerance = 1e-06
)
}
\arguments{
\item{atomic_coordinates}{A \code{data.table} of the primary (asymmetric) atom set.}

\item{expanded_coords}{A \code{data.table} of atoms in the expanded supercell.}

\item{unit_cell_metrics}{A \code{data.table} with cell parameters.}

\item{tolerance}{A numeric cutoff (default 1e-6). Distances smaller than this
value are considered floating-point noise (overlapping atoms) and are
filtered out.}
}
\value{
A \code{data.table} of all non-zero distances.
}
\description{
Computes the distances between a central set of atoms and an
expanded set, using the metric tensor for accuracy.
}
\examples{
ac <- data.table::data.table(Label = "Si1", x_a = 0, y_b = 0, z_c = 0)
ec <- data.table::data.table(Label = "O1_1", x_a = 0.5, y_b = 0.5, z_c = 0.5)
uc <- data.table::data.table(`_cell_length_a` = 10, `_cell_length_b` = 10,
                             `_cell_length_c` = 10, `_cell_angle_alpha` = 90,
                             `_cell_angle_beta` = 90, `_cell_angle_gamma` = 90)
calculate_distances(ac, ec, uc)
}
\seealso{
Other property calculators: 
\code{\link{calculate_angles}()},
\code{\link{calculate_neighbor_counts}()},
\code{\link{calculate_weighted_neighbor_counts}()},
\code{\link{filter_atoms_by_symbol}()},
\code{\link{filter_by_elements}()},
\code{\link{filter_by_wyckoff_symbol}()}
}
\concept{property calculators}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_cif_files.R
\name{read_cif_files}
\alias{read_cif_files}
\title{Read CIF Files into Memory}
\usage{
read_cif_files(file_paths)
}
\arguments{
\item{file_paths}{A character vector of paths to the CIF files.}
}
\value{
A list of \code{data.table} objects.
}
\description{
Reads one or more CIF files from disk and loads each into a
\code{data.table}. It includes encoding sanitization to handle legacy characters.
}
\examples{
cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
if (file.exists(cif_path)) {
  cifs <- read_cif_files(cif_path)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calculate_properties.R
\name{filter_by_wyckoff_symbol}
\alias{filter_by_wyckoff_symbol}
\title{Filter Data by Wyckoff Symbol}
\usage{
filter_by_wyckoff_symbol(
  data_table,
  atomic_coordinates,
  atom_col,
  wyckoff_symbols
)
}
\arguments{
\item{data_table}{A \code{data.table} object, such as one produced by
\code{minimum_distance} or
\code{calculate_angles}.}

\item{atomic_coordinates}{A \code{data.table} from \code{extract_atomic_coordinates}
containing \code{WyckoffMultiplicity} and \code{WyckoffSymbol} columns.}

\item{atom_col}{A character string specifying the column in \code{data_table}
that contains the atom labels to filter by (e.g., "Atom1", "CentralAtom").}

\item{wyckoff_symbols}{A character vector of the full Wyckoff symbols to
keep (e.g., \code{c("4c")} or \code{c("6c", "16i", "24k")}).}
}
\value{
A \code{data.table} filtered to include only rows where the specified
atom occupies one of the desired Wyckoff sites.
}
\description{
Filters a data table (e.g., bonds or angles) to include only
entries where a specified atom occupies one of the given Wyckoff sites.
}
\details{
This function is designed to facilitate analysis based on crystallographic
site symmetry. It is particularly useful for analyzing complex structures like
clathrates where different atoms perform distinct structural roles based
on their site symmetry.

The function works by:
\enumerate{
\item Creating a full Wyckoff label (e.g., "4c", "24k") by combining the
\code{WyckoffMultiplicity} and \code{WyckoffSymbol} columns from the
\code{atomic_coordinates} table.
\item Identifying the parent atom for each entry in the \code{data_table}.
\item Merging this Wyckoff information into the results table.
\item Filtering to keep only rows where the atom's full Wyckoff label matches
one of the symbols provided in the \code{wyckoff_symbols} vector.
}
}
\examples{
cif_file <- system.file("extdata", "1590946.cif", package = "crystract")
if (file.exists(cif_file)) {
  old_threads <- data.table::setDTthreads(2)
  # 1. Perform a standard analysis to get bond and coordinate tables
  cif_content <- read_cif_files(cif_file)[[1]]
  atoms <- extract_atomic_coordinates(cif_content)
  metrics <- extract_unit_cell_metrics(cif_content)
  sym_ops <- extract_symmetry_operations(cif_content)
  full_cell <- apply_symmetry_operations(atoms, sym_ops, metrics)
  super_cell <- expand_transformed_coords(full_cell)
  dists <- calculate_distances(atoms, super_cell, metrics)
  bonds <- minimum_distance(dists)

  # 2. Mock Wyckoff sites since 1590946 doesn't explicitly declare them
  atoms[, WyckoffSymbol := "c"]
  atoms[, WyckoffMultiplicity := 4]

  print("Original atomic coordinates showing Wyckoff sites:")
  print(atoms[, .(Label, WyckoffSymbol, WyckoffMultiplicity)])

  filtered_bonds <- filter_by_wyckoff_symbol(
    data_table = bonds,
    atomic_coordinates = atoms,
    atom_col = "Atom1",
    wyckoff_symbols = "4c"
  )

  cat("\nNumber of bonds in original table:", nrow(bonds), "\n")
  cat("Number of bonds after filtering for '4c' site:", nrow(filtered_bonds), "\n")
  data.table::setDTthreads(old_threads)
}
}
\seealso{
Other property calculators: 
\code{\link{calculate_angles}()},
\code{\link{calculate_distances}()},
\code{\link{calculate_neighbor_counts}()},
\code{\link{calculate_weighted_neighbor_counts}()},
\code{\link{filter_atoms_by_symbol}()},
\code{\link{filter_by_elements}()}
}
\concept{property calculators}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_coordinates.R
\name{calculate_expansion_factors}
\alias{calculate_expansion_factors}
\title{Calculate Supercell Expansion Factors}
\usage{
calculate_expansion_factors(unit_cell_metrics, radius)
}
\arguments{
\item{unit_cell_metrics}{A \code{data.table} with cell parameters.}

\item{radius}{Numeric. The cutoff radius (e.g., 13.0 for Voronoi).}
}
\value{
A named numeric vector \code{c(a=..., b=..., c=...)}.
}
\description{
Computes the number of unit cell repetitions required in each
direction (a, b, c) to encompass a sphere of a given radius.
}
\examples{
uc <- data.table::data.table(`_cell_length_a` = 10, `_cell_length_b` = 10,
                             `_cell_length_c` = 10, `_cell_angle_alpha` = 90,
                             `_cell_angle_beta` = 90, `_cell_angle_gamma` = 90)
calculate_expansion_factors(uc, radius = 13.0)
}
\seealso{
Other coordinate processors: 
\code{\link{apply_symmetry_operations}()},
\code{\link{expand_transformed_coords}()}
}
\concept{coordinate processors}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_coordinates.R
\name{merge_sites_pbc}
\alias{merge_sites_pbc}
\title{Merge Close Atoms (Pymatgen Style)}
\usage{
merge_sites_pbc(atomic_coordinates, unit_cell_metrics, tol = 1e-04)
}
\arguments{
\item{atomic_coordinates}{A data.table with columns \code{Label}, \code{x_a}, \code{y_b}, \code{z_c}.}

\item{unit_cell_metrics}{A data.table containing cell lengths and angles.}

\item{tol}{Numeric. Distance tolerance in Angstroms. Sites closer than this will be merged.
Default is 1e-4.}
}
\value{
A deduplicated data.table with coordinates averaged.
}
\description{
Merges atomic sites that are within a specified distance tolerance,
accounting for Periodic Boundary Conditions (PBC). This mimics pymatgen's
\code{Structure.merge_sites} logic: it performs hierarchical clustering and then
averages the coordinates of the merged sites, taking care of PBC wrapping.
}
\examples{
ac <- data.table::data.table(Label = c("A", "B"),
                             x_a = c(0, 0.00001),
                             y_b = c(0, 0),
                             z_c = c(0, 0))
uc <- data.table::data.table(`_cell_length_a` = 10, `_cell_length_b` = 10,
                             `_cell_length_c` = 10, `_cell_angle_alpha` = 90,
                             `_cell_angle_beta` = 90, `_cell_angle_gamma` = 90)
merge_sites_pbc(ac, uc, tol = 1e-4)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_coordinates.R
\name{expand_transformed_coords}
\alias{expand_transformed_coords}
\title{Expand Coordinates into a Supercell}
\usage{
expand_transformed_coords(transformed_coords, expansion_factors = c(1, 1, 1))
}
\arguments{
\item{transformed_coords}{A \code{data.table} of atom positions within one unit cell.}

\item{expansion_factors}{A numeric vector \code{c(a, b, c)} specifying the number of
images in each direction (+/-). Default \code{c(1,1,1)} creates a 3x3x3 grid.}
}
\value{
A \code{data.table} containing all atomic positions in the expanded supercell.
}
\description{
Takes a set of atomic coordinates within a single unit cell and
replicates them into a supercell grid defined by expansion factors.
}
\examples{
tc <- data.table::data.table(Label = "A", x_a = 0, y_b = 0, z_c = 0)
expand_transformed_coords(tc, expansion_factors = c(1, 1, 1))
}
\seealso{
Other coordinate processors: 
\code{\link{apply_symmetry_operations}()},
\code{\link{calculate_expansion_factors}()}
}
\concept{coordinate processors}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calculate_properties.R
\name{propagate_angle_error}
\alias{propagate_angle_error}
\title{Propagate Angle Error}
\usage{
propagate_angle_error(
  bond_angles,
  atomic_coordinates,
  expanded_coords,
  unit_cell_metrics
)
}
\arguments{
\item{bond_angles}{Data.table of calculated bond angles.}

\item{atomic_coordinates}{Data.table with fractional coordinates and errors.}

\item{expanded_coords}{Data.table of supercell atom coordinates.}

\item{unit_cell_metrics}{Data.table with unit cell parameters and errors.}
}
\value{
The input \code{bond_angles} data.table with a new 'AngleError' column.
}
\description{
Calculates the standard uncertainty for each bond angle.
}
\examples{
# 1. Create dummy bond angles
ba <- data.table::data.table(
  CentralAtom = "Si1", Neighbor1 = "O1", Neighbor2 = "O2", Angle = 109.5
)

# 2. Create dummy atomic coordinates with errors
ac <- data.table::data.table(
  Label = c("Si1", "O1", "O2"),
  x_a = c(0, 0.1, 0), y_b = c(0, 0, 0.1), z_c = c(0, 0, 0),
  x_error = c(0.01, 0.01, 0.01),
  y_error = c(0.01, 0.01, 0.01),
  z_error = c(0.01, 0.01, 0.01)
)

# 3. Create dummy expanded coordinates
ec <- data.table::data.table(
  Label = c("O1", "O2"),
  x_a = c(0.1, 0), y_b = c(0, 0.1), z_c = c(0, 0)
)

# 4. Create dummy unit cell metrics
uc <- data.table::data.table(
  `_cell_length_a` = 10, `_cell_length_a_error` = 0.1,
  `_cell_length_b` = 10, `_cell_length_b_error` = 0.1,
  `_cell_length_c` = 10, `_cell_length_c_error` = 0.1,
  `_cell_angle_alpha` = 90, `_cell_angle_alpha_error` = 0,
  `_cell_angle_beta` = 90, `_cell_angle_beta_error` = 0,
  `_cell_angle_gamma` = 90, `_cell_angle_gamma_error` = 0
)

# 5. Run the error propagation
propagate_angle_error(ba, ac, ec, uc)
}
\seealso{
Other error propagators: 
\code{\link{propagate_distance_error}()}
}
\concept{error propagators}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_cif_files.R
\name{analyze_single_cif}
\alias{analyze_single_cif}
\title{Analyze the Content of a Single CIF File}
\usage{
analyze_single_cif(
  cif_content,
  file_name = NULL,
  perform_extraction = TRUE,
  perform_calcs_and_transforms = TRUE,
  bonding_algorithms = c("minimum_distance"),
  calculate_bond_angles = TRUE,
  perform_error_propagation = TRUE,
  tolerance = 1e-04,
  minimum_distance_delta = 0.1
)
}
\arguments{
\item{cif_content}{Either a \code{data.table} containing the lines of a CIF file,
OR a character string specifying the file path to a CIF file.}

\item{file_name}{The name of the original CIF file, used for labeling output.}

\item{perform_extraction}{Logical. If \code{TRUE}, extracts all metadata.}

\item{perform_calcs_and_transforms}{Logical. If \code{TRUE}, generates unit cell
and supercell.}

\item{bonding_algorithms}{Character vector. Options: \code{"minimum_distance"},
\code{"brunner"}, \code{"econ"}, \code{"crystal_nn"}, \code{"voronoi"}.}

\item{calculate_bond_angles}{Logical.}

\item{perform_error_propagation}{Logical.}

\item{tolerance}{Numeric. Cutoff for floating-point noise and atom merging (default 1e-4).}

\item{minimum_distance_delta}{Numeric. Tolerance for min-dist algorithm.}
}
\value{
A one-row \code{data.table} with results.
}
\description{
The core worker function that orchestrates the analysis pipeline
for a single crystal structure's data. It dynamically adjusts supercell
size and coordinate sets based on the requested bonding algorithms.
For geometric algorithms (Voronoi/CrystalNN), it automatically merges atoms
occupying the same site to ensure mathematical validity.
}
\examples{
cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
if (file.exists(cif_path)) {
  old_threads <- data.table::setDTthreads(2)
  res <- analyze_single_cif(cif_path, perform_error_propagation = FALSE)
  data.table::setDTthreads(old_threads)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/crystract-package.R
\docType{package}
\name{crystract-package}
\alias{crystract}
\alias{crystract-package}
\title{crystract: Tools to Extract Data from CIF Files for Crystallography}
\description{
Parse Crystallographic Information Files (.cif) and extract essential
data such as chemical formulas, unit cell parameters, atomic coordinates,
and symmetry operations. Calculate interatomic distances, identify bonded
pairs using various algorithms (Minimum Distance, Brunner's, EconNN,
CrystalNN, Voronoi), determine nearest neighbor counts, and calculate bond
angles. Facilitates the preparation of crystallographic data for further
analysis, including machine learning applications in materials science.
}
\references{
Shannon, R. D. (1976). "Revised effective ionic radii and systematic studies of
interatomic distances in halides and chalcogenides." \emph{Acta Crystallographica
Section A: Crystal Physics, Diffraction, Theoretical and General Crystallography},
32(5), 751-767.

Pauling, L. (1960). \emph{The Nature of the Chemical Bond and the Structure of
Molecules and Crystals: An Introduction to Modern Structural Chemistry} (Vol. 18).
Cornell University Press.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://prabhulab.github.io/ml-crystals/}
}

}
\author{
\strong{Maintainer}: Don Ngo \email{dngo@carnegiescience.edu} (\href{https://orcid.org/0009-0001-2779-2146}{ORCID})

Authors:
\itemize{
  \item Anirudh Prabhu \email{aprabhu@carnegiescience.edu} (\href{https://orcid.org/0000-0002-9921-6084}{ORCID})
  \item Julia-Maria Hubner \email{julia-maria.huebner@tu-dresden.de} (\href{https://orcid.org/0000-0003-2048-6629}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calculate_properties.R
\name{filter_atoms_by_symbol}
\alias{filter_atoms_by_symbol}
\title{Filter Data by Atom Symbol Interactively}
\usage{
filter_atoms_by_symbol(data_table, atom_col = "CentralAtom")
}
\arguments{
\item{data_table}{A \code{data.table} object containing atomic information, such as
the output from \code{calculate_angles} or \code{minimum_distance}.}

\item{atom_col}{A character string specifying the name of the column in
\code{data_table} that contains the atom labels to filter by. Defaults to "CentralAtom".}
}
\value{
A \code{data.table} filtered to include only the rows where the atom
label in \code{atom_col} corresponds to one of the user-selected chemical symbols.
If the user provides no input, an empty \code{data.table} is returned.
}
\description{
Prompts the user to select chemical elements to keep in a data table
of bonds or angles. Filtering is based on matching the base chemical symbol
in a specified column (e.g., "CentralAtom").
}
\details{
The function first identifies all unique base chemical symbols from the atom
labels in the specified column (e.g., it extracts 'C' from 'C1', 'Si' from 'Si2_1').
It then presents these symbols to the user and asks for a comma-separated list
of the symbols they wish to retain.

The matching logic is designed to be specific to avoid ambiguity between elements.
For example, if the user enters 'C', the function will match labels like 'C1',
'C2', 'C_10', or a lone 'C'. However, it will \emph{not} match labels for different
elements that start with C, such as 'Cr1' or 'Ca2'. This is achieved by
constructing a regular expression that ensures the character(s) immediately
following the selected symbol are not alphabetical letters.

This function is intended for interactive use.
}
\examples{
# 1. Create a sample data.table of bond angles
sample_angles <- data.table::data.table(
  CentralAtom = c("C1", "C2", "Si1", "Cr1", "O1", "O2", "C"),
  Neighbor1 = c("O1", "O2", "O1", "N1", "C1", "C2", "H1"),
  Neighbor2 = c("H1", "H2", "O2", "N2", "H3", "H4", "H2"),
  Angle = c(109.5, 109.5, 120.0, 90.0, 104.5, 104.5, 120)
)

# 2. In an interactive R session, the function would prompt the user.
if (interactive()) {
  filtered_data <- filter_atoms_by_symbol(sample_angles, atom_col = "CentralAtom")
  print(filtered_data)
}

}
\seealso{
Other property calculators: 
\code{\link{calculate_angles}()},
\code{\link{calculate_distances}()},
\code{\link{calculate_neighbor_counts}()},
\code{\link{calculate_weighted_neighbor_counts}()},
\code{\link{filter_by_elements}()},
\code{\link{filter_by_wyckoff_symbol}()}
}
\concept{property calculators}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/radii_management.R
\name{set_radii_data}
\alias{set_radii_data}
\title{Set or Reset a Custom Atomic Radii Table}
\usage{
set_radii_data(radii_data = NULL)
}
\arguments{
\item{radii_data}{A \code{data.table} or \code{data.frame} containing the custom radii
data, or \code{NULL} to reset to the package default.}
}
\value{
Invisibly returns \code{NULL}. A message is printed to the console
confirming the action.
}
\description{
Allows the user to provide their own table of atomic radii for
the current R session. This custom table will be used by functions like
\code{filter_ghost_distances}.
}
\details{
The provided \code{data.table} or \code{data.frame} must contain at least three columns:
\itemize{
\item \code{Symbol} (character): The chemical symbol of the element (e.g., "Si").
\item \code{Radius} (numeric): The atomic radius in Angstroms (Å).
\item \code{Type} (character): A descriptor for the radius type (e.g., "covalent", "ionic").
}
This allows for storing multiple types of radii in the same table, which can
be selected using the \code{radii_type} argument in relevant functions.

To revert to using the package's default radii table, call the function
with \code{NULL} or without any arguments.
}
\examples{
# 1. Create a custom radii table with both covalent and ionic radii
my_radii <- data.table::data.table(
  Symbol = c("Si", "O", "O"),
  Radius = c(1.11, 0.66, 1.40),
  Type   = c("covalent", "covalent", "ionic")
)

# 2. Set this table for the current session
set_radii_data(my_radii)

# 3. Reset to the package's default covalent radii table
set_radii_data(NULL)

}
\seealso{
Other post-processing: 
\code{\link{calculate_weighted_average_network_distance}()},
\code{\link{filter_ghost_distances}()}
}
\concept{post-processing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calculate_properties.R
\name{calculate_angles}
\alias{calculate_angles}
\title{Calculate Bond Angles}
\usage{
calculate_angles(
  bonded_pairs,
  atomic_coordinates,
  expanded_coords,
  unit_cell_metrics
)
}
\arguments{
\item{bonded_pairs}{Data.table of bonded atoms.}

\item{atomic_coordinates}{Data.table of asymmetric atom coordinates.}

\item{expanded_coords}{Data.table of supercell atom coordinates.}

\item{unit_cell_metrics}{Data.table with unit cell parameters.}
}
\value{
A \code{data.table} of all unique bond angles.
}
\description{
Calculates all bond angles centered on each atom, formed by pairs
of its bonded neighbors.
}
\examples{
bp <- data.table::data.table(Atom1 = c("Si1", "Si1"), Atom2 = c("O1", "O2"))
ac <- data.table::data.table(Label = "Si1", x_a = 0, y_b = 0, z_c = 0)
ec <- data.table::data.table(Label = c("O1", "O2"),
                             x_a = c(1, 0), y_b = c(0, 1), z_c = c(0, 0))
uc <- data.table::data.table(`_cell_length_a` = 10, `_cell_length_b` = 10,
                             `_cell_length_c` = 10, `_cell_angle_alpha` = 90,
                             `_cell_angle_beta` = 90, `_cell_angle_gamma` = 90)
calculate_angles(bp, ac, ec, uc)
}
\seealso{
Other property calculators: 
\code{\link{calculate_distances}()},
\code{\link{calculate_neighbor_counts}()},
\code{\link{calculate_weighted_neighbor_counts}()},
\code{\link{filter_atoms_by_symbol}()},
\code{\link{filter_by_elements}()},
\code{\link{filter_by_wyckoff_symbol}()}
}
\concept{property calculators}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict_bonds.R
\name{minimum_distance}
\alias{minimum_distance}
\title{Identify Atomic Bonds using the Minimum Distance Method}
\usage{
minimum_distance(distances, delta = 0.1)
}
\arguments{
\item{distances}{A \code{data.table} of interatomic distances from \code{calculate_distances}.}

\item{delta}{The relative tolerance parameter (default 0.1).}
}
\value{
A \code{data.table} of bonded pairs.
}
\description{
Identifies bonded atoms by finding the nearest neighbor distance (d_min)
for each central atom and defining a cutoff distance (d_cut) as
d_cut = (1 + delta) * d_min.
}
\examples{
dists <- data.table::data.table(Atom1 = c("A", "A"), Atom2 = c("B", "C"),
                                Distance = c(1.5, 2.5),
                                DeltaX = c(1, 0), DeltaY = c(0, 1), DeltaZ = c(0, 0))
minimum_distance(dists, delta = 0.1)
}
\seealso{
Other bonding algorithms: 
\code{\link{brunner_nn_reciprocal}()},
\code{\link{crystal_nn}()},
\code{\link{econ_nn}()},
\code{\link{voronoi_nn}()}
}
\concept{bonding algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_cif_files.R
\name{aggregate_batch_results}
\alias{aggregate_batch_results}
\title{Aggregate Batched Analysis Results}
\usage{
aggregate_batch_results(input_dir, cols_to_keep = NULL)
}
\arguments{
\item{input_dir}{The directory containing RDS files from \code{analyze_cif_files}.}

\item{cols_to_keep}{Optional character vector of column names to keep.}
}
\value{
A single \code{data.table} containing the aggregated results.
}
\description{
Reads all \code{batch_*.rds} files from a directory and combines them.
}
\examples{
\donttest{
cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
if (file.exists(cif_path)) {
  old_threads <- data.table::setDTthreads(2)
  out_dir <- file.path(tempdir(), "cif_batches")
  analyze_cif_files(cif_path, output_dir = out_dir)

  aggr <- aggregate_batch_results(out_dir)

  unlink(out_dir, recursive = TRUE)
  data.table::setDTthreads(old_threads)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calculate_properties.R
\name{filter_ghost_distances}
\alias{filter_ghost_distances}
\title{Filter Ghost Distances Using Atomic Radii}
\usage{
filter_ghost_distances(
  distances,
  atomic_coordinates,
  margin = 0.1,
  radii_type = "covalent"
)
}
\arguments{
\item{distances}{A \code{data.table} of interatomic distances, typically from
\code{calculate_distances}. Must contain \code{Atom1}, \code{Atom2}, and \code{Distance}.}

\item{atomic_coordinates}{A \code{data.table} of asymmetric atoms from
\code{extract_atomic_coordinates}. Used to link atom labels to element types.}

\item{margin}{A numeric value (default 0.1) specifying the tolerance. A
distance \code{d} between atoms with radii \code{r1} and \code{r2} is kept if
\verb{(r1+r2)*(1-margin) <= d <= (r1+r2)*(1+margin)}.}

\item{radii_type}{A character string specifying the type of radius to use for
the calculation (e.g., \code{"covalent"}, \code{"ionic"}). This value must correspond
to an entry in the \code{Type} column of the active radii table. Defaults to \code{"covalent"}.
The radii table can be customized for the session using \code{set_radii_data()}.}
}
\value{
A list containing two \code{data.table}s:
\item{kept}{The distances considered physically plausible.}
\item{removed}{The "ghost" distances that were filtered out, with columns
explaining the reason for removal.}
}
\description{
Cleans a distance table by removing physically implausible distances.
It uses a table of atomic radii to establish a plausible bond length
range for each atom pair. Any calculated distance falling outside this range
(defined by a \code{margin}) is considered a "ghost" distance and is removed. This
is particularly useful for cleaning data from disordered crystal structures.
}
\examples{
# Create minimal dummy data for demonstration
distances <- data.table::data.table(
  Atom1 = c("Si1_1_0_0", "O1_1_0_0"),
  Atom2 = c("O1_1_0_0", "Si1_1_0_0"),
  Distance = c(1.6, 0.5) # 0.5 is implausibly short
)

atomic_coords <- data.table::data.table(
  Label = c("Si1", "O1")
)

# Run the filter
result <- filter_ghost_distances(distances, atomic_coords, margin = 0.1)

print(result$kept)
print(result$removed)
}
\seealso{
Other post-processing: 
\code{\link{calculate_weighted_average_network_distance}()},
\code{\link{set_radii_data}()}
}
\concept{post-processing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_cif_files.R
\name{analyze_cif_files}
\alias{analyze_cif_files}
\title{Analyze a Batch of CIF Files}
\usage{
analyze_cif_files(
  file_paths,
  workers = 1,
  output_dir = NULL,
  batch_size = 1000,
  ...
)
}
\arguments{
\item{file_paths}{Character vector of paths or list of data.tables.}

\item{workers}{Integer. Number of parallel workers.}

\item{output_dir}{Path to output directory (optional).}

\item{batch_size}{Integer.}

\item{...}{Args passed to \code{analyze_single_cif}.}
}
\value{
Data.table or output directory path.
}
\description{
A high-level wrapper that analyzes CIF files in batch, supporting
parallel processing and batch-to-disk operations.
}
\examples{
\donttest{
cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
if (file.exists(cif_path)) {
  old_threads <- data.table::setDTthreads(2)
  out_dir <- file.path(tempdir(), "cif_analysis")
  res <- analyze_cif_files(cif_path, output_dir = out_dir, workers = 1)

  # Clean up
  unlink(out_dir, recursive = TRUE)
  data.table::setDTthreads(old_threads)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calculate_properties.R
\name{propagate_distance_error}
\alias{propagate_distance_error}
\title{Propagate Distance Error}
\usage{
propagate_distance_error(bonded_pairs, atomic_coordinates, unit_cell_metrics)
}
\arguments{
\item{bonded_pairs}{Data.table of bonded atoms with their distances.}

\item{atomic_coordinates}{Data.table with fractional coordinates and errors.}

\item{unit_cell_metrics}{Data.table with unit cell parameters and errors.}
}
\value{
The input \code{bonded_pairs} data.table with a new 'DistanceError' column.
}
\description{
Calculates the standard uncertainty for each interatomic distance.
}
\examples{
bp <- data.table::data.table(Atom1 = "Si1", Atom2 = "O1_1", Distance = 1.6,
                             DeltaX = 0.1, DeltaY = 0.1, DeltaZ = 0)
ac <- data.table::data.table(Label = c("Si1", "O1"),
                             x_error = c(0.01, 0.01),
                             y_error = c(0.01, 0.01),
                             z_error = c(0.01, 0.01))
uc <- data.table::data.table(`_cell_length_a` = 10, `_cell_length_a_error` = 0.1,
                             `_cell_length_b` = 10, `_cell_length_b_error` = 0.1,
                             `_cell_length_c` = 10, `_cell_length_c_error` = 0.1,
                             `_cell_angle_alpha` = 90, `_cell_angle_alpha_error` = 0,
                             `_cell_angle_beta` = 90, `_cell_angle_beta_error` = 0,
                             `_cell_angle_gamma` = 90, `_cell_angle_gamma_error` = 0)
propagate_distance_error(bp, ac, uc)
}
\seealso{
Other error propagators: 
\code{\link{propagate_angle_error}()}
}
\concept{error propagators}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_cif_data.R
\name{extract_space_group_number}
\alias{extract_space_group_number}
\title{Extract Space Group Number from CIF Content}
\usage{
extract_space_group_number(cif_content)
}
\arguments{
\item{cif_content}{A \code{data.table} containing the lines of a CIF file.}
}
\value{
A character string of the space group number, or \code{NA} if not found.
}
\description{
Extracts the International Tables space group number (e.g., from \verb{_space_group_IT_number}).
}
\examples{
cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
if (file.exists(cif_path)) {
  cif_content <- read_cif_files(cif_path)[[1]]
  extract_space_group_number(cif_content)
}
}
\seealso{
Other extractors: 
\code{\link{extract_atomic_coordinates}()},
\code{\link{extract_chemical_formula}()},
\code{\link{extract_database_code}()},
\code{\link{extract_space_group_name}()},
\code{\link{extract_structure_type}()},
\code{\link{extract_symmetry_operations}()},
\code{\link{extract_unit_cell_metrics}()}
}
\concept{extractors}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_cif_data.R
\name{extract_atomic_coordinates}
\alias{extract_atomic_coordinates}
\title{Extract Atomic Coordinates}
\usage{
extract_atomic_coordinates(cif_content, chemical_formula = NA)
}
\arguments{
\item{cif_content}{A \code{data.table} containing the lines of a CIF file.}

\item{chemical_formula}{The chemical formula string for validation.}
}
\value{
A \code{data.table} with atomic coordinate data, or \code{NULL} if not found.
}
\description{
Parses atomic site info, including labels, fractional coordinates,
occupancies, oxidation states, and thermal parameters. It includes logic
to extract oxidation states from both site tags and type loops.
}
\examples{
cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
if (file.exists(cif_path)) {
  cif_content <- read_cif_files(cif_path)[[1]]
  extract_atomic_coordinates(cif_content)
}
}
\seealso{
Other extractors: 
\code{\link{extract_chemical_formula}()},
\code{\link{extract_database_code}()},
\code{\link{extract_space_group_name}()},
\code{\link{extract_space_group_number}()},
\code{\link{extract_structure_type}()},
\code{\link{extract_symmetry_operations}()},
\code{\link{extract_unit_cell_metrics}()}
}
\concept{extractors}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict_bonds.R
\name{voronoi_nn}
\alias{voronoi_nn}
\title{Identify Atomic Bonds using Voronoi Tessellation}
\usage{
voronoi_nn(
  atomic_coordinates,
  expanded_coords,
  unit_cell_metrics,
  cutoff = 13,
  tol = 0
)
}
\arguments{
\item{atomic_coordinates}{A \code{data.table} of the primary (asymmetric) atom set.}

\item{expanded_coords}{A \code{data.table} of atoms in the expanded supercell.}

\item{unit_cell_metrics}{A \code{data.table} with cell parameters.}

\item{cutoff}{Distance cutoff (default 13.0).}

\item{tol}{Tolerance for solid angle weights (default 0).}
}
\value{
A \code{data.table} of bonded pairs.
}
\description{
Performs 3D Voronoi analysis on the supercell.
}
\examples{
\donttest{
cif_path <- system.file("extdata", "1590946.cif", package = "crystract")
if (file.exists(cif_path)) {
  old_threads <- data.table::setDTthreads(2)
  res <- analyze_single_cif(cif_path, bonding_algorithms = "voronoi")
  print(head(res$bonds_voronoi[[1]]))
  data.table::setDTthreads(old_threads)
}
}
}
\references{
O'Keeffe, M. (1979). "A Proposed Rigorous Definition of Coordination Number."
\emph{Acta Crystallographica Section A}, 35(5), 772–775. \doi{10.1107/S0567739479001765}

Aurenhammer, F., Klein, R., & Lee, D.-T. (2013). \emph{Voronoi Diagrams and
Delaunay Triangulations}. World Scientific Publishing Company.
}
\seealso{
Other bonding algorithms: 
\code{\link{brunner_nn_reciprocal}()},
\code{\link{crystal_nn}()},
\code{\link{econ_nn}()},
\code{\link{minimum_distance}()}
}
\concept{bonding algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calculate_properties.R
\name{filter_by_elements}
\alias{filter_by_elements}
\title{Filter Distances by Element Symbols}
\usage{
filter_by_elements(distances, atomic_coordinates, elements_to_exclude)
}
\arguments{
\item{distances}{A \code{data.table} of interatomic distances.}

\item{atomic_coordinates}{A \code{data.table} of atomic coordinates.}

\item{elements_to_exclude}{A character vector of element symbols to exclude.}
}
\value{
A \code{data.table} of distances with the specified elements removed.
}
\description{
Removes any distance pair where at least one of the atoms
corresponds to a specified list of element symbols.
}
\examples{
dists <- data.table::data.table(Atom1 = "Si1_1_0_0", Atom2 = "O1_1_0_0", Distance = 1.6)
ac <- data.table::data.table(Label = c("Si1", "O1"))
filter_by_elements(dists, ac, elements_to_exclude = "O")
}
\seealso{
Other property calculators: 
\code{\link{calculate_angles}()},
\code{\link{calculate_distances}()},
\code{\link{calculate_neighbor_counts}()},
\code{\link{calculate_weighted_neighbor_counts}()},
\code{\link{filter_atoms_by_symbol}()},
\code{\link{filter_by_wyckoff_symbol}()}
}
\concept{property calculators}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calculate_properties.R
\name{calculate_weighted_average_network_distance}
\alias{calculate_weighted_average_network_distance}
\title{Calculate Weighted Average Network Bond Distance}
\usage{
calculate_weighted_average_network_distance(
  distances,
  atomic_coordinates,
  wyckoff_symbols
)
}
\arguments{
\item{distances}{A \code{data.table} of interatomic distances filtered to include
\strong{only bonded pairs} (e.g., from \code{minimum_distance}).}

\item{atomic_coordinates}{A \code{data.table} of asymmetric atoms from
\code{extract_atomic_coordinates}.}

\item{wyckoff_symbols}{A character vector of Wyckoff symbols defining the
atomic network (e.g., \code{c("6c", "16i", "24k")}). Must be the full symbol.}
}
\value{
A single numeric value representing the weighted average bond distance.
}
\description{
Computes a single, representative bond length for a specified atomic
network. This function precisely implements the validated logic that accounts
for site multiplicity and occupancy of the central atoms.
}
\examples{
dists <- data.table::data.table(Atom1 = "Si1", Atom2 = "O1", Distance = 1.6)
ac <- data.table::data.table(Label = c("Si1", "O1"),
                             WyckoffMultiplicity = c(4, 4),
                             WyckoffSymbol = c("c", "c"),
                             Occupancy = c(1, 1))
calculate_weighted_average_network_distance(dists, ac, wyckoff_symbols = "4c")
}
\seealso{
Other post-processing: 
\code{\link{filter_ghost_distances}()},
\code{\link{set_radii_data}()}
}
\concept{post-processing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_coordinates.R
\name{apply_symmetry_operations}
\alias{apply_symmetry_operations}
\title{Apply Symmetry Operations to Generate a Full Unit Cell}
\usage{
apply_symmetry_operations(
  atomic_coordinates,
  symmetry_operations,
  unit_cell_metrics,
  tolerance = 1e-04
)
}
\arguments{
\item{atomic_coordinates}{A \code{data.table} of asymmetric atoms.}

\item{symmetry_operations}{A \code{data.table} of operations.}

\item{unit_cell_metrics}{A \code{data.table} of cell parameters (needed for distance merging).}

\item{tolerance}{Numeric. The distance tolerance in Angstroms for merging close atoms.
Default is 1e-4.}
}
\value{
A \code{data.table} containing all unique atomic positions in the unit cell.
}
\description{
Generates all symmetry-equivalent atomic positions within the unit cell.
}
\examples{
ac <- data.table::data.table(Label = "A", x_a = 0, y_b = 0, z_c = 0)
ops <- data.table::data.table(x = c("x", "-x"), y = c("y", "-y"), z = c("z", "-z"))
uc <- data.table::data.table(`_cell_length_a` = 10, `_cell_length_b` = 10,
                             `_cell_length_c` = 10, `_cell_angle_alpha` = 90,
                             `_cell_angle_beta` = 90, `_cell_angle_gamma` = 90)
apply_symmetry_operations(ac, ops, uc)
}
\seealso{
Other coordinate processors: 
\code{\link{calculate_expansion_factors}()},
\code{\link{expand_transformed_coords}()}
}
\concept{coordinate processors}
